/***********************************************************************
*
* sim980.c - Simulator for the TI 980 computer.
*
* Changes:
*   06/23/11   DGP   Original
*   09/01/11   DGP   Change register names R0-R7 to match arch, A, E, ...
*   09/14/11   DGP   Added cpu models and Calder 980 support.
*   10/25/11   DGP   Added CDD Common Data Feature.
*   08/23/16   DGP   Added SCSI disk support.
*   08/24/16   DGP   Revamp console support. Misc. fixes.
*
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#if defined(UNIX)
#include <unistd.h>
#endif

#include "simdef.h"
#include "screen.h"
#include "simdsk.h"

uint32 breakaddr;		/* Break point address */
uint32 watchaddr;		/* Watch point address */
int pancount;			/* Instructions executed for panel */
unsigned long instcount;	/* Instructions executed */
unsigned long mipcount;		/* Instructions executed for MIP calc */
unsigned long tracestart = 0;	/* Trace instruction count start */

int run = FALSE;		/* CPU running */
int idle = FALSE;		/* CPU running IDLE */
int runled = FALSE;		/* CPU running LED */
int idleled = FALSE;		/* CPU idle LED */
int idlenum = 0;		/* CPU idle number */
int enablepanel = FALSE;	/* Enable panel */
int showstatus = FALSE;		/* Show CPU status */
int breakflag = FALSE;		/* No breakpoint enabled */
int watchflag = FALSE;		/* No watch point enabled */
int traceenable = FALSE;	/* No trace enabled */
int traceit = FALSE;		/* No trace enabled */
int trclatch = FALSE;		/* Error Trace latch */
int errindex = 0;		/* Error Trace Memory index */
int trcindex = 0;		/* Error Trace Memory locked index */
int devcnt = 0;			/* Attached device count */
int deferint = 0;		/* Defer interrupts inst count */
int tracelim[2];		/* Trace memory limits */
int windowlen = 25;		/* Window length */
int viewlen;			/* View length */
int prtviewlen;			/* prtview length */
int prtviewcol;			/* prtview column */
int usertclock = FALSE;		/* Use Real Time clock */
int dispaddr = 0;		/* Display area address */
int model = CPU_TI980B;		/* CPU model */

uint16 curinst = 0;		/* Current instruction */
uint16 ssw;			/* Sense Switches */
uint16 statreg;			/* The program status register */

uint32 intrequest = 0;		/* Interrupt requests */
uint32 dmadevint = 0;		/* DMA device interrupts */
uint32 memlen = MEMSIZE;	/* Length of system memory */
uint32 lowlimitreg = 0;		/* PIF/MP Low limit register */
uint32 upperlimitreg = 0xFFFF;	/* PIF/MP Upper limit register */

int cddcmnenable = 0;		/* CDD Common enable */
uint16 cddcmnlimit = 0;		/* CDD Common Limit */
uint32 cddcmnbank = 0;		/* CDD Common Bank */

uint32 cddlowlimitext = 0;	/* CDD PIF/MP Low limit register extension */
uint32 cddupperlimitext = 0;	/* CDD PIF/MP Upper limit register extension */
uint32 cpumemmap[2];		/* CDD980 CPU memory map */
uint32 dmamemmap[8];		/* CDD980 DMA memory map */

long delayclock = 0;		/* Delay clock interval */

char view[MAXVIEW][MAXVIEWLEN+1]; /* Messages panel */
char prtview[MAXVIEW][MAXVIEWLEN+1]; /* Print messages panel */
uint16 regs[16];		/* The Register file */
uint16 memory[MAXMEMSIZE];	/* The CPU memory */
Device devices[MAXDEVICES];	/* The attached devices */
FILE *tracefd = NULL;		/* Trace file fd */
int traceline = 60;		/* Trace file line */

int trcidx = 0;			/* Inst trace index */
int trcopr = 0;			/* Inst trace operand index */
TraceInst trcbuf[TRCBUFMAX];	/* Inst trace buffer */

static int dispcnt  = 0;	/* Display count */
static int dismaddr = 0;	/* Disassembly address */
static int dismcnt = 0;		/* Disassembly count */
static uint32 displen;
static char input[80];
static char oinput[80];

static char *cpu_models[] = 
{
   "980", "980A", "980B", "CDD-980"
};

typedef struct 
{
   char command;
   char *brief;
   char *helpmsg;
} HelpMsg;

static HelpMsg helpmsgs[] =
{
   { 'a', "attach",
     "attach device: a device[/switches] devaddr [dmaport] dev.file" },
   { 'B', "begin",
     "begin trace pc (default 0): B location" },
   { 'b', "boot",
     "boot from device: b[n] device" },
   { 'c', "clr b|w",
     "clear breakpoint/watchpoint: c b|w" },
   { 'D', "disasm",
     "disassemble instructions: D location [count]" },
   { 'd', "display",
     "display memory: d[b] location [length]" },
   { 'E', "end",
     "End trace pc (default >FFFF): E location" },
   { 'g', "go",
     "go (run program)" },
   { 'h', "help",
     "help: h [command]" },
   { 'i', "insts",
     "instructions, dumps instruction buffer (last 60 instructions)" },
   { 'k', "break",
     "set breakpoint: k location" },
   { 'l', "load",
     "load program: l[n] [loadaddr] object.file" },
   { 'm', "modify",
     "modify memory: m location newvalue" },
   { 'N', "nxt dis",
     "disassemble next instructions: N [count]" },
   { 'n', "next",
     "display next memory location: n [length]" },
   { 'p', "set pc",
     "set pc: p location" },
   { 'q', "quit",
     "quit the simulator" },
   { 'R', "reset",
     "reset the system" },
   { 'r', "reg d|m",
     "register display/modify: r d|m reg [newvalue]" },
   { 's', "step",
     "single step program" },
   { 't', "trace",
     "trace program: t [count] trace.file" },
   { 'w', "switch",
     "switch toggle: w n" },
   { 'W', "watch",
     "set watchpoint: W location" },
   { 'z', "panel",
     "toggle panel mode: z [windowlen]" },
   { 0, NULL, NULL }
};

/***********************************************************************
* sigintr - Process interrupt signal.
***********************************************************************/

void
sigintr (int sig)
{
   run = FALSE;
   showstatus = TRUE;
   signal (SIGINT, sigintr);
}

/***********************************************************************
* puthelp - Put help on screen.
***********************************************************************/

static void
puthelp (char *bp)
{
   int i, j;
   int pall = FALSE;
   int psum = TRUE;

   if (bp != NULL)
   {
      while (*bp && isspace (*bp)) bp++;
      if (*bp == '*')
         pall = TRUE;
      for (i = j = 0; helpmsgs[i].command; i++)
      {
	 if (pall || (*bp == helpmsgs[i].command))
	 {
	    snprintf (view[j], MAXVIEWLEN, "%c - %s",
		     helpmsgs[i].command, helpmsgs[i].helpmsg);
	    psum = FALSE;
	    if (!pall) break;
	    j++;
	    if (j == MAXVIEW)
	    {
	       putview (TRUE, TRUE);
	       j = 0;
	    }
	 }
      }
   }
   if (psum)
   {
      for (j = i = 0; helpmsgs[i].command; i++)
      {
	 char phrase[32];

	 if (i % 6 == 0)
	 {
	    if (i) j++;
	    view[j][0] = 0;
	 }
	 sprintf (phrase, "%c - %-8s ",
		  helpmsgs[i].command, helpmsgs[i].brief);
         strcat (view[j], phrase);
      }
   }
}

/***********************************************************************
* getnum - Get a number from the command.
***********************************************************************/

char *
getnum (char *bp, long *val)
{
   char *cp;
   long i = 0;

   while (*bp && isspace (*bp)) bp++;
   cp = bp;
   if (*bp == HEXSYM)
   {
      i = strtol (++bp, &cp, 16);
   }
   else if (isdigit (*bp) || *bp == '-' || *bp == '+')
   {
      i = strtol (bp, &cp, 10);
   }
   *val = i;
   return (cp);
}

/***********************************************************************
* panel - Put up the panel.
***********************************************************************/

void
panel (void)
{
   char idlestr[12];

   if (idleled)
      sprintf (idlestr, "IDLE: %d ", idlenum);
   else
      strcpy (idlestr, "         ");

   if (enablepanel)
   {
      screenposition (enablepanel, TOPLINE, 1);
      clearline (enablepanel);
      printf ("%s %s   %s CPU %dK",
	      INTRO, VERSION, cpu_models[model], displen/1024);

      screenposition (enablepanel, STATUSLINE, 1);
      printf ("Status: %04X  SSW: %d%d%d%d  %s",
	      statreg, (ssw >> 3) & 1, (ssw >> 2) & 1, (ssw >> 1) & 1, ssw & 1,
	      idlestr);

      screenposition (enablepanel, BREAKLINE,1);
      printf ("Break: %04X  ", breakaddr);
      printf (" Watch: %04X  ", watchaddr);

      screenposition (enablepanel, REGISTERLINE,1);
      printf (" A %04X  E %04X  X %04X  M %04X", R0, R1, R2, R3);
      screenposition (enablepanel, REGISTERLINE+1,1);
      printf (" S %04X  L %04X  B %04X  P %04X", R4, R5, R6, R7);
      if (model == CPU_CDD980)
      {
	 screenposition (enablepanel, REGISTERLINE+2,1);
	 printf (" 8 %04X  9 %04X 10 %04X 11 %04X", RF0, RF1, RF2, RF3);
	 screenposition (enablepanel, REGISTERLINE+3,1);
	 printf ("12 %04X 13 %04X 14 %04X 15 %04X", RF4, RF5, RF6, RF7);
      }
      screenposition (enablepanel, OUTPUTLINE + prtviewlen - 1, prtviewcol);
      fflush (stdout);
      pancount = FALSE;
   }
   else if (showstatus)
   {
      printf ("\nStatus: %04X  SSW: %d%d%d%d %s\n",
	      statreg, (ssw >> 3) & 1, (ssw >> 2) & 1, (ssw >> 1) & 1, ssw & 1,
	      idlestr);
      printf ("Break: %04X  ", breakaddr);
      printf (" Watch: %04X\n", watchaddr);
      printf (" A %04X  E %04X  X %04X  M %04X\n", R0, R1, R2, R3);
      printf (" S %04X  L %04X  B %04X  P %04X\n", R4, R5, R6, R7);
      if (model == CPU_CDD980)
      {
	 printf (" 8 %04X  9 %04X 10 %04X 11 %04X\n", RF0, RF1, RF2, RF3);
	 printf ("12 %04X 13 %04X 14 %04X 15 %04X\n", RF4, RF5, RF6, RF7);
      }
   }
}

/***********************************************************************
* clearall - Clear out system.
***********************************************************************/

static int 
clearall ()
{
   int i;

   /*
   ** Clear registers
   */

   memset (&regs, 0, sizeof (regs));
   ssw = 0;
   statreg = 0;

   /*
   ** Clear memory
   */

   memset (&memory, 0, sizeof (memory));
   
   instcount = 0;
   intrequest = 0;
   lowlimitreg = 0;
   upperlimitreg = 0xFFFF;

   /*
   ** Clear out display panels.
   */

   memset (prtview, 0, sizeof(prtview));
   memset (view, 0, sizeof (view));

   trcidx = 0;
   for (i = 0; i < TRCBUFMAX; i++)
   {
      trcbuf[i].pc = -1;
   }

   /*
   ** CDD 980 "registers"
   */

   cddlowlimitext = 0;
   cddupperlimitext = 0;
   cddcmnenable = 0;
   cddcmnlimit = 0;
   cddcmnbank = 0;
   memset (&cpumemmap, 0, sizeof (cpumemmap));
   memset (&dmamemmap, 0, sizeof (dmamemmap));

   return (0);
}

/***********************************************************************
* printinst - Print an instruction with disassembly.
***********************************************************************/

void
printinst (char *outbuf, int index)
{
   char opr[12];

   sprintf (outbuf, "%04X %04X:  %04X ", 
	    trcbuf[index].pc, trcbuf[index].st, trcbuf[index].inst);

   if (trcbuf[index].ops[0] != -1)
   {
      sprintf (opr, "%04X ", trcbuf[index].ops[0] & 0xFFFF);
      strcat (outbuf, opr);
   }
   else
      strcat (outbuf, "     ");

   strcat (outbuf, "  ");
   disinst (&trcbuf[index]);
   strcat (outbuf, trcbuf[index].dis);
}

/***********************************************************************
* printsingle - Print a single instruction with disassembly.
***********************************************************************/

static void
printsingle (char *outbuf)
{
   int daddr;
   int insttype;
   TraceInst inst;
   char opr[12];

   daddr = dismaddr;

   inst.pc = daddr;
   inst.st = 0;
   inst.inst = memory[dismaddr];
   inst.ops[0] = -1;
   insttype = decodeinst (inst.inst);
   switch (insttype)
   {
   case TYPE_1:
      {
	 uint16 ixb;
	 uint16 disp;

	 ixb = (inst.inst & 0x0700) >> 8;
	 disp = (inst.inst & 0x00FF);
	 if ((disp == 0) && ((ixb == 0) || (ixb == 4) || (ixb == 6)))
	    inst.ops[0] = memory[dismaddr + 1];
      }
      break;
   case TYPE_5:
   case TYPE_7:
      inst.ops[0] = memory[dismaddr + 1];
      break;
   case TYPE_6:
      {
         uint16 op = ((inst.inst & 0x00F0) >> 4);
	 switch (op)
	 {
	 case 2:
	 case 3:
	 case 6:
	 case 7:
	    inst.ops[0] = memory[dismaddr + 1];
	    break;
	 default: ;
	 }
      }
      break;
   default: ;
   }
   disinst (&inst);
   sprintf (outbuf, "%04X   %04X ", daddr, inst.inst);
   dismaddr++;
   if (inst.ops[0] != -1)
   {
      sprintf (opr, "%04X ", inst.ops[0]);
      strcat (outbuf, opr);
      dismaddr++;
   }
   else
      strcat (outbuf, "     ");

   strcat (outbuf, inst.dis);
   if (dismaddr >= memlen)
      dismaddr = 0;
}

/***********************************************************************
* dumpsingle - Dump a single line.
***********************************************************************/

void
dumpsingle (char *outbuf)
{
   int i, j;
   char opr[8];
   char chd[32];

   sprintf (outbuf, "%05X   ", dispaddr);
   memset (chd, ' ', sizeof(chd));
   for (j = 2, i = 0; i < 8; i++, j += 3)
   {
      uint16 wd = memory[dispaddr];
      uint8 ch;

      sprintf (opr, "%04X ", wd);
      ch = (wd >> 8) & 0xFF;
      if (ch >= 0x20 && ch < 0x7F) chd[j] = ch;
      else chd[j] = '.';
      ch = wd & 0xFF;
      if (ch >= 0x20 && ch < 0x7F) chd[j+1] = ch;
      else chd[j+1] = '.';
      strcat (outbuf, opr);
      dispaddr += 1;
      if (dispaddr >= memlen)
         dispaddr = 0;
   }
   chd[j] = 0;
   strcat (outbuf, chd);
}

/***********************************************************************
* docommand - Do user commands.
***********************************************************************/

static int
docommand (int interactive)
{
   char *bp;
   long val, val1;
   int action;
   int i, j, k;
   int norun;
   char subcmd;

   bp = input;
   action = NULL_ACTION;
   showstatus = TRUE;
   while (*bp && isspace(*bp)) bp++;
#ifdef DEBUGCMD
   printf ("docmommand: entered: input = '%s'\n", input);
#endif

   switch (*bp++)
   {
   case '\0':
   case '#': /* Comment */
      break;

   case 'a': /* Attach device */
      showstatus = FALSE;
      while (*bp && isspace (*bp)) bp++;
      if (*bp)
	 attachdev (bp);
      if (interactive && view[0][0] == '\0')
      {
	 snprintf (view[0], MAXVIEWLEN, "DEV  SW  ADDR DMA FILE");
	 for (i = 0, j = 1; i < devcnt; i++, j++)
	 {
	    Device *dev = &devices[i];

	    if (dev->infd)
	    {
	       char temp [MAXDEVPATHLEN+2];

	       if ((dev->type == DISK) && (dev->infd != DEV_NULL))
	          sprintf (temp, "%s:%s",
		           disks[dev->info.dskinfo.model].model,
			   dev->file);
	       else
		  strcpy (temp, dev->file);

	       if (j == MAXVIEW)
	       {
	          putview (interactive, TRUE);
		  j = 0;
	       }
	       if (dev->devaddr & DMA_DEVICE)
		  snprintf (view[j], MAXVIEWLEN, "%s   %02X  %04X  %d  %s",
		     dev->name, dev->switches, dev->devaddr,
		     dev->dmaport, temp);
	       else
		  snprintf (view[j], MAXVIEWLEN, "%s   %02X  %04X  -  %s",
		     dev->name, dev->switches, dev->devaddr,
		     temp);
	    }
	 }
      }
      oinput[0] = '\0';
      break;

   case 'b': /* Boot */
      showstatus = FALSE;
      if (*bp == 'n') /* No start option */
      {
	 bp++;
	 norun = TRUE;
      }
      else norun = FALSE;
      if (!*bp)
      {
	 strcpy (view[0], "You must specify a boot device");
      }
      else if (boot (bp) == 0)
      {
	 if (!norun && run)
	 {
	    action = RUN_ACTION;
	    runprog ();
#ifdef DEBUGCMD
	    printf ("docmommand: runprog returns\n");
#endif
	 }
      }
      oinput[0] = '\0';
      break;

   case 'B': /* trace limit Begin pc */
      showstatus = FALSE;
      bp = getnum (bp, &val);
      tracelim[0] = val;
      oinput[0] = '\0';
      break;

   case 'c': /* Clear breakpoint/watchpoint */
      while (*bp && isspace (*bp)) bp++;
      if (*bp == 'b')
	 breakflag = FALSE;
      else if (*bp == 'w')
	 watchflag = FALSE;
      oinput[0] = '\0';
      break;

   case 'd': /* Display memory */
      showstatus = FALSE;
      while (*bp && isspace (*bp)) bp++;
      subcmd = 0;
      if (*bp == 'b')
         subcmd = *bp++;
      bp = getnum (bp, &val);
      if (subcmd == 'b')
         val >>= 1;
      if (val < memlen)
      {
	 dispaddr = val;
	 bp = getnum (bp, &val);
	 if (val > 16)
	 {
	    dispcnt = val;
	    if (dispcnt % 16) dispcnt += 16;
	 }
	 else dispcnt = 16;
	 if (enablepanel && dispcnt/16 > MAXVIEW) dispcnt = MAXVIEW * 16;
	 action = DISP_ACTION;

	 for (i = j = 0; i < (dispcnt/16); i++, j++)
	 {
	    if (j == MAXVIEW)
	    {
	       putview (interactive, TRUE);
	       j = 0;
	    }
	    dumpsingle (view[j]);
	 }
      }
      oinput[0] = '\0';
      break;

   case 'D': /* Disassemble instruction block */
      showstatus = FALSE;
      bp = getnum (bp, &val);
      if (val < memlen)
      {
	 dismaddr = val;
	 bp = getnum (bp, &val);
	 if (val > 0) dismcnt = val;
	 else dismcnt = 1;
	 action = DISP_ACTION;
	 for (i = j = 0; i < dismcnt; i++, j++)
	 {
	    if (j == MAXVIEW)
	    {
	       putview (interactive, TRUE);
	       j = 0;
	    }
	    printsingle (view[j]);
	 }
      }
      oinput[0] = '\0';
      break;

   case 'E': /* trace limit End pc */
      showstatus = FALSE;
      bp = getnum (bp, &val);
      tracelim[1] = val;
      oinput[0] = '\0';
      break;

   case 'g': /* Go - run program */
      if (enablepanel)
         prtpanel (prtviewcol, TRUE);
      action = RUN_ACTION;
      run = TRUE;
      runprog ();
#ifdef DEBUGCMD
      printf ("docmommand: runprog returns\n");
#endif
      break;

   case 'h': /* Help */
   case '?':
      showstatus = FALSE;
      puthelp (bp);
      action = HELP_ACTION;
      oinput[0] = '\0';
      break;

   case 'i': /* dump Instruction trace buffer */
      showstatus = FALSE;
      sprintf (view[0], " PC   ST       INST           CODE");
      for (k = 0, j = 1, i = trcidx; k < TRCBUFMAX; i++, k++)
      {
         if (i >= TRCBUFMAX) i = 0;
	 if (trcbuf[i].pc != -1)
	 {
	    printinst (view[j], i);
	    j++;
	 }
      }
      break;

   case 'k': /* set breaKpoint */
      bp = getnum (bp, &val);
      if (val < memlen)
      {
	 breakaddr = val;
	 breakflag = TRUE;
      }
      oinput[0] = '\0';
      break;

   case 'l': /* Load - use simloader */
      norun = FALSE;
      while (*bp && !isspace (*bp))
      {
	 if (*bp == 'n') /* No start option */
	    norun = TRUE;
	 bp++;
      }
      bp = getnum (bp, &val);
      if (val < memlen)
      {
	 action = val;
	 if (binloader (bp, action) == 0)
	 {
	    if (!norun && run)
	    {
	       action = RUN_ACTION;
	       runprog ();
#ifdef DEBUGCMD
	       printf ("docmommand: runprog returns\n");
#endif
	    }
	    else action = LOAD_ACTION;
	 }
      }
      run = FALSE;
      oinput[0] = '\0';
      break;

   case 'm': /* Modify memory */
      showstatus = FALSE;
      bp = getnum (bp, &val);
      dispaddr = val;
      bp = getnum (bp, &val);
      memory[dispaddr] = val & 0xFFFF;
      action = DISP_ACTION;
      if (!enablepanel)
	 printf (" (%05X) = >%04X\n", dispaddr, memory[dispaddr]);
      oinput[0] = '\0';
      break;

   case 'N': /* Disassemble next instruction block */
      showstatus = FALSE;
      bp = getnum (bp, &val);
      if (val > 0) dismcnt = val;

      for (i = j = 0; i < dismcnt; i++, j++)
      {
	 if (j == MAXVIEW)
	 {
	    putview (interactive, TRUE);
	    j = 0;
	 }
	 printsingle (view[j]);
      }
      break;

   case 'n': /* display Next memory location */
      showstatus = FALSE;
      bp = getnum (bp, &val);
      if (val > 16)
      {
	 dispcnt = val;
	 if (dispcnt % 16) dispcnt += 16;
      }
      else if (val > 0) dispcnt = 16;
      action = DISP_ACTION;

      for (i = j = 0; i < (dispcnt/16); i++, j++)
      {
	 if (j == MAXVIEW)
	 {
	    putview (interactive, TRUE);
	    j = 0;
	 }
	 dumpsingle (view[j]);
      }
      break;

   case 'p': /* set PC */
      bp = getnum (bp, &val);
      SETPC (val & 0xFFFF);
      action = PC_ACTION;
      if (!enablepanel)
	 printf (" R%d = >%04X\n", 7, GETPC);
      oinput[0] = '\0';
      break;

   case 'q': /* Quit */
      action = QUIT_ACTION;
      break;

   case 'R': /* Reset system */
      showstatus = FALSE;
      clearall ();
      action = CLEAR_ACTION;
      oinput[0] = '\0';
      break;

   case 'r': /* Register display/modify */
      while (*bp && isspace (*bp)) bp++;
      subcmd = *bp++;
      bp = getnum (bp, &val1);
      val1 &= 7;
      if (subcmd == 'm')
      {
	 bp = getnum (bp, &val);
	 regs[val1] = val & 0xFFFF;
      }
      if (!enablepanel)
	 printf (" R%ld = >%04X\n", val1, regs[val1]);
      action = DISP_ACTION;
      oinput[0] = '\0';
      break;

   case 's': /* Step program */
      action = STEP_ACTION;
      run = FALSE;
      stepprog ();
      printinst (view[0], trcidx-1);
      break;

   case 't': /* Trace program */
      showstatus = FALSE;
      traceenable = traceenable ? FALSE : TRUE;
      if (traceenable)
      {
	 traceline = 60;
	 bp = getnum (bp, (long *)&tracestart);
	 while (*bp && isspace (*bp)) bp++;
	 if ((tracefd = fopen (bp, "w")) == NULL)
	 {
	    traceenable = FALSE;
	    sprintf (view[0], "Can't open trace file: %s",
		  strerror (ERRNO));
	    sprintf (view[1], "filename: %s", bp);
	 }
      }
      else
      {
	 fclose (tracefd);
	 tracefd = NULL;
      }
      oinput[0] = '\0';
      break;

   case 'w': /* toggle switch */
      bp = getnum (bp, &val);
      if ((val > 0) && (val < 5))
      {
	 val = 1 << (4 - val);
	 ssw ^= val;
      }
      if (!enablepanel)
	 printf ("SSW: %d%d%d%d",
		 (ssw >> 3) & 1, (ssw >> 2) & 1, (ssw >> 1) & 1, ssw & 1);
      break;

   case 'W': /* set Watchpoint */
      bp = getnum (bp, &val);
      if (val < memlen)
      {
	 watchaddr = val;
	 watchflag = TRUE;
      }
      oinput[0] = '\0';
      break;

   case 'z': /* it's Zee panel */
      bp = getnum (bp, &val);
      if (val > 0)
      {
	 windowlen = val;
	 viewlen = windowlen - OUTPUTLINE - 3;
	 prtviewlen = windowlen - OUTPUTLINE - 1;
      }
      if (!enablepanel)
      {
	 enablepanel = TRUE;
	 if (dismcnt > MAXVIEW) dismcnt = MAXVIEW;
	 clearscreen (enablepanel);
	 panel ();
      }
      else
      {
         screenposition (enablepanel, windowlen, 1);
	 enablepanel = FALSE;
      }
      oinput[0] = '\0';
      break;

   default:
      showstatus = FALSE;
      puthelp (NULL);
      oinput[0] = '\0';
   }
   fflush(stdout);
   return (action);
}

/***********************************************************************
* readconfig - Read config file.
***********************************************************************/

static int
readconfig (char *config)
{
   FILE *fd;

   if ((fd = fopen (config, "r")) == NULL)
   {
      sprintf (view[0], "readconfig: open failed: %s",
	    strerror (ERRNO));
      sprintf (view[1], "filename: %s", config);
      return (-1);
   }
   while (fgets (input, sizeof (input), fd))
   {
      char *cp;

      *(strchr (input, '\n')) = '\0';
      if ((cp = strchr (input, '\r')) != NULL)
	 *cp = '\0';
      docommand (FALSE);
      if (view[0][0] != '\0')
      {
	 putview (FALSE, FALSE);
	 break;
      }
   }
   fclose (fd);
   return (0);
}

/***********************************************************************
* main - Main Procedure.
***********************************************************************/

int
main (int argc, char **argv)
{
   char *bp;
   char *pp;
   char *objfile = NULL;
   char *config = NULL;
   int action;
   int i;

   /*
   ** Process args
   */

   for (i = 1; i < argc; i++)
   {
      bp = argv[i];

      if (*bp == '-')
      {
         for (bp++; *bp; bp++) switch (*bp)
         {
	 case 'C': /* delay Clock interval */
	    i++;
	    if (i >= argc) goto USAGE;
	    delayclock = atoi (argv[i]);
	    break;

	 case 'c': /* Config file spec */
	    i++;
	    if (i >= argc) goto USAGE;
	    config = argv[i];
	    break;

	 case 'm': /* CPU Model */
	    i++;
	    if (i >= argc) goto USAGE;
	    if (!strcmp (argv[i], "TI980"))
	       model = CPU_TI980;
	    else if (!strcmp (argv[i], "TI980A"))
	       model = CPU_TI980A;
	    else if (!strcmp (argv[i], "TI980B"))
	       model = CPU_TI980B;
	    else if (!strcmp (argv[i], "CDD980"))
	    {
	       model = CPU_CDD980;
	       memlen = MAXMEMSIZE;
	    }
	    else
	    {
	       fprintf (stderr, "Unsupported CPU model: %s\n", argv[i]);
	       goto USAGE;
	    }
	    break;

	 case 'R': /* Use Real Time clock */
            usertclock = TRUE;
            break;

	 case 's': /* Size of memory */
	    i++;
	    if (i >= argc) goto USAGE;
	    memlen = 0;
	    for (pp = argv[i]; *pp; pp++)
	    {
	       if (isdigit(*pp))
	       {
		  memlen = (memlen * 10) + (*pp - '0');
	       }
	       else
	       {
		  if (*pp == 'k' || *pp == 'K')
		  {
		     memlen *= 1024;
		  }
		  else
		  {
		     fprintf (stderr,
			      "Invalid memory specification %s\n", argv[i]);
		     exit (ABORT);
		  }
	       }
	    }
	    if (((model < CPU_CDD980) && (memlen > MEMSIZE)) ||
	        ((model == CPU_CDD980) && (memlen > MAXMEMSIZE)))
	    {
	       fprintf (stderr,
			"Memory specification, %s, too large for a %s CPU\n",
			argv[i], cpu_models[model]);
	       exit (ABORT);
	    }
	    break;

	 case 'w': /* Set window length */
	    i++;
	    if (i >= argc) goto USAGE;
	    windowlen = atoi (argv[i]);
            /* fall through */

         case 'p': /* Panel enable */
	    enablepanel = TRUE;
            clearscreen (enablepanel);
	    break;

         default:
      USAGE:
	    fprintf (stderr, "usage:  sim980 [-options][object.file]\n");
	    fprintf (stderr, " options:\n");
	    fprintf (stderr, "   -c configfile  - Config file\n");
	    fprintf (stderr, "   -m model       - CPU model\n");
	    fprintf (stderr,
		  "            CPU = TI980, TI980A, TI980B, CDD980\n");
	    fprintf (stderr, "   -p             - Enable panel\n");
	    fprintf (stderr, "   -s memsize     - Memory size\n");
	    fprintf (stderr, "   -w NN          - Window size for panel\n");
	    return (ABORT);
         }
      }
      else
      {
	 if (objfile) goto USAGE;
	 objfile = argv[i];
      }
   }

   /*
   ** Check memory size for CPU type
   */

   displen = memlen;

   /*
   ** Clear system
   */

   clearall ();

   tracelim[0] = 0;
   tracelim[1] = 0xFFFF;

   viewlen = windowlen - OUTPUTLINE - 3;
   prtviewlen = windowlen - OUTPUTLINE - 1;
   prtviewcol = 1;

   /*
   ** Set up initial devices
   */

   for (i = 0; i < MAXDEVICES; i++)
   {
      memset (&devices[i], '\0', sizeof (Device));
   }

   devices[0].type = CONSOLE;
   devices[0].infd = stdin;
   devices[0].outfd = stdout;
   devices[0].devaddr = TERMADDR;
   devices[0].intloc = DBUSINTLOC;
   devices[0].switches = SWUPCASE;
   strcpy (devices[0].name, "CO");
   strcpy (devices[0].file, "TTY");
   devcnt = 1;

   ttinit ();

   /*
   ** Fire up clock thread
   */

   startclk (FALSE);

   /*
   ** Print intro
   */

   if (!enablepanel)
      printf ("%s %s: %s CPU %dK\n",
	       INTRO, VERSION, cpu_models[model], displen/1024);

   signal (SIGINT, sigintr);

   /*
   ** Config file specified, do the commands.
   */

   if (config)
   {
      readconfig (config);
   }
   showstatus = TRUE;

   /*
   ** If object file specified, load and run it.
   */

   if (objfile)
   {
      if (binloader (objfile, 0) == 0)
      {
	 if (run)
	 {
	    runprog ();
	 }
      }
      else
      {
         putview (FALSE, FALSE);
      }
   }

   /*
   ** Command loop
   */

   input[0] = oinput[0] = '\0';
   action = NULL_ACTION;

   while (!objfile && action != QUIT_ACTION)
   {
      panel ();

      if (view[0][0] != '\0') putview (TRUE, FALSE);

      if (enablepanel)
      {
	 screenposition (enablepanel, windowlen-2, 1);
	 clearline (enablepanel);
	 printf (":%s", oinput);
	 screenposition (enablepanel, windowlen-2, 2);
      }
      else
      {
	 if (oinput[0])
	    printf ("\n:(%s) ", oinput);
	 else
	    printf ("\n: ");
      }

      fflush(stdout);
      action = NULL_ACTION;
      if (fgets (input, sizeof (input), stdin) != NULL)
      {
	 *(strchr (input,'\n')) = '\0';

	 if (input[0] == '\0')
	    strcpy (input, oinput);
	 else
	    strcpy (oinput, input);
	 putview (TRUE, FALSE);

	 action = docommand (TRUE);
      }
      else
      {
         printf ("fget returned NULL\n");
      }

   }

   /*
   ** Close open devices
   */

   for (i = 0; i < devcnt; i++)
   {
      if (devices[i].infd != NULL && devices[i].infd != DEV_NULL)
      {
	 if (devices[i].infd != stdin)
	    fclose (devices[i].infd);
      }
   }

   ttclose ();
   stopclk (TRUE);
   screenposition (enablepanel, windowlen, 1);
   return (NORMAL);
}
