/***********************************************************************
*
* simdef.h - Simulator header for the TI 980 computer.
*
* Changes:
*   06/23/11   DGP   Original.
*   08/03/11   DGP   Added 980 card format object support.
*   08/26/11   DGP   Changed LOADADDRESS to 0x100.
*   08/31/11   DGP   Added Status register bit names.
*   09/14/11   DGP   Added cpu models and Calder 980 support.
*   10/25/11   DGP   Added CDD Common Data Feature.
*
***********************************************************************/

#include <stdio.h>

/*
** Definitions
*/

#define VERSION "2.0.0"
#define INTRO "TI 980 Simulator"

#define NORMAL 0
#ifndef ABORT
#define ABORT  12
#endif


#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif


/*
** Configuration definitions
*/

#define MAXMEMSIZE 256*1024	/* Maximum size of physical memory */
#define MEMSIZE 64*1024		/* Size of physical memory */

#define MAXDEVICES 20		/* Number of devices permitted */
#define MAXVIEW 100		/* Number of messages in view panel */
#define MAXVIEWLEN 80		/* Length of messages in view panel */
#define MAXDEVPATHLEN 256

#define TOPLINE       1
#define STATUSLINE    3
#define BREAKLINE     4
#define REGISTERLINE  6
#define OUTPUTLINE    11

#define TRCBUFMAX     50

/*
** System definitions.
*/

#define LOADADDRESS 0x100

#define EOFSYM ':'
#define HEXSYM '>'

/*
** Default device addresses and interrupt levels.
*/

#define PIFLOWADD  0x00		/* PIF lower address boundry */
#define PIFUPPADD  0x01		/* PIF upper address boundry */
#define CLOCKADDR  0x04		/* Clock address */
#define TERMADDR   0x05		/* Serial Terminal address */
#define INTEXADDR  0x13		/* Interrupt expander */

/* CDD980 devices */

#define CDDMEMMAP  0x03		/* CDD980 Memory Mapper */
#define CDDCDBLIM  0x06		/* CDD980 CDB Limit */
#define CDDBNKSEL  0x07		/* CDD980 Bank Select */
#define CDDDMMAP0  0x18		/* CDD980 DMA MAP 0 */
#define CDDDMMAP1  0x19		/* CDD980 DMA MAP 1 */
#define CDDDMMAP2  0x1A		/* CDD980 DMA MAP 2 */
#define CDDDMMAP3  0x1B		/* CDD980 DMA MAP 3 */
#define CDDDMMAP4  0x1C		/* CDD980 DMA MAP 4 */
#define CDDDMMAP5  0x1D		/* CDD980 DMA MAP 5 */
#define CDDDMMAP6  0x1E		/* CDD980 DMA MAP 6 */
#define CDDDMMAP7  0x1F		/* CDD980 DMA MAP 7 */

#define PWRINTLOC  0x00		/* Power interrupt location */
#define INTINTLOC  0x02		/* Internal interrupt location */
#define DMAINTLOC  0x04		/* DMA device interrupt location */
#define DBUSINTLOC 0x06		/* Data Bus device interrupt location */

#define DMASTATLOC 0x96		/* DMA interrupt status */

/*
** CPU Model
*/

enum cpumodels
{
   CPU_TI980, CPU_TI980A, CPU_TI980B, CPU_CDD980
};

/*
** Data type definitions
*/

#define int8            signed char
#define int16           short
#define int32           int
typedef int             t_stat;                         /* status */
typedef int             t_bool;                         /* boolean */
typedef unsigned char   uint8;
typedef unsigned short  uint16;
typedef unsigned int    uint32, t_addr;                 /* address */

#if defined (WIN32)                                     /* Windows */
#define t_int64 __int64
#elif defined (__ALPHA) && defined (VMS)                /* Alpha VMS */
#define t_int64 __int64
#elif defined (__ALPHA) && defined (__unix__)           /* Alpha UNIX */
#define t_int64 long
#else                                                   /* default GCC */
#define t_int64 long long
#endif                                                  /* end OS's */
typedef unsigned t_int64        t_uint64, t_value;      /* value */
typedef t_int64                 t_svalue;               /* signed value */

/*
** Access macros and functions.
*/

extern uint16 getinst (int);
extern uint16 getreg (uint16);
extern void putreg (uint16, uint16);
extern uint16 getmem (uint16);
extern void putmem (uint16, uint16);

#define INCPC(v) regs[7]+=(v)
#define DECPC(v) regs[7]-=(v)
#define SETPC(v) regs[7]=(v)
#define GETPC regs[7]

#define R0  regs[0]
#define R1  regs[1]
#define R2  regs[2]
#define R3  regs[3]
#define R4  regs[4]
#define R5  regs[5]
#define R6  regs[6]
#define R7  regs[7]

#define RF0 regs[8]
#define RF1 regs[9]
#define RF2 regs[10]
#define RF3 regs[11]
#define RF4 regs[12]
#define RF5 regs[13]
#define RF6 regs[14]
#define RF7 regs[15]

/*
** Status register set and check
*/

#define GT_BIT     0x8000
#define EQ_BIT     0x4000
#define OVER_BIT   0x2000
#define CARRY_BIT  0x1000
#define PRIV_BIT   0x0800
#define PROT_BIT   0x0400
#define PIFVIO_BIT 0x0200
#define DBINT_BIT  0x0100
#define VECINT_BIT 0x0080
#define PIF_BIT    0x0040
#define PRENDX_BIT 0x0020
#define PARINT_BIT 0x0010
#define DMAINT_BIT 0x0008
#define MEMMAP_BIT 0x0004
#define PARERR_BIT 0x0002
#define PWRDWN_BIT 0x0001

#define CMP_BITS   (GT_BIT | EQ_BIT)
#define CLR_CMP    (statreg &= ~CMP_BITS)
#define CLR_OVER   (statreg &= ~OVER_BIT)
#define CLR_CARRY  (statreg &= ~CARRY_BIT)
#define CLR_PRIV   (statreg &= ~PRIV_BIT)
#define CLR_PROT   (statreg &= ~PROT_BIT)
#define CLR_PIFVIO (statreg &= ~PIFVIO_BIT)
#define CLR_DBINT  (statreg &= ~DBINT_BIT)
#define CLR_VECINT (statreg &= ~VECINT_BIT)
#define CLR_PIF    (statreg &= ~PIF_BIT)
#define CLR_PRENDX (statreg &= ~PRENDX_BIT)
#define CLR_PARINT (statreg &= ~PARINT_BIT)
#define CLR_DMAINT (statreg &= ~DMAINT_BIT)
#define CLR_MEMMAP (statreg &= ~MEMMAP_BIT)
#define CLR_PARERR (statreg &= ~PARERR_BIT)
#define CLR_PWRDWN (statreg &= ~PWRDWN_BIT)

#define SET_LT     (statreg = (statreg & ~CMP_BITS))
#define SET_EQ     (statreg = (statreg & ~CMP_BITS) | EQ_BIT)
#define SET_GT     (statreg = (statreg & ~CMP_BITS) | GT_BIT)
#define SET_OVER   (statreg |= OVER_BIT)
#define SET_CARRY  (statreg |= CARRY_BIT)
#define SET_PRIV   (statreg |= PRIV_BIT)
#define SET_PROT   (statreg |= PROT_BIT)
#define SET_PIFVIO (statreg |= PIFVIO_BIT)
#define SET_DBINT  (statreg |= DBINT_BIT)
#define SET_VECINT (statreg |= VECINT_BIT)
#define SET_PIF    (statreg |= PIF_BIT)
#define SET_PRENDX (statreg |= PRENDX_BIT)
#define SET_PARINT (statreg |= PARINT_BIT)
#define SET_DMAINT (statreg |= DMAINT_BIT)
#define SET_MEMMAP (statreg |= MEMMAP_BIT)
#define SET_PARERR (statreg |= PARERR_BIT)
#define SET_PWRDWN (statreg |= PWRDWN_BIT)

#define IS_LT      ((statreg & CMP_BITS) == 0x0000)
#define IS_GT      ((statreg & CMP_BITS) == GT_BIT)
#define IS_EQ      ((statreg & CMP_BITS) == EQ_BIT)
#define IS_OVER    (statreg & OVER_BIT)
#define IS_CARRY   (statreg & CARRY_BIT)
#define IS_PRIV    (statreg & PRIV_BIT)
#define IS_PROT    (statreg & PROT_BIT)
#define IS_PIFVIO  (statreg & PIFVIO_BIT)
#define IS_DBINT   (statreg & DBINT_BIT)
#define IS_VECINT  (statreg & VECINT_BIT)
#define IS_PIF     (statreg & PIF_BIT)
#define IS_PRENDX  (statreg & PRENDX_BIT)
#define IS_PARINT  (statreg & PARINT_BIT)
#define IS_DMAINT  (statreg & DMAINT_BIT)
#define IS_MEMMAP  (statreg & MEMMAP_BIT)
#define IS_PARERR  (statreg & PARERR_BIT)
#define IS_PWRDWN  (statreg & PWRDWN_BIT)

#define IS_INTENAB ((statreg & (DBINT_BIT | VECINT_BIT | DMAINT_BIT)) != 0)

/*
** Instruction types
*/

#define MAX_INST_TYPES 11
enum optypes
{
   TYPE_1=0, TYPE_2,  TYPE_3,  TYPE_4,  TYPE_5,  TYPE_6,  TYPE_7,
   TYPE_8,   TYPE_9,  TYPE_10, TYPE_11, TYPE_ILL
};

/*
** Panel actions
*/

enum actions
{
   NULL_ACTION=0, QUIT_ACTION,   STEP_ACTION,   RUN_ACTION,    HELP_ACTION,
   LOAD_ACTION,   PC_ACTION,     DISP_ACTION,   CLEAR_ACTION
};

#define IDLEINST 0xCE00
#define SSBINST  0xD8C0
#define IS_IDLE(x) (((x) & 0xFFF0) == IDLEINST)

/*
** Object tags
*/

#define BINIDT_TAG	01
#define IDT_TAG		'0'
#define ABSENTRY_TAG	'1'
#define RELENTRY_TAG	'2'
#define RELEXTRN_TAG	'3'
#define ABSEXTRN_TAG	'4'
#define RELGLOBAL_TAG	'5'
#define ABSGLOBAL_TAG	'6'
#define CKSUM_TAG	'7'
#define NOCKSUM_TAG	'8'
#define ABSORG_TAG	'9'
#define RELORG_TAG	'A'
#define ABSDATA_TAG	'B'
#define RELDATA_TAG	'C'
#define EXTVALUE_TAG	'D'
#define RELBYTE_TAG	'E'
#define EOR_TAG		'F'

#define CHARSPERREC	66  /* Chars per object record */
#define WORDTAGLEN	5   /* Word + Object Tag length */
#define WORDLEN		4   /* Word length */
#define BINWORDTAGLEN	3   /* Binary Word + Object Tag length */
#define SYMLEN		6   /* REF/DEF Symbol length */
#define IDTLEN		6   /* IDT length */

/*
** Device table entry.
*/

/* Character definitions */

#define BACKSPACE  0x08
#define BLANK      0x20
#define ESCAPE     0x1B
#define TILDE      0x7E
#define DBACKSPACE 0x7F

#define DEV_NULL ((FILE *)1)
#define DEV_VDT ((FILE *)2)

/* Device switch masks. */

#define SWUPCASE  0x0001
#define SWPROTECT 0x0002
#define SWCREATE  0x0004
#define SWLOGOUT  0x0008
#define SWSCSI    0x0010

/* Data bus definitions RDS/WDS */

#define DBUS_BUSY 0x0080
#define DBUS_INC  0x0020
#define DBUS_ADDR 0x0008

/* DMA device definitions */

#define DMA_DEVICE 0x0100

/* DMA common status word bits */

#define OPCOMP	0x8000		/* Complete */
#define OFFLIN	0x2000		/* Off Line */

typedef struct 
{
   uint16 subcmd;
   uint16 genint;
   uint16 chain;
   uint16 cyl;
   uint16 sector;
   uint16 head;
   uint16 seccnt;
} DiskParams;

/* Devices supported by the system */

enum dev_t
{
   CONSOLE=1,  CARDREADER, CARDPUNCH,  PRINTER,
   SERIAL,     DISK,       TAPE,
};

typedef struct
{
   int type;
   int model;
   int cyls;
   int heads;
   int secttrk;
   int sectlen;
   int overhead;
   int tracklen;
   int curcyl;
   int curhead;
   int cursect;
} DiskInfo;

typedef struct
{
   long curpos;
   int online;
} TapeInfo;

typedef struct
{
   int cardeof;
   int cardndx;
   char carddat[82];
} CARDInfo;

typedef struct
{
   FILE *infd;
   FILE *outfd;
   int type;
   int unit;
   int select;
   int count;
   int inchar;
   int outchar;
   int cbreak;
   int viewndx;
   int switches;
   uint16 devaddr;
   uint16 intloc;
   uint16 dmaport;
   char file[MAXDEVPATHLEN+2];
   char name[80];
   union
   {
      DiskInfo dskinfo;
      TapeInfo tapeinfo;
      CARDInfo cardinfo;
   } info;
} Device;

/*
** Trace Instruction 
*/

typedef struct
{
   int  pc;
   int  st;
   int  inst;
   int  ops[3];
   int  regs[16];
   char dis[64];
} TraceInst;

/*
** Externals
*/

/* sim980 externals */

extern char *getnum (char *, long *);
extern void panel (void);
extern void dumpsingle (char *);
extern void printinst (char *, int);

/* simloader externals */

extern int loadrec (FILE *, uint8 *, int, int *);
extern int binloader (char *, int);

/* simops externals */

extern void stepprog (void);
extern void runprog (void);
extern void cmpwordzero (uint16);
extern int decodeinst (uint16);

/* simio externals */

void dmaputword (Device *, uint16, uint16);
uint16 dmagetword (Device *, uint16);
void dmaputbyte (Device *, uint16, uint16, uint8);
uint8 dmagetbyte (Device *, uint16, uint16);

extern int doRDS (uint16, uint16);
extern int doWDS (uint16, uint16);
extern int doATI (uint16, uint16);

extern void attachdev (char *);
extern void resetdev (void);
extern int boot (char *);

extern void checkinterrupts (void);
extern int geninterrupt (uint16, uint16);

/* simterm externals */

extern void prtpanel (int, int);
extern void putview (int, int);

extern int serialpollchar (Device *, long);

extern int serialgetchar (Device *);
extern void serialputchar (Device *, uint16);

extern int ttinit (void);
extern int ttcmdstate (void);
extern int ttrunstate (void);
extern int ttclose (void);

/* simclock externals */

extern void startclk (int);
extern void stopclk (int);
extern void smallsleep (long);

/* simdsk externals */

extern int dskopen (Device *);
extern int dskdocmd (Device *, uint16, uint16);
extern int dskboot (Device *);
extern void dskreset (Device *);

/* simsdsk externals */

extern int sdskopen (Device *);
extern int sdskdocmd (Device *, uint16, uint16);
extern int sdskboot (Device *);
extern void sdskreset (Device *);

/* simtape externals */

extern int tapeopen (Device *);
extern int tapedocmd (Device *, uint16, uint16);
extern int tapeboot (Device *);
extern void tapereset (Device *);

/* simdis externals */

extern void disinst (TraceInst *);

/*
** HEXDUMP - Macro used in debugging
*/

extern void hexdump (FILE *, char *, int, int);
#define HEXDUMP(file,ptr,size,offset) hexdump(file,ptr,offset,size)

/*
** Define things specific to the system environment
*/

#if defined(VAXVMS)
#if defined(TGVTCP)
#define ERRNO		socket_errno
#define CLOSE		socket_close
#define READ		socket_read
#define WRITE		socket_write
#define PERROR		socket_perror
#endif /* TGVTCP */
#if defined(UCXTCP)
/*
** Select uses bit masks of file descriptors in longs.
** These macros manipulate such bit fields (the filesystem macros use
** chars). FD_SETSIZE may be defined by the user, but the default here
** should be >= NOFILE (param.h).
*/
#ifndef FD_SETSIZE
#define FD_SETSIZE      256
#endif
#ifndef NBBY
#define NBBY    8
#endif

#define NFDBITS (sizeof(fd_mask) * NBBY)	/* bits per mask */
#ifndef howmany
#define howmany(x, y)   (((x)+((y)-1))/(y))
#endif

#if defined(VMSV5) || defined(VMSV6)
typedef long    fd_mask;

typedef struct fd_set
{
   fd_mask         fds_bits[howmany (FD_SETSIZE, NFDBITS)];
}
fd_set;
#endif /* VMSV5 || VMSV6 */

#define  FD_SET(n, p)   ((p)->fds_bits[((unsigned)n)/NFDBITS] \
                        |= (1 << (((unsigned)n) % NFDBITS)))
#define  FD_CLR(n, p)   ((p)->fds_bits[((unsigned)n)/NFDBITS] \
                        &= ~(1 << (((unsigned)n) % NFDBITS)))
#define  FD_ISSET(n, p) ((p)->fds_bits[((unsigned)n)/NFDBITS] \
                        & (1 << (((unsigned)n) % NFDBITS)))
#define FD_ZERO(p) memset((char *)(p), '\0', sizeof(*(p)))

#define APPTCPBLOCKMODE	0
#endif /* UCXTCP */
#endif /* VAXVMS */

#if defined(WIN32) 
#include <winsock.h>

#define CLOSE		closesocket
#define IOCTL(a,b,c,d)	ioctlsocket((a),(b),(unsigned long *)(c))
#define ERRNO		WSAGetLastError()
#define PERROR(m)	fprintf (stderr, "%s: errno = %d\n", (m), ERRNO);
typedef unsigned long	caddr_t;
#ifndef ETIMEDOUT
#define ETIMEDOUT	WSAETIMEDOUT
#endif
#ifndef EADDRINUSE
#define EADDRINUSE	WSAEADDRINUSE
#endif
#ifndef ECONNREFUSED
#define ECONNREFUSED	WSAECONNREFUSED
#endif
#ifndef EINPROGRESS
#define EINPROGRESS	WSAEWOULDBLOCK
#endif
#ifndef ENOTCONN
#define ENOTCONN	WSAENOTCONN
#endif
#ifndef ECONNREFUSED
#define ECONNREFUSED	WSAECONNREFUSED
#endif
#ifndef ENOBUFS
#define ENOBUFS	        WSAENOBUFS
#endif
#endif /* WIN32  */

#if defined(HPUXV9)
#define SELECT(a,b,c,d,e) select((a),(int *)(b),(int *)(c),(int *)(d),(e))
#endif /* HPUX */

#if defined(RISCOS)
extern int errno;
#endif /* RISCOS */

/*
** Define things left undefined above
*/

#ifndef ERRNO
#define ERRNO		errno
#endif
#ifndef CLOSE
#define CLOSE		close
#endif
#ifndef READ
#define READ		read
#endif
#ifndef WRITE
#define WRITE		write
#endif
#ifndef PERROR
#define PERROR		perror
#endif
#ifndef SELECT
#define SELECT		select
#endif
#ifndef IOCTL
#define IOCTL		ioctl
#endif

