/***********************************************************************
*
* simdis.c - Disassemble instructions
*
* Changes:
*   06/28/11   DGP   Original.
*
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>

#include "simdef.h"

static int opidx;

/*
** Opcode tables.
*/

static char *ops1op[] =
{
   "LDA",  "LDE",  "LDX",  "LDM",  "ADD",  "SUB",  "IOR",  "AND",
   "BIX",  "DMT",  "IMO",  "DIV",  "CPL",  "CPA",  "BRL",  "BRU",
   "STA",  "STE",  "STX",  "MPY",  "DST",  "DSB",  "DLD",  "DAD"
};

static char *ops2op[] =
{
   "RSU",  "RAD",  "RCO",  "",     "RIV",  "REO",  "RIN",  "",
   "RCA",  "ROR",  "RMO",  "",     "RCL",  "RAN",  "RDE",  "REX"
};

static char *ops3op[] =
{
   "ARA",  "ARD",  "LRA",  "LRD",  "ALA",  "ALD",  "LLA",  "LLD",
   "RTO",  "",     "RTZ",  "",     "LTO",  "",     "LTZ",  "",
   "CRA",  "CRE",  "CRX",  "CRM",  "NRM",  "",     "",     "", 
   "",     "CRS",  "CRL",  "CRB",  "CLD",  "",     "CRD",
};

static char *ops4op[] =
{
   "SZE",  "SSE",  "SOO",  "",     "SOD",  "",     "SMI",  "",
   "SNZ",  "SSN",  "SNO",  "",     "SEV",  "",     "SPL",  "",
   "SLT",  "",     "SEQ",  "",     "SGT",  "",     "SOV",  "",
   "SGE",  "",     "SNE",  "",     "SLE",  "",     "SNV",  "",
   "IDL",  "SOC",  "SNC",
};

static char *ops5op[] =
{
   "RDS",  "",     "WDS",  "",     "",     "",     "",     "",
   "LSB",  "LSR",  "LRF",  "",     "SSB",  "",     "SRF",
};

static char *ops6op[] =
{
   "TABZ", "TABO", "TMBZ", "TMBO", "SABZ", "SABO", "SMBZ", "SMBO",
};

static char *ops7op[] =
{
   "ATI",  "API"
};

static char *ops8op[] =
{
   "MVC",  "CLC"
};

static char *ops10op[] =
{
   "LAR", "SAR", "RFI", "RFD"
};

static char *ops11op[] =
{
   "SWTX", "SWT", "SWTS"
};

static char *movmodes[] =
{
   "*", "*+", "-*", "", "#", "#+", "-#", ""
};

/*
** Function definitions and branch table
*/

static int dis1op  (TraceInst *trcdata);
static int dis2op  (TraceInst *trcdata);
static int dis3op  (TraceInst *trcdata);
static int dis4op  (TraceInst *trcdata);
static int dis5op  (TraceInst *trcdata);
static int dis6op  (TraceInst *trcdata);
static int dis7op  (TraceInst *trcdata);
static int dis8op  (TraceInst *trcdata);
static int dis9op  (TraceInst *trcdata);
static int dis10op (TraceInst *trcdata);
static int dis11op (TraceInst *trcdata);

typedef struct {
   int (*proc) (TraceInst *trcdata);
} Dis_Proc;

static Dis_Proc dis_proc[MAX_INST_TYPES] = {
   dis1op,  dis2op,  dis3op,  dis4op,  dis5op,  dis6op,  dis7op,
   dis8op,  dis9op,  dis10op, dis11op
};

/***********************************************************************
* dis1op - Process type 1 operations.
***********************************************************************/

static int
dis1op (TraceInst *trcdata)
{
   int extended;
   uint16 inst;
   uint16 ixb;
   uint16 disp;
   uint16 op;
   char temp[32];

   inst = trcdata->inst;
   ixb = (inst & 0x0700) >> 8;
   disp = (inst & 0x00FF);
   op = (inst >> 11);

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", op, op);
#endif

   extended = FALSE;
   if ((disp == 0) && ((ixb == 0) || (ixb == 4) || (ixb == 6)))
   {
      extended = TRUE;
   }

   switch (op)
   {
   case 0: /* LDA */
   case 1: /* LDE */
   case 2: /* LDX */
   case 3: /* LDM */
   case 4: /* ADD */
   case 5: /* SUB */
   case 6: /* IOR */
   case 7: /* AND */
   case 8: /* BIX */
   case 9: /* DMT */
   case 10: /* IMO */
   case 11: /* DIV */
   case 12: /* CPL */
   case 13: /* CPA */
   case 14: /* BRL */
   case 15: /* BRU */
   case 16: /* STA */
   case 17: /* STE */
   case 18: /* STX */
   case 19: /* MPY */
   case 20: /* DST */
   case 21: /* DSB */
   case 22: /* DLD */
   case 23: /* DAD */
      break;

   default:
      return (-1);
   }

   if (extended)
      sprintf (temp, "@%s", ops1op[op]);
   else
      strcpy (temp, ops1op[op]);

   sprintf (trcdata->dis, "%-8s >%02X,%d", temp, disp, ixb);
   if (extended)
   {
      sprintf (temp, "  >%04X", trcdata->ops[0] & 0xFFFF);
      strcat (trcdata->dis, temp);
   }
      
   return (0);
}

/***********************************************************************
* dis2op - Process type 2 operations.
***********************************************************************/

static int
dis2op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 op;
   uint16 sr, dr;

   inst = trcdata->inst;
   dr = inst & 0x0007;
   sr = (inst >> 4) & 0x0007;
   op = ((inst & 0x0F80) >> 7);

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", op, op);
#endif

   switch (op)
   {
   case 0: /* RSU */
   case 1: /* RAD */
   case 2: /* RCO */
   case 4: /* RIV */
   case 5: /* REO */
   case 6: /* RIN */
   case 8: /* RCA */
   case 9: /* ROR */
   case 10: /* RMO */
   case 12: /* RCL */
   case 13: /* RAN */
   case 14: /* RDE */
   case 15: /* REX */
      break;

   default:
      return (-1);
   }

   sprintf (trcdata->dis, "%-8s %d,%d", ops2op[op], sr, dr);
   return (0);
}

/***********************************************************************
* dis3op - Process type 3 operations.
***********************************************************************/

static int
dis3op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 count;
   uint16 op;

   inst = trcdata->inst;
   count = inst & 0x001F;
   op = ((inst & 0x03F0) >> 5);

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", op, op);
#endif

   switch (op)
   {
   case 0: /* ARA */
   case 1: /* ARD */
   case 2: /* LRA */
   case 3: /* LRD */
   case 4: /* ALA */
   case 5: /* ALD */
   case 6: /* LLA */
   case 7: /* LLD */
   case 8: /* RTO */
   case 10: /* RTZ */
   case 12: /* LTO */
   case 14: /* LTZ */
   case 16: /* CRA */
   case 17: /* CRE */
   case 18: /* CRX */
   case 19: /* CRM */
   case 25: /* CRS */
   case 26: /* CRL */
   case 27: /* CRB */
   case 28: /* CLD */
   case 30: /* CRD */
      break;

   case 20: /* NRM */
      sprintf (trcdata->dis, "%-8s", ops3op[op]);
      return (0);

   default:
      return (-1);
   }

   sprintf (trcdata->dis, "%-8s %d", ops3op[op], count);
   return (0);
}

/***********************************************************************
* dis4op - Process type 4 operations.
***********************************************************************/

static int
dis4op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 op;

   inst = trcdata->inst;
   op =  ((inst & 0x3F0) >> 4);

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", op, op);
#endif

   switch (op)
   {
   case 0: /* SZE */
   case 1: /* SSE */
   case 2: /* SOO */
   case 4: /* SOD */
   case 6: /* SMI */
   case 8: /* SNZ */
   case 9: /* SSN */
   case 10: /* SNO */
   case 12: /* SEV */
   case 14: /* SPL */
   case 32: /* IDL */
      sprintf (trcdata->dis, "%-8s %d", ops4op[op], inst & 0x000F);
      return (0);
      break;

   case 16: /* SLT */
   case 18: /* SEQ */
   case 20: /* SGT */
   case 22: /* SOV */
   case 24: /* SGE */
   case 26: /* SNE */
   case 28: /* SLE */
   case 30: /* SNV */
      break;

   case 54: /* SOC */
      op = 33;
      break;

   case 62: /* SNC */
      op = 34;
      break;

   default:
      return (-1);
   }

   sprintf (trcdata->dis, "%-8s ", ops4op[op]);
   return (0);
}

/***********************************************************************
* dis5op - Process type 5 operations.
***********************************************************************/

static int
dis5op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 dev;
   uint16 op;

   inst = trcdata->inst;
   dev = (inst & 0x065F);
   op = ((inst & 0x00F0) >> 4);

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", op, op);
#endif

   if ((inst & 0xD9A0) == 0xD800) /* RDS */
      op = 0;
   else if ((inst & 0xD9A0) == 0xD820) /* WDS */
      op = 2;
   else switch (op)
   {
   case 8: /* LSB */
   case 9: /* LSR */
   case 10: /* LRF */
   case 12: /* SSB */
   case 14: /* SRF */
      break;

   default:
      return (-1);
   }

   if (op < 4)
      sprintf (trcdata->dis, "%-8s >%04X  >%04X",
	       ops5op[op], dev, trcdata->ops[0] & 0xFFFF);
   else
      sprintf (trcdata->dis, "@%-7s >%04X",
	       ops5op[op], trcdata->ops[0] & 0xFFFF);
   return (0);
}

/***********************************************************************
* dis6op - Process type 6 operations.
***********************************************************************/

static int
dis6op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 op;
   uint16 bit;

   inst = trcdata->inst;
   bit = inst & 0x000F;
   op = ((inst & 0x00F0) >> 4);

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", op, op);
#endif

   switch (op)
   {
   case 0: /* TABZ */
   case 1: /* TABO */
   case 4: /* SABZ */
   case 5: /* SABO */
      sprintf (trcdata->dis, "%-8s >%X", ops6op[op], bit);
      break;

   case 2: /* TMBZ */
   case 3: /* TMBO */
   case 6: /* SMBZ */
   case 7: /* SMBO */
      sprintf (trcdata->dis, "%-8s >%X  >%04X",
	       ops6op[op], bit, trcdata->ops[0] & 0xFFFF);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* dis7op - Process type 7 operations.
***********************************************************************/

static int
dis7op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 op = 0;
   uint16 dev;

   inst = trcdata->inst;
   dev = inst & 0x00FF;

   if ((inst & 0xFF00) == 0xD900) /* ATI */
      op = 0;
   else if ((inst & 0xFF00) == 0xDD00) /* API */
      op = 1;
   else return (-1);

   sprintf (trcdata->dis, "%-8s >%02X  >%04X",
            ops7op[op], dev, trcdata->ops[0] & 0xFFFF);
   return (0);
}

/***********************************************************************
* dis8op - Process type 8 operations.
***********************************************************************/

static int
dis8op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 op = 0;

   inst = trcdata->inst;

   if ((inst & 0xFF80) == 0xDF00) /* MVC */
      op = 0;
   else if ((inst & 0xFF80) == 0xDF80) /* CLC */
      op = 1;
   else return (-1);

   sprintf (trcdata->dis, "%-8s ", ops8op[op]);
   return (0);
}

/***********************************************************************
* dis9op - Process type 9 operations.
***********************************************************************/

static int
dis9op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 sr, dr;
   uint16 smode, dmode;

   inst = trcdata->inst;

   sr = ((inst >> 4) & 0xF);
   smode = ((inst >> 10) & 0x3) | ((inst >> 4) & 0x4);
   dr = (inst & 0xF);
   dmode = ((inst >> 8) & 0x3) | (inst & 0x4);

   sprintf (trcdata->dis, "%-8s %s%d,%s%d",
   	    "MOV", movmodes[smode], sr, movmodes[dmode], dr);
   return (0);
}

/***********************************************************************
* dis10op - Process type 10 operations.
***********************************************************************/

static int
dis10op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 sr, dr;

   inst = trcdata->inst;

   sr = ((inst >> 4) & 7);
   dr = (inst & 7);

   if ((inst & 0x0088) == 0x0088) /* LAR */
   {
      sprintf (trcdata->dis, "@%-7s >%04X",
	       ops10op[0], trcdata->ops[0] & 0xFFFF);
   }
   else if ((inst & 0x0088) == 0x0080) /* SAR */
   {
      sprintf (trcdata->dis, "@%-7s >%04X",
	       ops10op[1], trcdata->ops[0] & 0xFFFF);
   }
   else if ((inst & 0x0088) == 0x0008) /* RFI */
   {
      sprintf (trcdata->dis, "%-8s %d,%d", ops10op[2], sr, dr);
   }
   else if ((inst & 0x0088) == 0x0000) /* RFD */
   {
      sprintf (trcdata->dis, "%-8s %d,%d", ops10op[3], sr, dr);
   }
   else
      return (-1);
   
   return (0);
}

/***********************************************************************
* dis11op - Process type 11 operations.
***********************************************************************/

static int
dis11op (TraceInst *trcdata)
{
   uint16 inst;
   uint16 op = 0;

   inst = trcdata->inst;
   op = (inst & 0x0F00) >> 8;

   switch (op)
   {
   case 0:
   case 1:
   case 2:
      sprintf (trcdata->dis, "%-8s >%02X", ops11op[op], inst & 0xFF);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* disinst - Disassemble an instruction.
***********************************************************************/

void 
disinst (TraceInst *trcdata)
{
   int type;
   int error;

   type = decodeinst (trcdata->inst);
   trcdata->dis[0] = '\0';

   opidx = 0;
   error = -1;
#ifdef DEBUGINSTR
   printf ("PC = %04X INST = %04X TYPE = %d",
   	   trcdata->pc, trcdata->inst, type + 1);
#endif
   if (type != TYPE_ILL)
   {
      error = dis_proc[type].proc (trcdata);
   }

#ifdef DEBUGINSTR
   printf ("\n\r");
#endif
   if (error < 0)
      strcpy (trcdata->dis, "*illop*");

}
