/***********************************************************************
*
* simdsk.c - Disk IO processing for the TI 980 Simulator.
*
* Changes:
*   06/24/11   DGP   Original.
*   08/31/11   DGP   Fix up interrupts.
*   09/02/11   DGP   Added disk geometry checks and type.
*   08/23/16   DGP   Added SCSI disk support.
*
* The disk is formatted as follows:
*    4 bytes (big-endian): # cylinders
*    4 bytes (big-endian): disk type << 16 | # heads
*    4 bytes (big-endian): overhead << 8 | # sectors/track
*    4 bytes (big-endian): # bytes/sector
*    N bytes: raw data for each sector - the order is:
*      cylinder 0 head 0 sector 0
*      cylinder 0 head 0 sector 1, ...
*      cylinder 0 head 0 sector N,
*      cylinder 0 head 1 sector 0, ... 
*      cylinder 0 head M sector N,
*      cylinder 1 head 0 sector 0, ...
*
***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <errno.h>

#if defined (WIN32)
#define __TTYROUTINES 0
#include <conio.h>
#include <windows.h>
#include <signal.h>
#endif

#if defined(UNIX)
#include <sys/time.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <pthread.h>
#endif

#include "simdef.h"

#define EXTERN extern
#include "simdsk.h"

extern uint16 statreg;	/* The program status register */

extern int run;
extern int devcnt;
extern int model;
extern uint32 memlen;
extern unsigned long instcount;
extern char view[MAXVIEW][MAXVIEWLEN+1];

extern uint16 regs[16];
extern uint16 memory[MAXMEMSIZE];
extern Device devices[MAXDEVICES];

/*
** Disk status word bits
*/

#define WPERR	0x4000		/* Write Protect Error */
#define XFRERR  0x1000		/* Data transfer error */
#define EODERR  0x0800		/* End of Disk error */
#define PGMERR	0x0400		/* Program error */
#define CMPERR  0x0200		/* Compare error */
#define PARERR  0x0100		/* Parity error */
#define IDERR   0x0080		/* ID error */
#define DEVBSY  0x0040		/* Device Busy */
#define CHNTKN  0x0020		/* Chain taken */
#define WPBIT	0x0010		/* Write Protect */

/***********************************************************************
* dskreadint - Read an integer.
***********************************************************************/

static int
dskreadint (FILE *fd)
{
   int r;
   int i;

   r = 0;
   for (i = 0; i < 4; i++)
   {
      int c;
      if ((c = fgetc (fd)) < 0)
      {
	 sprintf (view[0], "dskreadint: read failed: %s",
	       strerror (ERRNO));
         return (-1);
      }
      r = (r << 8) | (c & 0xFF);
   }
   return (r);
}

/***********************************************************************
* dskopen - Open disk and read geometry.
***********************************************************************/

int 
dskopen (Device *dev)
{
   int diskdata;

#ifdef DEBUGDSK
   fprintf (stderr, "dskopen: ENTERED\n");
#endif

   if (dev->infd == DEV_NULL)
   {
      return (0);
   }

   /*
   ** Seek to beginning of the disk
   */

   if (fseek (dev->infd, 0, SEEK_SET) < 0)
   {
      sprintf (view[0], "dskopen: seek failed: %s",
	    strerror (ERRNO));
      snprintf (view[1], MAXVIEWLEN, "filename: %s", dev->file);
      return (-1);
   }

   /*
   ** Read the disk geometry
   */

   if ((dev->info.dskinfo.cyls = dskreadint (dev->infd)) < 0) return (-1);
   if ((diskdata = dskreadint (dev->infd)) < 0) return (-1);
   dev->info.dskinfo.type = (diskdata >> 28) & 0xF;
   dev->info.dskinfo.model = (diskdata >> 16) & 0xFFF;
   dev->info.dskinfo.heads = diskdata & 0xFFFF;
   if ((diskdata = dskreadint (dev->infd)) < 0) return (-1);
   dev->info.dskinfo.overhead = 0;
   dev->info.dskinfo.secttrk = diskdata & 0xFFFF;
   if ((dev->info.dskinfo.sectlen = dskreadint (dev->infd)) < 0) return (-1);
   dev->info.dskinfo.tracklen = dev->info.dskinfo.secttrk * 
	       (dev->info.dskinfo.sectlen + dev->info.dskinfo.overhead);
	       
   /*
   ** Check geometry against model
   */

   if ((dev->info.dskinfo.model > MAXDISKS) ||
       (dev->info.dskinfo.type != DISK_CHS) ||
       (dev->info.dskinfo.type != disks[dev->info.dskinfo.model].type) ||
       (dev->info.dskinfo.cyls != disks[dev->info.dskinfo.model].cyls) ||
       (dev->info.dskinfo.heads != disks[dev->info.dskinfo.model].heads) ||
       (dev->info.dskinfo.secttrk != disks[dev->info.dskinfo.model].sectrk) ||
       (dev->info.dskinfo.sectlen != disks[dev->info.dskinfo.model].bytsec) ||
       (dev->info.dskinfo.overhead != disks[dev->info.dskinfo.model].overhead))
   {
      sprintf (view[0], "Unknown disk type or geometry mismatch");
      snprintf (view[1], MAXVIEWLEN, "filename: %s", dev->file);
      sprintf (view[2], "   model     = %d", dev->info.dskinfo.model);
      sprintf (view[3], "   type      = %d", dev->info.dskinfo.type);
      sprintf (view[4], "   cyls      = %d", dev->info.dskinfo.cyls);
      sprintf (view[5], "   heads     = %d", dev->info.dskinfo.heads);
      sprintf (view[6], "   secttrk   = %d", dev->info.dskinfo.secttrk);
      sprintf (view[7], "   sectlen   = %d", dev->info.dskinfo.sectlen);
      sprintf (view[8], "   overhead  = %d", dev->info.dskinfo.overhead);
      return (-1);
   }

#ifdef DEBUGDSK
   fprintf (stderr, "Disk geometry:\n");
   fprintf (stderr, "  model     = %s\n", disks[dev->info.dskinfo.model].model);
   fprintf (stderr, "  type      = %d\n", dev->info.dskinfo.type);
   fprintf (stderr, "  cyls      = %d\n", dev->info.dskinfo.cyls);
   fprintf (stderr, "  heads     = %d\n", dev->info.dskinfo.heads);
   fprintf (stderr, "  secttrk   = %d\n", dev->info.dskinfo.secttrk);
   fprintf (stderr, "  sectlen   = %d\n", dev->info.dskinfo.sectlen);
   fprintf (stderr, "  overhead  = %d\n", dev->info.dskinfo.overhead);
   fprintf (stderr, "  tracklen  = %d\n", dev->info.dskinfo.tracklen);
   fprintf (stderr, "  disk size = %d bytes\n", dev->info.dskinfo.tracklen *
	       dev->info.dskinfo.heads * dev->info.dskinfo.cyls);
#endif

   return (0);
}

/***********************************************************************
* dskreset - Reset device
***********************************************************************/

void
dskreset (Device *dev)
{
#ifdef DEBUGDSK
   fprintf (stderr, "dskreset: ENTERED\n");
#endif

   return;
}

/***********************************************************************
* dskseek - Seek to specified disk location.
***********************************************************************/

static int
dskseek (Device *dev, int cyl, int head, int sector, int statloc)
{
   int dskloc;

   /*
   ** Ensure cylinder is OK
   */

   memory[statloc] = 0;
   if (cyl >= dev->info.dskinfo.cyls)
   {
#ifdef DEBUGDSK
      fprintf (stderr, "dskseek: ERROR: cyl = %d, head = %d, sector = %d\n",
	       cyl, head, sector);
#endif
      memory[statloc] |= EODERR | ((dev->switches & SWPROTECT) ? WPBIT : 0);
      return (-1);
   }

   /*
   ** Calculate location
   */

   dskloc = (dev->info.dskinfo.tracklen * head) +
      (dev->info.dskinfo.tracklen * dev->info.dskinfo.heads * cyl) +
      ((dev->info.dskinfo.overhead + dev->info.dskinfo.sectlen) * sector) +
      DSKOVERHEAD;

#ifdef DEBUGDSK
   fprintf (stderr, "dskseek: cyl = %d, head = %d, sector = %d\n",
	    cyl, head, sector);
   fprintf (stderr, "   dskloc = %d\n", dskloc);
#endif

   /*
   ** Go to the specified location on disk
   */

   if (fseek (dev->infd, dskloc, SEEK_SET) < 0)
   {
      sprintf (view[0], "dskseek: seek failed: %s",
	    strerror (ERRNO));
      sprintf (view[1], "filename: %s", dev->file);
      return (-1);
   }

   return (0);
}

/***********************************************************************
* dskgetparms - Get parms from command list, based on disk model.
***********************************************************************/

static void
dskgetparms (Device *dev, uint16 word2, uint16 word3, DiskParams *params)
{
   switch (dev->info.dskinfo.model)
   {
   case 0: /* FD800 */
   case 1: /* FD1000 */
   case 2: /* DS31 */
      params->genint = word2 & 0x8000;
      params->sector = (word2 >> 8) & 0x7F;
      params->seccnt = word2 & 0xFF;

      params->chain = word3 & 0x8000;
      params->subcmd = (word3 >> 12) & 7;
      params->head = (word3 >> 8) & 0x1;
      params->cyl = word3 & 0xFF;
      break;

   case 3: /* DS44 */
      params->genint = word2 & 0x8000;
      params->sector = (word2 >> 8) & 0x7F;
      params->seccnt = word2 & 0xFF;

      params->chain = word3 & 0x8000;
      params->subcmd = (word3 >> 12) & 7;
      params->head = (word3 >> 8) & 0x1;
      params->cyl = (word3 & 0xFF) | ((word3 & 0x200) >> 1);
      break;

   default: ;
   }
}

/***********************************************************************
* dskbumpsector - Bump to next sector
***********************************************************************/

static void 
dskbumpsector (Device *dev, uint16 *cyl, uint16 *head, uint16 *sector)
{
   (*sector)++;
   if (*sector == dev->info.dskinfo.secttrk)
   {
      *sector = 0;
      (*head)++;
      if (*head == dev->info.dskinfo.heads)
      {
	 *head = 0;
	 (*cyl)++;
	 if (*cyl == dev->info.dskinfo.cyls)
	    *cyl = 0;
      }
   }
}

/***********************************************************************
* dskdocmd - Do disk command
***********************************************************************/

int 
dskdocmd (Device *dev, uint16 device, uint16 word2)
{
   Device *udev;
   int i;
   uint16 statloc;
   uint16 addr;
   uint16 len;
   uint16 chain;
   uint16 cmd;
   uint16 unit;
   uint16 genint;
   DiskParams params;

   statloc = ((device & 0x0007) << 1) + 0x0098;
   memory[statloc] = 0;
   memory[statloc+1] = word2;

   unit = (device & 0x0030) >> 4;
   cmd = (device >> 7) & 1;
   genint = 0;

#ifdef DEBUGDSK
   fprintf (stderr,
   	    "dskdocmd: device = %d, word2 = >%04X, unit = %d, cmd = %d\n",
	    device & 7, word2, unit, cmd);
   fprintf (stderr, "   statloc = >%04X\n", statloc);
#endif
   if (cmd)
      goto dsk_exit;

   /*
   ** Find selected unit
   */

   for (i = 0; i < devcnt; i++) /* find selected unit */
      if (devices[i].devaddr == dev->devaddr && devices[i].unit == unit)
	 break;
   if (i == devcnt)
   {
#ifdef DEBUGDSK
      fprintf (stderr, "Unit not configured\n", unit);
#endif
      /* Drive not ready */
      memory[statloc] = OPCOMP | OFFLIN;
      goto dsk_exit;
   }
   udev = &devices[i];

   do
   {
      int i;
      uint16 subcmd;
      uint16 sector, seccnt;
      uint16 head, cyl;
      uint16 d, c;

      memory[statloc+1] = word2;
      addr = getmem (word2);

      dskgetparms (dev, getmem (word2+1), getmem (word2+2), &params);
      genint = params.genint;
      sector = params.sector;
      seccnt = params.seccnt;
      chain = params.chain;
      subcmd = params.subcmd;
      head = params.head;
      cyl = params.cyl;

      len = udev->info.dskinfo.sectlen * seccnt;
#ifdef DEBUGDSK
      fprintf (stderr,
      	       "   subcmd = %d, buffer = >%04X, len = %d, chain = %x\n",
      	       subcmd, addr, len, chain);
#endif

      /*
      ** Check disk parameters
      */

      if (cyl >= udev->info.dskinfo.cyls)
      {
#ifdef DEBUGDSK
	 fprintf (stderr, "dskdocmd: ERROR: unit = %d: Invalid cylinder: %d\n",
		  unit, cyl);
#endif
	    memory[statloc] |= OPCOMP | EODERR |
			((udev->switches & SWPROTECT) ? WPBIT : 0);
	    return (genint);
      }

      if (subcmd == 3) /* Seeks only use cylinder */
      {
	 head = 0;
	 sector = 0;
      }
      else
      {
	 if (head >= udev->info.dskinfo.heads)
	 {
#ifdef DEBUGDSK
	    fprintf (stderr, "dskdocmd: ERROR: unit = %d: Invalid head: %d\n",
		     unit, head);
#endif
	    memory[statloc] |= OPCOMP | PGMERR |
			((udev->switches & SWPROTECT) ? WPBIT : 0);
	    return (genint);
	 }
	 if (sector >= udev->info.dskinfo.secttrk)
	 {
#ifdef DEBUGDSK
	    fprintf (stderr, "dskdocmd: ERROR: unit = %d: Invalid sector: %d\n",
		     unit, sector);
#endif
	    memory[statloc] |= OPCOMP | PGMERR |
			((udev->switches & SWPROTECT) ? WPBIT : 0);
	    return (genint);
	 }
      }

      /*
      ** Parms OK, seek
      */

      if (dskseek (udev, cyl, head, sector, statloc) < 0)
      {
	 memory[statloc] |= OPCOMP | PGMERR |
		     ((udev->switches & SWPROTECT) ? WPBIT : 0);
	 return (genint);
      }

      /*
      ** Set initial sate
      */

      udev->select = TRUE;

      switch (subcmd)
      {
      case 0: /* Read */
#ifdef DEBUGDSK
	 fprintf (stderr,
		  "   READ: cyl = %d, head = %d, sector = %d, len = %d\n",
		  cyl, head, sector, len);
#endif
	 while (len)
	 {
	    int slen;

	    /*
	    ** Read one sector at a time
	    */

	    slen = (len < udev->info.dskinfo.sectlen) ?
		  len : udev->info.dskinfo.sectlen;

	    for (i = 0; i < slen; i += 2)
	    {
	       d = fgetc (udev->infd);
	       c = fgetc (udev->infd);
	       d = (d << 8) | c;
	       dmaputword (dev, addr, d);
	       addr++;
	    }

	    len -= slen;
	    dskbumpsector (udev, &cyl, &head, &sector);

	    if (len)
	       if (dskseek (udev, cyl, head, sector, statloc) < 0)
		  return (genint);
	 }

	 udev->info.dskinfo.curcyl = cyl;
	 udev->info.dskinfo.curhead = head;
	 udev->info.dskinfo.cursect = sector;
         break;

      case 1: /* Read ID */
	 d = ((udev->info.dskinfo.curcyl & 0xFF) << 7) | 
	     ((udev->info.dskinfo.curcyl & 0x100) >> 2) | 
	     (udev->info.dskinfo.cursect & 0x3F);
	 dmaputword (dev, addr, d);
         break;

      case 2: /* Compare disk and memory */
#ifdef DEBUGDSK
	 fprintf (stderr,
		  "   READCMP: cyl = %d, head = %d, sector = %d, len = %d\n",
		  cyl, head, sector, len);
#endif
	 while (len)
	 {
	    int slen;

	    /*
	    ** Read one sector at a time
	    */

	    slen = (len < udev->info.dskinfo.sectlen) ?
		  len : udev->info.dskinfo.sectlen;

	    for (i = 0; i < slen; i += 2)
	    {
	       d = fgetc (udev->infd);
	       c = fgetc (udev->infd);
	       d = (d << 8) | c;
	       if (d != dmagetword (dev, addr))
	          memory[statloc] |= CMPERR;
	       addr++;
	    }

	    len -= slen;
	    dskbumpsector (udev, &cyl, &head, &sector);

	    if (len)
	       if (dskseek (udev, cyl, head, sector, statloc) < 0)
		  return (genint);
	 }

	 udev->info.dskinfo.curcyl = cyl;
	 udev->info.dskinfo.curhead = head;
	 udev->info.dskinfo.cursect = sector;
         break;

      case 3: /* Seek */
	 memory[statloc] |= (1 << unit);
	 udev->info.dskinfo.curcyl = cyl;
	 udev->info.dskinfo.curhead = head;
	 udev->info.dskinfo.cursect = sector;
         break;

      case 4: /* Write */
      case 7: /* Write protect override */
#ifdef DEBUGDSK
	 fprintf (stderr,
		  "   WRITE: cyl = %d, head = %d, sector = %d, len = %d\n",
		  cyl, head, sector, len);
#endif
	 if (!(udev->switches & SWPROTECT));
	 {
	    while (len)
	    {
	       int slen;

	       /*
	       ** Write one sector at a time
	       */

	       slen = (len < udev->info.dskinfo.sectlen) ?
		     len : udev->info.dskinfo.sectlen;

	       for (i = 0; i < slen; i += 2)
	       {
		  d = dmagetword (dev, addr);
		  fputc ((d >> 8) & 0xFF, udev->infd);
		  fputc (d & 0xFF, udev->infd);
		  addr++;
	       }

	       /*
	       ** Zero fill short sectors
	       */

#ifdef DEBUGDSKDATA
	       if (i < udev->info.dskinfo.sectlen)
		  fprintf (stderr, "Pad %d bytes\n",
			udev->info.dskinfo.sectlen - i);
#endif
	       for (; i < udev->info.dskinfo.sectlen; i++)
		  fputc ('\0', udev->infd);

	       len -= slen;
	       dskbumpsector (udev, &cyl, &head, &sector);

	       if (len)
		  if (dskseek (udev, cyl, head, sector, statloc) < 0)
		     return (genint);
	    }

	    udev->info.dskinfo.curcyl = cyl;
	    udev->info.dskinfo.curhead = head;
	    udev->info.dskinfo.cursect = sector;
	 }
         break;

      case 5: /* Write ID w/o protect */
         break;

      case 6: /* Write ID with protect */
         break;

      }
      if (chain)
         word2 = getmem (word2+3);
   } while (chain);

dsk_exit:
   memory[statloc] |= OPCOMP;
   return (genint);
}

/***********************************************************************
* dskboot - Boot from device.
***********************************************************************/

int
dskboot (Device *dev)
{
   int i;
   int addr;
   uint16 d, c;

   if (fseek (dev->infd, DSKOVERHEAD, SEEK_SET) < 0)
   {
      sprintf (view[0], "Device %s: seek failed: %s",
	       dev->name, strerror (ERRNO));
      sprintf (view[1], "filename: %s", dev->file);
      return (-1);
   }

   addr = 0;
   SETPC (addr);
   for (i = 0; i < dev->info.dskinfo.sectlen; i += 2)
   {
      d = fgetc (dev->infd);
      c = fgetc (dev->infd);
      d = (d << 8) | c;
      dmaputword (dev, addr, d);
      addr++;
   }
   run = TRUE;

   return (0);
}
