/***********************************************************************
*
* simio.c - IO processing for the TI 980 Simulator.
*
* Changes:
*   06/24/11   DGP   Original.
*   08/31/11   DGP   Added PIF/MP support. Fix interrupts.
*   09/14/11   DGP   Added cpu models and Calder 980 support.
*   10/25/11   DGP   Added CDD Common Data Feature.
*   01/12/11   DGP   Fix test bit operations.
*   03/26/13   DGP   Fix up card punch.
*   08/23/16   DGP   Expand RDS status.
*                    Added SCSI disk support.
*   08/24/16   DGP   Revamp console support.
*
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>

#if defined (WIN32)
#define __TTYROUTINES 0
#include <conio.h>
#include <windows.h>
#include <signal.h>
#include <process.h>
#ifndef HZ
#define HZ 100
#endif
#endif

#if defined(UNIX)
#include <unistd.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <pthread.h>
#endif

#include "simdef.h"
#include "simcra.h"

extern uint16 statreg;	/* The program status register */

extern int run;
extern int idle;
extern int devcnt;
extern int enablepanel;
extern int deferint;
extern int model;
extern char view[MAXVIEW][MAXVIEWLEN+1];

extern uint32 intrequest;
extern uint32 dmadevint;
extern uint32 memlen;
extern uint16 regs[16];
extern uint16 memory[MAXMEMSIZE];
extern Device devices[MAXDEVICES];

extern int cddcmnenable;
extern uint16 cddcmnlimit;
extern uint32 cddcmnbank;

extern uint32 lowlimitreg;
extern uint32 upperlimitreg;
extern uint32 cddlowlimitext;
extern uint32 cddupperlimitext;

extern uint32 cpumemmap[2];
extern uint32 dmamemmap[8];

#ifdef UNIX
static pthread_mutex_t genint_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif
#ifdef WIN32
static HANDLE mhandle;
#endif

/***********************************************************************
* geninterrupt - Generate interrupt
***********************************************************************/

int
geninterrupt (uint16 loc, uint16 setbit)
{
#ifdef UNIX
   pthread_mutex_lock (&genint_mutex);
#endif
#ifdef WIN32
   if (WaitForSingleObject (mhandle, INFINITE) != WAIT_OBJECT_0)
   {
      fprintf (stderr, "geninterrupt: mutex aquire failed: err = %d",
               GetLastError());
      exit (1);
   }
#endif

   intrequest |= (1 << (loc / 2));
   if (loc == INTINTLOC)
      statreg |= setbit;
   else if (loc == DMAINTLOC)
      dmadevint |= (0x0080 << (8 - (setbit & 7)));

#ifdef UNIX
   pthread_mutex_unlock (&genint_mutex);
#endif
#ifdef WIN32
    ReleaseMutex (mhandle);
#endif

   return (0);
}

/***********************************************************************
* clrinterrupt - Clear interrupt
***********************************************************************/

static void
clrinterrupt (uint16 loc)
{
#ifdef UNIX
   pthread_mutex_lock (&genint_mutex);
#endif
#ifdef WIN32
   if (WaitForSingleObject (mhandle, INFINITE) != WAIT_OBJECT_0)
   {
      fprintf (stderr, "clrinterrupt: mutex aquire failed: err = %d",
               GetLastError());
      exit (1);
   }
#endif

   intrequest &= ~(1 << (loc / 2));

#ifdef UNIX
   pthread_mutex_unlock (&genint_mutex);
#endif
#ifdef WIN32
    ReleaseMutex (mhandle);
#endif
}

/***********************************************************************
* chkinterrupt - Check interrupt
***********************************************************************/

static int
chkinterrupt (uint16 loc)
{
   uint16 i;

#ifdef UNIX
   pthread_mutex_lock (&genint_mutex);
#endif
#ifdef WIN32
   if (WaitForSingleObject (mhandle, INFINITE) != WAIT_OBJECT_0)
   {
      fprintf (stderr, "chkinterrupt: mutex aquire failed: err = %d",
               GetLastError());
      exit (1);
   }
#endif

   i = intrequest & (1 << (loc / 2));

#ifdef UNIX
   pthread_mutex_unlock (&genint_mutex);
#endif
#ifdef WIN32
    ReleaseMutex (mhandle);
#endif
#ifdef DEBUGINTS
   if (i)
      fprintf (stderr,"chkinterrupt: loc = %x TRUE\n", loc);
#endif
   return (i);
}

/***********************************************************************
* checkinterrupts - If any pending and not masked, take it.
***********************************************************************/

void 
checkinterrupts (void)
{
   Device *dev;
   int curloc;
   static int scancount = 0;
   int i;
   int keyhit;
   uint16 curinst;

   do
   {
      scancount++;
      curinst = getinst (FALSE);

      /*
      ** Check the keyboards for interrupts.
      */

      keyhit = FALSE;
      if (idle || scancount > 10000)
      {
	 int c;

         /*
         ** Check for console input, console always device 0.
         */

	 dev = &devices[0];
	 if (dev->infd != DEV_NULL && IS_DBINT)
	 {
	    c = serialpollchar (dev, 10);
	    if (c != EOF)
	    {
	       dev->select = TRUE;
	       keyhit = TRUE;
	       dev->inchar = c;
	       geninterrupt (dev->intloc, dev->devaddr);
	    }
	    else if (run == FALSE)
	    {
	       return;
	    }
	 }
	 else
	 {
	    c = serialgetchar (dev);
	    if (c != EOF)
	    {
	       dev->select = TRUE;
	       keyhit = TRUE;
	       dev->inchar = c;
	    }
	    if (run == FALSE)
	    {
	       return;
	    }
	 }

	 scancount = 0;
      }

      /*
      ** If we have interrupts, process them.
      */

      if (intrequest)
      {
	 curloc = 48;
	 for (i = 2; intrequest && (i < curloc); i += 2)
	 {
	    if (chkinterrupt (i))
	    {
	       uint16 opc, npc;
	       uint16 ost;

	       npc = i;
	       if (idle && IS_IDLE (curinst))
		  INCPC(1);
	       curinst = memory[npc];

	       /*
	       ** If trap location is not set or illegal, stop execution.
	       */

	       if (!curinst || (decodeinst (curinst) == TYPE_ILL))
	       {
		  run = FALSE;
		  return;
	       }

	       if (npc == DMAINTLOC)
	          memory[DMASTATLOC] = dmadevint & 0xFFFF;
	       ost = statreg;
	       opc = GETPC;
	       clrinterrupt (i);
	       idle = FALSE;
	       deferint = 2;

	       /*
	       ** If the trap inst is an SSB, do it here.
	       */

	       if (curinst == SSBINST)
	       {
	          if (npc == INTINTLOC)
		     statreg &= 0xF034;
	          else if (npc == DMAINTLOC)
		     statreg &= 0xF0B7;
	          else if (npc == DBUSINTLOC)
		     statreg &= 0xF0BF;
		  npc = memory[npc+1];

		  memory[npc] = opc;
		  memory[npc+1] = ost;
		  SETPC (memory[npc+2]);
	       }

	       else
	       {
	          SETPC (opc);
		  statreg = ost;
	       }
	       return;
	    }
	 }
      }

      /*
      ** If nothing to do snooze a bit
      */

      if (idle && scancount && !keyhit)
      {
#ifdef LINUX
         sched_yield();
#else
	 smallsleep(1);
#endif
      }

   } while (idle);
}

/***********************************************************************
* finddev - Find attached device.
***********************************************************************/

static int
finddev (uint16 devaddr)
{
   int i;

   for (i = 0; i < devcnt; i++)
   {
      if (devices[i].devaddr == devaddr && devices[i].infd != NULL) break;
   }

   if (i == devcnt)
   {
#ifdef DEBUGFINDDEV
      fprintf (stderr, "Device for devaddr %04X is not attached\n", devaddr);
#endif
      return (-1);
   }
   return (i);
}

/***********************************************************************
* findtype - Find attached device by type.
***********************************************************************/

static int
findtype (int type, char *dt)
{
   int i;

   for (i = 0; i < devcnt; i++)
   {
      if (devices[i].type == type && devices[i].infd != NULL &&
	  !strcmp (devices[i].name, dt) ) return (i);
   }

   return (-1);
}

/***********************************************************************
* attachit - Attach a file to a device (do the dirty work).
***********************************************************************/

static void
attachit (char *bp, int type, char *mode, char *fdev)
{
   Device *dev;
   FILE *infd;
   FILE *onfd;
   char *attype;
   long val;
   int i;
   int devaddr;
   int dmaport;
   int protectflag;
   int upcaseflag;
   int createflag;
   int scsiflag;
   char lclmode[32];

   /*
   ** Initialize locals.
   */

   strcpy (lclmode, mode);
   protectflag = FALSE;
   upcaseflag = FALSE;
   createflag = FALSE;
   scsiflag = FALSE;

   /*
   ** Check for device switches
   */

   if (*bp == '/')
   {
      bp++;
      switch (type)
      {
      case DISK:
	 while (*bp && !isspace (*bp))
	 {
	    if (*bp == 'r') protectflag = TRUE;
	    else if (*bp == 's') scsiflag = TRUE;
	    bp++;
	 }
	 break;

      case TAPE:
	 while (*bp && !isspace (*bp))
	 {
	    if (*bp == 'c') createflag = TRUE;
	    else if (*bp == 'r') protectflag = TRUE;
	    bp++;
	 }
	 break;

      case SERIAL:
      case CONSOLE:
	 while (*bp && !isspace (*bp))
	 {
	    if (*bp == 'u') upcaseflag = TRUE;
	    bp++;
	 }
	 break;

      default:
	 while (*bp && !isspace(*bp)) bp++;
      }
   }

   /*
   ** Get device address
   */

   bp = getnum (bp, &val);
   devaddr = val & 0xFFFF;

   /*
   ** Sanity check device address
   */

   dmaport = 0;
   switch (type)
   {
   case PRINTER:
   case TAPE:
   case DISK:
      if (devaddr > 7)
      {
	 sprintf (view[0], "attach: illegal device address %d", devaddr);
	 return;
      }
      bp = getnum (bp, &val);
      dmaport = val & 0xFFFF;
      if (dmaport > 7)
      {
	 sprintf (view[0], "attach: illegal DMA port %d", dmaport);
	 return;
      }
      break;

   case SERIAL:
   case CONSOLE:
   case CARDREADER:
   case CARDPUNCH:
      if ((devaddr < 33) || ((devaddr > 63) && (devaddr < 97)))
	 break;
      sprintf (view[0], "attach: illegal device address %d", devaddr);
      return;

   default:
      return;
   }

   /*
   ** Do the open
   */

   while (*bp && isspace (*bp)) bp++;

   attype = "filename";
   if (!strcmp (bp, "NULL") || !strcmp (bp, "null"))
   {
      strcpy (bp, "NULL");
      infd = onfd = DEV_NULL;
   }
   else if (!strcmp (bp, "TTY") || !strcmp (bp, "tty"))
   {
      strcpy (bp, "TTY");
      infd = stdin;
      onfd = stdout;
   }
   else
   {
      if (createflag)
	 lclmode[0] = 'w';
      infd = onfd = fopen (bp, lclmode);
   }

   if (infd == NULL)
   {
      sprintf (view[0], "attach: %s open failed: %s", fdev, strerror (ERRNO));
      sprintf (view[1], "%s = %s", attype, bp);
      return;
   }

   /*
   ** Locate a device table slot
   */

   if ((i = findtype (type, fdev)) >= 0)
   {
      if (devices[i].infd != stdin && devices[i].infd != DEV_NULL)
	    fclose (devices[i].infd);
      if (devices[i].infd != devices[i].outfd)
      {
	 if (devices[i].outfd != stdout && devices[i].outfd != DEV_NULL)
	    fclose (devices[i].outfd);
      }
   }
   else
   {
      if (devcnt == MAXDEVICES)
      {
	 sprintf (view[0], "attach: Too many devices, MAXDEVICES = %d",
		  MAXDEVICES);
	 return;
      }
      else
      {
	 i = devcnt;
      }
   }

   /*
   ** Fill it in
   */

   dev = &devices[i];
   memset ((char *)dev, '\0', sizeof (Device));

   dev->infd = infd;
   dev->outfd = onfd;
   dev->devaddr = devaddr;
   dev->type = type;
   dev->select = 0;
   dev->count = 0;
   dev->unit = 0;
   dev->viewndx = 0;

   if (upcaseflag)
      dev->switches |= SWUPCASE;
   if (createflag)
      dev->switches |= SWCREATE;
   if (protectflag)
      dev->switches |= SWPROTECT;
   if (scsiflag)
      dev->switches |= SWSCSI;

   strcpy (dev->file, bp);
   strcpy (dev->name, fdev);

   switch (type)
   {
   case DISK:
      dev->unit = atoi(&fdev[1]) & 3;
      dev->devaddr |= DMA_DEVICE;
      dev->intloc = DMAINTLOC;
      dev->dmaport = dmaport;
      if (dev->switches & SWSCSI)
	 sdskopen (dev);
      else
	 dskopen (dev);
#ifdef DEBUGDSK
      fprintf (stderr, "   unit = %02X\n", dev->unit);
#endif
      break;

   case TAPE:
      dev->unit = atoi(&fdev[1]) & 3;
      dev->devaddr |= DMA_DEVICE;
      dev->intloc = DMAINTLOC;
      dev->dmaport = dmaport;
      tapeopen (dev);
#ifdef DEBUGTAPE
      fprintf (stderr, "   unit = %02X\n", dev->unit);
#endif
      break;

   case PRINTER:
      dev->intloc = DMAINTLOC;
      dev->devaddr |= DMA_DEVICE;
      dev->dmaport = dmaport;
      break;

   case CARDPUNCH:
      dev->intloc = DBUSINTLOC;
      dev->info.cardinfo.cardndx = 0;
      break;

   case CARDREADER:
      dev->intloc = DBUSINTLOC;
      dev->info.cardinfo.cardeof = FALSE;
      dev->info.cardinfo.cardndx = 100;
      break;

   case CONSOLE:
   case SERIAL:
      dev->intloc = DBUSINTLOC;
      break;

   default: ;
   }

   devcnt++;
}

/***********************************************************************
* attachdev - Attach a file to a device.
***********************************************************************/

void
attachdev (char *bp)
{
   char *ep;
   long val;
   int  devnum;
   char fdev[10];

   while (*bp && isspace (*bp)) bp++;
   ep = bp;
   switch (*bp++)
   {
      case 'c':
	 switch (*bp++)
	 {
	 case 'o': /* Console */
	    attachit (bp, CONSOLE, "r", "CO");
	    break;

	 case 'p': /* Card punch */
	    attachit (bp, CARDPUNCH, "w", "CP");
	    break;

	 case 'r': /* Card reader */
	    attachit (bp, CARDREADER, "r", "CR");
	    break;

	 default: 
	    goto BADDEV;
	 }
	 break;

      case 'd': /* Disk */
	 if (!isdigit (*bp)) goto BADDEV;
	 bp = getnum (bp, &val);
	 devnum = val & 0xFFFF;
	 sprintf (fdev, "D%d", devnum);
	 attachit (bp, DISK, "r+b", fdev);
	 break;

      case 'm': /* MagTape */
	 if (!isdigit (*bp)) goto BADDEV;
	 bp = getnum (bp, &val);
	 devnum = val & 0xFFFF;
	 sprintf (fdev, "M%d", devnum);
	 attachit (bp, TAPE, "r+b", fdev);
	 break;

      case 'p':
	 switch (*bp++)
	 {
	 case 'r': /* Printer */
	    attachit (bp, PRINTER, "w", "PR");
	    break;

	 default:
	    goto BADDEV;
	 }
	 break;

      case 's': /* Serial */
	 attachit (bp, SERIAL, "r+", "SE");
	 break;

      default:
BADDEV:
	 sprintf (view[0], "Invalid device %s, devices are:", ep);
	 sprintf (view[1], " co = console      pr = printer ");
	 sprintf (view[2], " cp = card punch   cr = card reader");
	 sprintf (view[3], " dn = DISK n=0-3   mn = TAPE n=0-3");
	 break;
   }
}

/***********************************************************************
* resetdev - Reset devices for RSET.
***********************************************************************/

void
resetdev (void)
{
   Device *dev;
   int i;

   intrequest = 0; /* Reset all interrupts */

   for (i = 0; i < devcnt; i++)
   {
      dev = &devices[i];
      switch (dev->type)
      {
      case TAPE:
	 tapereset (dev);
	 break;
      case DISK:
	 if (dev->switches & SWSCSI)
	    sdskreset (dev);
	 else
	    dskreset (dev);
	 break;
      default: ;
      }
   }
}

/***********************************************************************
* dmamapaddr - Map DMA address
***********************************************************************/

static uint32
dmamapaddr (Device *dev, uint32 ma)
{
   if ((model == CPU_CDD980) && (ma >= 0x0100))
   {
      uint32 dmacmnlimit;

      dmacmnlimit = dmamemmap[dev->dmaport] & 0xFF00;
      if (dmacmnlimit > 0)
      {
	 if (ma >= dmacmnlimit)
	    ma += (dmamemmap[dev->dmaport] & 0xF) << 14;
      }
      else
	 ma += (dmamemmap[dev->dmaport] & 0xF) << 14;
   }
   return (ma);
}

/***********************************************************************
* dmaputword - Put data into memory.
***********************************************************************/

void
dmaputword (Device *dev, uint16 pa, uint16 v)
{
   uint32 ma = pa;

   ma = dmamapaddr (dev, ma);

   if (ma < memlen)
      memory[ma] = v;
}

/***********************************************************************
* dmagetword - Get data from memory.
***********************************************************************/

uint16
dmagetword (Device *dev, uint16 pa)
{
   uint32 ma = pa;

   ma = dmamapaddr (dev, ma);

   if (ma < memlen)
      return (memory[ma]);
   return (0xFFFF);
}

/***********************************************************************
* dmaputbyte - Put a byte in memory.
***********************************************************************/

void
dmaputbyte (Device *dev, uint16 pa, uint16 bno, uint8 v)
{
   uint32 ma = pa;
   uint16 word;

   ma = dmamapaddr (dev, ma);

   if (ma < memlen)
   {
      word = memory[ma];
      if (bno)
	 word = (word & 0xFF00) | v;
      else
	 word = (word & 0x00FF) | (v << 8);
      memory[ma] = word;
   }
}

/***********************************************************************
* dmagetbyte - Get a byte from memory.
***********************************************************************/

uint8
dmagetbyte (Device *dev, uint16 pa, uint16 bno)
{
   uint32 ma = pa;
   uint16 word;

   ma = dmamapaddr (dev, ma);

   if (ma < memlen)
   {
      word = memory[ma];
      return (bno ? (word & 0x00FF) : (word >> 8));
   }
   return (0xFF);
}

/***********************************************************************
* doterminal - Process terminal inst.
***********************************************************************/

static int
doterminal (int rw, Device *dev, uint16 device, uint16 word2)
{
   uint8 ch;

   if (rw == 0) /* RDS */
   {
      if (dev->infd == DEV_NULL) return (0);

      if (!dev->select) {
      	 regs[word2 & 0x7] = 0x5000;
	 if (word2 & DBUS_BUSY)
            return (1);
         return (0);
      }

      dev->select = FALSE;
      if (dev->switches & SWUPCASE && islower (dev->inchar))
	 dev->inchar = toupper (dev->inchar);
      if (dev->inchar == '\n') dev->inchar = '\r';
      regs[word2 & 0x7] = dev->inchar;
   }

   else /* WDS */
   {
      if (dev->outfd == DEV_NULL) return (0);

      ch = (uint8)regs[word2 & 0x7] & 0x7F;
#ifdef DEBUGCON
      fprintf (stderr,
	 "WRS: device = >%04X, word2 = >%04X, reg[%d] = >%04X, ch = %c(%02X)\n",
               device, word2, word2 & 0x7, regs[word2 & 0x7],
	       (ch > 0x1F && ch < 0x7F) ? ch : '.', ch);
#endif

      serialputchar (dev, ch);
      if (IS_DBINT)
	 geninterrupt (dev->intloc, dev->devaddr);
   }
   return (0);
}

/***********************************************************************
* docardreader - Process card reader.
***********************************************************************/

static void
docardreader (int rw, Device *dev, uint16 device, uint16 word2)
{
   if (rw == 0) /* RDS */
   {
      int cdat;

      /*
      ** If record exhausted, read another
      */

      if (dev->info.cardinfo.cardndx >= 80)
      {
	 char *cbuf;
	 int i1, o1;
	 int len;
	 char temp[1024];

	 /*
	 ** Read record
	 */

	 cbuf = dev->info.cardinfo.carddat;
	 if (!(dev->info.cardinfo.cardeof) && dev->infd)
	 {
	    if (fgets (temp, 1024, dev->infd) == NULL)
	    {
	       strcpy (temp, "/*\n");
	       dev->info.cardinfo.cardeof = TRUE;
	    }
	 }
	 else
	 {
	    strcpy (temp, "/*\n");
	 }

	 /*
	 ** Detab input record
	 */

	 len = strlen (temp);
	 for (i1 = o1 = 0; i1 < len && o1 < 80; i1++)
	 {
	    if (temp[i1] == '\t')
	    {
	       int o2;

	       for (o2 = 8 - (o1 % 8); o2; o2--)
	       {
		  cbuf[o1++] = ' ';
	       }
	    }
	    else
	    {
	       cbuf[o1++] = temp[i1];
	    }
	 }
	 cbuf[o1] = '\0';

	 /*
	 ** Pad with blanks
	 */

	 while (o1 < 80)
	 {
	    cbuf[o1++] = ' ';
	 }
	 cbuf[80] = '\0';

#ifdef DEBUGCARD
	 fprintf (stderr, "CARDREC:\n");
	 HEXDUMP (stderr, cbuf, 80, 0);
#endif
	 dev->info.cardinfo.cardndx = 0;
      }

      /*
      ** Get a character from the record
      */

      cdat = dev->info.cardinfo.carddat[dev->info.cardinfo.cardndx];
      dev->info.cardinfo.cardndx ++;

      /*
      ** Translate to card code
      */

      regs[0] = tocard[cdat];
      if (dev->info.cardinfo.cardndx >= 80)
         regs[0] |= 0x4000;
   }

   else /* WDS */
   {
      /* Check to Feed a new card or skip to end of card */
      if (regs[0] & 3)
      {
	 dev->info.cardinfo.cardeof = FALSE;
	 dev->info.cardinfo.cardndx = 100;
      }
   }
   if (IS_DBINT)
      geninterrupt (dev->intloc, dev->devaddr);
}

/***********************************************************************
* docardpunch - Process card punch.
***********************************************************************/

static void
docardpunch (int rw, Device *dev, uint16 device, uint16 word2)
{
   if (rw == 1) /* WDS */
   {
      int i;
      int cdat;

      /*
      ** Translate from card code
      */

      cdat = regs[0] & 0xFFF;
      if (cdat == 0)
      {
	 cdat = ' ';
      }
      else
      {
	 for (i = 0x21; i < 0x80; i++)
	 {
	    if (cdat == tocard[i])
	    {
	       cdat = i;
	       break;
	    }
	 }
	 if (i == 0x80)
	    cdat = ' ';
      }
      
      fputc (cdat, dev->outfd);
      dev->info.cardinfo.cardndx ++;

      /*
      ** Time to punch the card
      */

      if ((regs[0] & 0x4000) || (dev->info.cardinfo.cardndx >= 80))
      {
	 fputc ('\n', dev->outfd);
	 dev->info.cardinfo.cardndx = 0;
      }
   }
   if (IS_DBINT)
      geninterrupt (dev->intloc, dev->devaddr);
}

/***********************************************************************
* busincaddr - Process I,A and reg in command.
***********************************************************************/

static void
busincaddr (uint16 word)
{
   uint16 addr;
   uint16 reg;

   reg = word & 0x0007;
   addr = 0;

   if (reg == 0) return;
   if (word & DBUS_ADDR)
   {
      addr = regs[reg];
      putmem (addr, getmem (addr) + ((word & DBUS_INC) ? 1 : (-1)));
   }
   else
   {
      regs[reg] = regs[reg] + ((word & DBUS_INC) ? 1 : (-1));
   }
}

/***********************************************************************
* doRDS - Process RDS inst.
***********************************************************************/

int
doRDS (uint16 device, uint16 word2)
{
   Device *dev;
   int status;
   int devindex;

   status = 0;

   /*
   ** Check if PM/PIF device.
   */

   if (device == PIFLOWADD)
   {
      regs[0] = lowlimitreg;
   }
   else if (device == PIFUPPADD)
   {
      regs[0] = upperlimitreg;
   }

   /*
   ** Regular device.
   */

   else
   {
      if ((devindex = finddev (device)) < 0)
      {
#ifdef DEBUGDEVICES
	 fprintf (stderr, "UNKNOWN RDS device: device = >%04X, word2 = >%04X\n",
		  device, word2);
#endif
	 status = -1;
	 regs[word2 & 0x7] = 0;
	 return (status);
      }

      dev = &devices[devindex];
      switch (dev->type)
      {
      case CONSOLE:
      case SERIAL:
	 if (doterminal (0, dev, device, word2))
	    return (status);
	 break;

      case CARDPUNCH:
	 docardpunch (0, dev, device, word2);
	 break;

      case CARDREADER:
	 docardreader (0, dev, device, word2);
	 break;

      default: ;
      }
   }

   busincaddr (word2);

   if (word2 & DBUS_BUSY)
      INCPC (1);

   return (status);
}

/***********************************************************************
* doWDS - Process WDS inst.
***********************************************************************/

int
doWDS (uint16 device, uint16 word2)
{
   Device *dev;
   int status;
   int devindex;

   status = 0;

   /*
   ** Check if PM/PIF device.
   */

   if (device == PIFLOWADD)
   {
      lowlimitreg = regs[0];
   }
   else if (device == PIFUPPADD)
   {
      upperlimitreg = regs[0];
   }

   /*
    * CDD980 devices
    */

   else if ((device == CDDMEMMAP) && (model == CPU_CDD980))
   {
      if (regs[0] & 0x8000)
         cpumemmap[1] = (uint32)(regs[0] & 0x00F0) << 10;
      if (regs[0] & 0x4000)
         cpumemmap[0] = (uint32)(regs[0] & 0x000F) << 14;
   }
   else if ((device == CDDCDBLIM) && (model == CPU_CDD980))
   {
      cddcmnenable = regs[0] & 0x0080;
      cddcmnlimit = regs[0] & 0xF800;
      cddlowlimitext = ((uint32)(regs[0] & 3) << 16);
   }
   else if ((device == CDDBNKSEL) && (model == CPU_CDD980))
   {
      cddcmnbank = ((uint32)(regs[0] & 0xF000) << 4);
      cddupperlimitext = ((uint32)(regs[0] & 3) << 16);
   }
   else if ((device >= CDDDMMAP0) && (device <= CDDDMMAP7) &&
            (model == CPU_CDD980))
   {
      dmamemmap[device - CDDDMMAP0] = (uint32)regs[0];
   }

   /*
   ** Regular device.
   */

   else
   {
      if ((devindex = finddev (device)) < 0)
      {
#ifdef DEBUGDEVICES
	 fprintf (stderr, "UNKNOWN WDS device: device = >%04X, word2 = >%04X\n",
		  device, word2);
#endif
	 status = -1;
	 goto exit_WDS;
      }

      dev = &devices[devindex];
      switch (dev->type)
      {
      case CONSOLE:
      case SERIAL:
	 doterminal (1, dev, device, word2);
	 break;

      case CARDPUNCH:
	 docardpunch (1, dev, device, word2);
	 break;

      case CARDREADER:
	 docardreader (1, dev, device, word2);
	 break;

      default: ;
      }
   }

   busincaddr (word2);

exit_WDS:
   if (word2 & DBUS_BUSY)
      INCPC (1);

   return (status);
}

/***********************************************************************
* doprinter - Process printer.
***********************************************************************/

static int
doprinter (Device *dev, uint16 device, uint16 word2)
{
   uint16 statloc;
   uint16 buffer;
   uint16 length;
   uint16 chain;
   uint16 genint;

   statloc = ((device & 0x0007) << 1) + 0x0098;
   memory[statloc] = 0;

#ifdef DEBUGPRINTER
   fprintf (stderr,
   	    "doprinter: device = %d, word2 = >%04X, statloc = >%04X\n",
      	    device & 7, word2, statloc);
#endif

   do {
      uint16 bit;
      uint8 ch;

      buffer = getmem (word2);
      length = getmem (word2+1);
      chain = getmem (word2+2) & 0x4000;
      genint = getmem (word2+2) & 0x8000;

#ifdef DEBUGPRINTER
      fprintf (stderr, "   buffer = >%04X, length = %d, chain = %x\n",
      	       buffer, length, chain);
#endif
      bit = 0;
      for (; length; length--)
      {
	 ch = dmagetbyte (dev, buffer, bit) & 0x7F;
#ifdef DEBUGPRINTERDATA
	 fprintf (stderr,
	 	  "     length = %d, buffer = >%04X, bit = %d, ch = %c(%02X)\n",
	 	  length, buffer, bit, (ch < ' ') ? '.' : ch, ch);
#endif
	 bit ^= 1;
	 if (!bit) buffer++;
	 if (ch == 0x0C) fputc (0x0A, dev->outfd);
	 if ((ch > 0x1F && ch < 0x7F) || (ch > 0x07 && ch < 0x0D))
	    fputc (ch, dev->outfd);
      }
      if (chain)
         word2 = getmem (word2+3);
   } while (chain);

   memory[statloc] = 0x0001;
   return (genint);
}

/***********************************************************************
* doATI - Process ATI inst.
***********************************************************************/

int
doATI (uint16 device, uint16 word2)
{
   Device *dev;
   int status;
   int devindex;
   int genint;
   uint16 devaddr;
   uint16 statloc;

   status = 0;
   genint = 0;
   devaddr = (device & 0x7) | DMA_DEVICE;
   statloc = ((device & 0x7) << 1) + 0x0098;
   dmadevint &= ~(0x0080 << (8 - (device & 7)));
   memory[statloc] = 0;

   if ((devindex = finddev (devaddr)) < 0)
   {
#ifdef DEBUGDEVICES
      fprintf (stderr, "UNKNOWN ATI device: device = >%04X, word2 = >%04X\n",
	       device, word2);
#endif
      status = -1;
      memory[statloc] = OPCOMP | OFFLIN;
      goto exit_ATI;
   }

   dev = &devices[devindex];
   switch (dev->type)
   {
   case DISK:
      if (dev->switches & SWSCSI)
	 genint = sdskdocmd (dev, device, word2);
      else
	 genint = dskdocmd (dev, device, word2);
      break;

   case TAPE:
      genint = tapedocmd (dev, device, word2);
      break;

   case PRINTER:
      genint = doprinter (dev, device, word2);
      break;

   default:
      memory[statloc] = OPCOMP | OFFLIN;
   }

exit_ATI:
   if (genint && IS_DMAINT)
      geninterrupt (dev->intloc, dev->devaddr);

   return (status);
}

/***********************************************************************
* boot - Boot from device.
***********************************************************************/

int
boot (char *device)
{
   Device *dev;
   int i;

   run = FALSE;

   while (*device && isspace (*device)) device++;
   if (islower (*device)) *device = toupper (*device);

   for (i = 0; i < devcnt; i++)
   {
      if (!strcmp (devices[i].name, device)) break;
   }

   if (i == devcnt)
   {
      sprintf (view[0], "Boot device %s is not attached", device);
      return (-1);
   }

   dev = &devices[i];

   if (*device == 'D') /* Boot from disk */
   {
      if (dev->switches & SWSCSI)
	 return (sdskboot (dev));
      return (dskboot (dev));
   }
   else if (*device == 'M') /* Boot from tape */
   {
      return (tapeboot (dev));
   }
   else
   {
      sprintf (view[0], "Boot from device %s is not supported", device);
      return (-1);
   }

   return (0);
}
