/***********************************************************************
*
* simloader.c - Binary loader for the TI 980 simulator.
*
* Changes:
*   06/24/11   DGP   Original.
*   08/03/11   DGP   Added 980 card format object support.
*
***********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>

#include "simdef.h"

extern int run;
extern int dispaddr;
extern uint16 regs[16];
extern uint32 memlen;

extern char view[MAXVIEW][MAXVIEWLEN+1];
extern uint16 memory[MAXMEMSIZE];

/************************************************************************
* getnumfield - Gets the next numeric field.
************************************************************************/

static uint16
getnumfield (unsigned char *op, int binarymode)
{
   uint16 wdata;
   char item[8];

   if (binarymode)
   {
      wdata = (*op << 8) | *(op+1);
   }
   else
   {
      int wd;

      strncpy (item, (char *)op, 4);
      item[4] = '\0';
      sscanf (item, "%4X", &wd);
      wdata = wd & 0xFFFF;
   }
   return (wdata);
}

/***********************************************************************
* loadcardfile - Load a card mode object file.
***********************************************************************/

static int
loadcardfile (FILE *fd, FILE *lfd, int loadaddr, int *curraddr)
{
   int status = 0;
   int idtrec = TRUE;
   int reloprog;
   int refcnt, defcnt;
   uint16 reloword;
   uint8 inbuf[82];

   defcnt = refcnt = 0;
   while (fgets ((char *)inbuf, 82, fd) != NULL)
   {
      uint8 *op = (uint8 *)inbuf;
      uint16 wdata;

#ifdef DUMPLOADREC
      reclen = strlen (inbuf);
      fprintf (lfd, "Record:\n");
      HEXDUMP (lfd, inbuf, reclen, 0);
#endif

      /*
      ** The first "card" is the IDT record.
      */

      if (idtrec)
      {
	 int i;
	 char *tp;

         idtrec = FALSE;
	 op += IDTLEN;

	 /* Sanity check */

	 tp = (char*)op;
	 if (*tp != '0' && *tp != '1')
	    goto BAD_OBJECT;
	 tp++;
	 for (i = 0; i < 8; i++)
	    if (!isxdigit(*tp++))
	       goto BAD_OBJECT;
	 for (i = 0; i < 3; i++)
	 {
	    if (!isdigit(*tp))
	       goto BAD_OBJECT;
	    defcnt = (defcnt * 10) + (*tp - '0');
	    tp++;
	 }
	 for (i = 0; i < 3; i++)
	 {
	    if (!isdigit(*tp))
	       goto BAD_OBJECT;
	    refcnt = (refcnt * 10) + (*tp - '0');
	    tp++;
	 }
	 if (refcnt || defcnt)
	 {
	    sprintf (view[0], "binloader: Object contains REF/DEF");
	    status = -23;
	    break;
	 }

	 reloprog = *op++ == '1' ? TRUE : FALSE;
	 wdata = getnumfield (op, FALSE);
	 if (wdata >= memlen)
	 {
	    sprintf (view[0], "binloader: Program too large for memory");
	    status = -21;
	    break;
	 }
      }

      /*
      ** Process loadblocks
      */

      else
      {
	 int i, j;

	 if (!strncmp((char *)op, "/*", 2))
	 {
	    status = EOF;
	    break;
	 }

	 if (!strncmp((char *)op, "&&&&", 4))
	 {
	    op += WORDLEN;
	    wdata = getnumfield (op, FALSE);
	    if (reloprog)
	       wdata += loadaddr;
	    SETPC (wdata);
	    run = TRUE;
	    continue;
	 }

         *curraddr = getnumfield (op, FALSE);
	 if (reloprog)
	    *curraddr += loadaddr;
	 op += WORDLEN;

	 for (i = 0; i < 2; i++)
	 {
	    reloword = getnumfield (op, FALSE);
	    op += WORDLEN;
	    for (j = 0; j < 8; j++)
	    {
	       int relo;

	       if (*op == ' ')
	       {
	          i = j = 9;
		  continue;
	       }
	       wdata = getnumfield (op, FALSE);
	       op += WORDLEN;
	       relo = (reloword >> (14 - (j*2))) & 3;
	       if (reloprog && (relo == 2))
	          wdata += loadaddr;
	       putmem (*curraddr, wdata);
	       *curraddr += 1;
	    }
	 }
      }
   }
   return (status);

BAD_OBJECT:
   sprintf (view[0], "binloader: Bad object file");
   return (-22);
}

/***********************************************************************
* loadrec - Load a tagged object record.
***********************************************************************/

int
loadrec (FILE *lfd, uint8 *inbuf, int loadaddr, int *curraddr)
{
   static int wordtaglen = WORDTAGLEN;
   uint8 *op = (uint8 *)inbuf;
   int i, j;
   int relbyte;
   int16 cksum;

   if (*op == EOFSYM)
   {
      wordtaglen = WORDTAGLEN;
      return (EOF);
   }

   cksum = 0;
   relbyte = FALSE;
   for (i = 0; i < 80; i++)
   {
      uint16 wdata;
      uint16 bit;
      uint8 otag;
      uint8 item[16];

      otag = *op++;
      if (otag == EOR_TAG)
      {
	 return (0);
      }
      cksum += otag;
      memcpy (item, op, 4);
      item[4] = '\0';
      if (otag != CKSUM_TAG)
	 for (j = 0; j < 4; j++) cksum += item[j];
      wdata = getnumfield (op, FALSE);

#ifdef DEBUGLOADER
      if (lfd != NULL)
      {
	 fprintf (lfd, "loadaddr = %04X, curraddr = %04X, otag = %c\n",
		  loadaddr, *curraddr,
		  otag == 0x01 ? '0' : otag);
	 fprintf (lfd, "data = %04X\n", wdata & 0xFFFF);
      }
#endif

      switch (otag)
      {
      case IDT_TAG:
	 wdata += loadaddr;
	 if (wdata >= memlen)
	 {
	    sprintf (view[0], "binloader: Program too large for memory");
	    return (-21);
	 }
	 op += wordtaglen-1;
	 for (j = 0; j < IDTLEN; j++) cksum += *op++;
	 break;

      case RELORG_TAG:
	 wdata += loadaddr;
      case ABSORG_TAG:
	 *curraddr = wdata;
	 op += wordtaglen-1;
	 break;

      case RELDATA_TAG:
	 wdata += loadaddr;
      case ABSDATA_TAG:
	 if (relbyte)
	 {
	    relbyte = FALSE;
	    bit = wdata & 0x0001;
	    wdata = (getmem (*curraddr - 1) << 14) | (wdata >> 1);
	    wdata += loadaddr;
	    putmem (*curraddr - 1, wdata >> 14);
	    wdata = ((wdata << 1) & 0x7FFE) | bit;
	 }
	 putmem (*curraddr, wdata);
	 *curraddr += 1;
	 op += wordtaglen-1;
	 break;

      case RELBYTE_TAG:
	 relbyte = TRUE;
	 putmem (*curraddr, wdata);
	 *curraddr += 1;
	 op += wordtaglen-1;
         break;

      case RELENTRY_TAG:
	 wdata += loadaddr;
      case ABSENTRY_TAG:
	 SETPC (wdata);
	 run = TRUE;
	 op += wordtaglen-1;
	 break;

      case CKSUM_TAG:
#ifdef DEBUGLOADER
         if (lfd != NULL)
	    fprintf (lfd, "cksum = %04X\n", cksum);
#endif
         cksum += (int16)wdata;
#ifdef DEBUGLOADER
         if (lfd != NULL)
	    fprintf (lfd, "cksum+wdata = %04X\n", cksum);
#endif
	 if (cksum != 0)
	 {
	    sprintf (view[0], "binloader: Checksum error");
	    return (-20);
	 }
	 op += wordtaglen-1;
         break;

      case RELEXTRN_TAG:
      case ABSEXTRN_TAG:
      case ABSGLOBAL_TAG:
      case RELGLOBAL_TAG:
	 sprintf (view[0], "binloader: Object contains REF/DEF");
	 return (-23);

      default: 
	 sprintf (view[0], "binloader: Illegal object tag: %c(%02X)",
	 	  otag, otag);
	 return (-22);
      }
   }
   return (0);
}

/***********************************************************************
* binloader - Load binary programs.
***********************************************************************/

int
binloader (char *file, int loadpt)
{
   FILE *fd;
   FILE *lfd = NULL;
#ifdef DUMPLOADREC
   int reclen;
#endif
   int status = 0;
   int i;
   int loadaddr = LOADADDRESS;
   int curraddr = LOADADDRESS;
   uint8 inbuf[82];

   while (isspace (*file)) file++;
#ifdef DEBUGLOADER
   lfd = fopen ("fileload.log", "w");
   fprintf (lfd, "binloader: file = '%s', loadpd = %d(%04X)\n",
	    file, loadpt, loadpt);
#endif

   if (loadpt > 0)
   {
      loadaddr = loadpt;
      curraddr = loadpt;
   }

   if ((fd = fopen (file, "r")) == NULL)
   {
      sprintf (view[0], "binloader: open failed: %s",
	       strerror (ERRNO));
      sprintf (view[1], "filename: %s", file);
      run = FALSE;
#ifdef DEBUGLOADER
      fprintf (lfd, "error: %s\n", view[0]);
      fprintf (lfd, "       %s\n", view[1]);
      fclose (lfd);
#endif
      return (-1);
   }

   i = fgetc (fd);
   ungetc (i, fd);

   if (i != IDT_TAG)
   {
      status = loadcardfile (fd, lfd, loadaddr, &curraddr);
      if (status == EOF)
         status = 0;
      if (status != 0)
      {
	 sprintf (view[1], "filename: %s", file);
#ifdef DEBUGLOADER
	 fprintf (lfd, "error: %s\n", view[0]);
	 fprintf (lfd, "       %s\n", view[1]);
#endif
      }
   }

   else
   {
      while (fgets ((char *)inbuf, 82, fd) != NULL)
      {
#ifdef DUMPLOADREC
	 reclen = strlen (inbuf);
	 fprintf (lfd, "Record:\n");
	 HEXDUMP (lfd, inbuf, reclen, 0);
#endif
	 if ((status = loadrec (lfd, inbuf, loadaddr, &curraddr)) != 0)
	 {
	    if (status == EOF)
	    {
	       status = 0;
	       break;
	    }
	    snprintf (view[1], MAXVIEWLEN, "filename: %s", file);

	    run = FALSE;
#ifdef DEBUGLOADER
	    fprintf (lfd, "error: %s\n", view[0]);
	    fprintf (lfd, "       %s\n", view[1]);
#endif
	    break;
	 }
      }
   }

#ifdef DUMPPROG
   dispaddr = loadaddr;
   for (i = loadaddr; i < curraddr; i += 8)
   {
      dumpsingle (inbuf);
      fprintf (stderr, "%s\n", inbuf);
   }
   run = FALSE;
#endif

#ifdef DEBUGLOADER
   fclose (lfd);
#endif
   fclose (fd);

   return (status);
}
