/***********************************************************************
*
* simops.c - Simulate instruction operations for the TI 980 computer.
*
* Changes:
*   06/24/11   DGP   Original.
*   08/31/11   DGP   Added PIF/MP support.
*   09/14/11   DGP   Added cpu models and Calder 980 support.
*   10/25/11   DGP   Added CDD Common Data Feature.
*   04/05/12   DGP   Correct TMBO, TABO, etc.
*   08/24/16   DGP   Fixed process_ixb. Fixed MOV regsister mode.
*
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>

#include "simdef.h"

extern uint16 statreg;	/* The program status register */
extern uint16 lights;	/* The panel lights */
extern uint32 breakaddr;/* The break point address */
extern uint32 watchaddr;/* The watch point address */
extern int pancount;	/* Instructions executed for panel */
extern unsigned long instcount;
extern unsigned long tracestart;

extern int run;
extern int idle;
extern int runled;
extern int idleled;
extern int idlenum;
extern int breakflag;
extern int watchflag;
extern int enablepanel;
extern int mapenabled;
extern int traceenable;
extern int tracelim[2];
extern int deferint;
extern int traceit;
extern int windowlen;
extern uint16 curinst;
extern uint32 memlen;
extern long delayclock;
extern char view[MAXVIEW][81];
extern FILE *tracefd;
extern int errindex;
extern int trcindex;
extern int traceline;
extern int model;
extern int showstatus;

extern uint16 ssw;
extern uint16 regs[16];
extern uint16 memory[MAXMEMSIZE];

extern int cddcmnenable;
extern uint16 cddcmnlimit;
extern uint32 cddcmnbank;
extern uint32 lowlimitreg;
extern uint32 upperlimitreg;
extern uint32 cddlowlimitext;
extern uint32 cddupperlimitext;
extern uint32 cpumemmap[2];

static void runinst (uint16);

static int proc1op (uint16 inst);
static int proc2op (uint16 inst);
static int proc3op (uint16 inst);
static int proc4op (uint16 inst);
static int proc5op (uint16 inst);
static int proc6op (uint16 inst);
static int proc7op (uint16 inst);
static int proc8op (uint16 inst);
static int proc9op (uint16 inst);
static int proc10op (uint16 inst);
static int proc11op (uint16 inst);

extern int trcidx;
extern int trcopr;
extern TraceInst trcbuf[TRCBUFMAX];

typedef struct {
   int (*proc) (uint16 inst);
} Inst_Proc;

Inst_Proc inst_proc[MAX_INST_TYPES] = {
   proc1op,  proc2op,  proc3op,  proc4op,  proc5op,  proc6op,  proc7op,
   proc8op,  proc9op,  proc10op, proc11op
};

static char outbuf[128];

/***********************************************************************
* decodeinst - Decode instructions.
***********************************************************************/

int
decodeinst (uint16 inst)
{
   uint16 op;

   if ((inst & 0xF800) < 0xC000)
   {
      op = (inst >> 11);
      switch (op)
      {
      case 0:  /* LDA */
      case 1:  /* LDE */
      case 2:  /* LDX */
      case 3:  /* LDM */
      case 4:  /* ADD */
      case 5:  /* SUB */
      case 6:  /* IOR */
      case 7:  /* AND */
      case 8:  /* BIX */
      case 9:  /* DMT */
      case 10: /* IMO */
      case 11: /* DIV */
      case 12: /* CPL */
      case 13: /* CPA */
      case 14: /* BRL */
      case 15: /* BRU */
      case 16: /* STA */
      case 17: /* STE */
      case 18: /* STX */
      case 19: /* MPY */
      case 20: /* DST */
      case 21: /* DSB */
      case 22: /* DLD */
      case 23: /* DAD */
	 return (TYPE_1);

      default: ;
      }
   }
   else if ((inst & 0xFF00) < 0xC800)
   {
      op = ((inst & 0x0F80) >> 7);
      switch (op)
      {
      case 0:  /* RSU */
      case 1:  /* RAD */
      case 2:  /* RCO */
      case 4:  /* RIV */
      case 5:  /* REO */
      case 6:  /* RIN */
      case 8:  /* RCA */
      case 9:  /* ROR */
      case 10: /* RMO */
      case 12: /* RCL */
      case 13: /* RAN */
      case 14: /* RDE */
      case 15: /* REX */
	 return (TYPE_2);

      default: ;
      }
   }
   else if ((inst & 0xFF00) < 0xCC00)
   {
      op = ((inst & 0x03F0) >> 5);
      switch (op)
      {
      case 0:  /* ARA */
      case 1:  /* ARD */
      case 2:  /* LRA */
      case 3:  /* LRD */
      case 4:  /* ALA */
      case 5:  /* ALD */
      case 6:  /* LLA */
      case 7:  /* LLD */
      case 8:  /* RTO */
      case 10: /* RTZ */
      case 12: /* LTO */
      case 14: /* LTZ */
      case 16: /* CRA */
      case 17: /* CRE */
      case 18: /* CRX */
      case 19: /* CRM */
      case 20: /* NRM */
      case 25: /* CRS */
      case 26: /* CRL */
      case 27: /* CRB */
      case 28: /* CLD */
      case 30: /* CRD */
	 return (TYPE_3);

      default: ;
      }
   }
   else if ((inst & 0xFF00) < 0xD000)
   {
      op =  ((inst & 0x3F0) >> 4);
      switch (op)
      {
      case 0:  /* SZE */
      case 1:  /* SSE */
      case 2:  /* SOO */
      case 4:  /* SOD */
      case 6:  /* SMI */
      case 8:  /* SNZ */
      case 9:  /* SSN */
      case 10: /* SNO */
      case 12: /* SEV */
      case 14: /* SPL */
      case 16: /* SLT */
      case 18: /* SEQ */
      case 20: /* SGT */
      case 22: /* SOV */
      case 24: /* SGE */
      case 26: /* SNE */
      case 28: /* SLE */
      case 30: /* SNV */
      case 32: /* IDL */
      case 54: /* SOC */
      case 62: /* SNC */
	 return (TYPE_4);

      default: ;
      }
   }
   else if ((inst & 0xFF00) < 0xD900)
   {
      op = ((inst & 0x00F0) >> 4);
      if ((inst & 0xD9A0) == 0xD800) /* RDS */
	 return (TYPE_5);
      else if ((inst & 0xD9A0) == 0xD820) /* WDS */
	 return (TYPE_5);
      else switch (op)
      {
      case 8:  /* LSB */
      case 12: /* SSB */
	 return (TYPE_5);

      case 9:  /* LSR */
      case 10: /* LRF */
      case 14: /* SRF */
	 if (model >= CPU_TI980A)
	    return (TYPE_5);

      default: ;
      }
   }
   else if ((inst & 0xFF00) == 0xD900)
   {
      if ((inst & 0xFF00) == 0xD900) /* ATI */
	 return (TYPE_7);
   }
   else if ((inst & 0xFF00) == 0xDB00)
   {
      if (model >= CPU_TI980A)
      {
	 op = ((inst & 0x00F0) >> 4);
	 switch (op)
	 {
	 case 0: /* TABZ */
	 case 1: /* TABO */
	 case 4: /* SABZ */
	 case 5: /* SABO */
	 case 2: /* TMBZ */
	 case 3: /* TMBO */
	 case 6: /* SMBZ */
	 case 7: /* SMBO */
	    return (TYPE_6);

	 default: ;
	 }
      }
   }
   else if ((inst & 0xFF00) == 0xDD00)
   {
      if (model >= CPU_TI980A)
      {
	 if ((inst & 0xFF00) == 0xDD00) /* API */
	    return (TYPE_7);
      }
   }
   else if ((inst & 0xFF00) == 0xDF00)
   {
      if (model >= CPU_TI980A)
      {
	 if ((inst & 0xFF80) == 0xDF00) /* MVC */
	    return (TYPE_8);
	 else if ((inst & 0xFF80) == 0xDF80) /* CLC */
	    return (TYPE_8);
      }
   }
   else if ((inst & 0xF000) == 0xE000)
   {
      if (model == CPU_CDD980)
	 return (TYPE_9);
   }
   else if ((inst & 0xFF00) == 0xF800)
   {
      if (model == CPU_CDD980)
      {
	 if (((inst & 0x0088) == 0x0088) || /* LAR */
	     ((inst & 0x0088) == 0x0080) || /* SAR */
	     ((inst & 0x0088) == 0x0008) || /* RFI */
	     ((inst & 0x0088) == 0x0000))   /* RFD */
	    return (TYPE_10);
      }
   }
   else if ((inst & 0xF000) == 0xF000)
   {
      if (model == CPU_CDD980)
      {
	 if (((inst & 0x0F00) == 0x0000) || /* SWTX */
	     ((inst & 0x0F00) == 0x0100) || /* SWT */
	     ((inst & 0x0F00) == 0x0200))   /* SWTS */
	    return (TYPE_11);
      }
   }
   return (TYPE_ILL);
}

/***********************************************************************
* mapaddr - Map address.
***********************************************************************/

uint32
mapaddr (uint32 ma, int mapsel, int forcemap)
{
   int chklimit;
   uint32 upplimit;

   chklimit = FALSE;
   upplimit = upperlimitreg;

   if (model == CPU_CDD980)
   {
      if (cddcmnenable && (IS_PIF || IS_MEMMAP))
      {
         if (ma >= cddcmnlimit)
	    ma += cddcmnbank;
      }
      else
      {
	 upplimit += cddupperlimitext;
	 if (!forcemap && IS_PIF && !IS_MEMMAP)
	 {
	    ma += (lowlimitreg + 1) & 0xFFFF;
	    ma += cddlowlimitext;
	    chklimit = TRUE;
	 }
	 else if (forcemap || (!IS_PIF && IS_MEMMAP))
	 {
	    if (ma >= 0x100)
	    {
	       ma += cpumemmap[mapsel];
	       if (!forcemap)
		  chklimit = TRUE;
	    }
	 }
	 else if (!forcemap && IS_PIF && IS_MEMMAP)
	 {
	    ma += (lowlimitreg + 1) & 0xFFFF;
	    ma += cddlowlimitext;
	    ma += cpumemmap[mapsel];
	    chklimit = TRUE;
	 }
      }
   }

   else
   {
      if (IS_PIF)
      {
	 ma += (lowlimitreg + 1) & 0xFFFF;
	 chklimit = TRUE;
      }
   }

   if (chklimit && (ma > upplimit))
   {
      geninterrupt (INTINTLOC, PROT_BIT);
      ma = 0xFFFF;
   }
   return (ma);
}

/***********************************************************************
* process_ixb - Process address using IXB modes.
***********************************************************************/

uint32
process_ixb (uint16 ixb, uint16 disp, int inc)
{
   uint32 ma;

   switch (ixb)
   {
   case 0: /* PC relative */

      disp += inc;
      ma = (int16)GETPC + (int8)disp;
      ma &= 0xFFFF;
      break;

   case 1: /* Base register relative */

      disp += inc;
      ma = regs[6] + disp;
      break;

   case 2: /* Indexed PC relative */

      disp += inc;
      ma = (int16)(GETPC + regs[2]) + (int8)disp;
      ma &= 0xFFFF;
      break;

   case 3: /* Indexed base register relative */

      disp += inc;
      ma = regs[6] + regs[2] + disp;
      break;

   case 4: /* Indirect PC relative */

      ma = getmem ((int16)GETPC + (int8)disp);
      ma &= 0xFFFF;
      if ((inc == 0) && (disp == 0))
      {
	 trcbuf[trcidx].ops[0] = ma;
	 trcopr = 1;
      }
      ma += inc;
      break;

   case 5: /* Indirect base register relative */

      ma = getmem (regs[6] + disp);
      if ((inc == 0) && (disp == 0))
      {
	 trcbuf[trcidx].ops[0] = ma;
	 trcopr = 1;
      }
      ma += inc;
      break;

   case 6: /* Indirect, indexed, PC relative */

      if (IS_PRENDX)
         ma = getmem ((int16)(GETPC + regs[2]) + (int8)disp);
      else
         ma = getmem (((int16)GETPC + (int8)disp) + regs[2]);
      ma &= 0xFFFF;
      if ((inc == 0) && (disp == 0))
      {
	 trcbuf[trcidx].ops[0] = ma;
	 trcopr = 1;
      }
      ma += inc;
      break;

   case 7: /* Immediate */

      ma = disp;
      ma += inc;
      break;
   }

   ma = mapaddr (ma, 0, FALSE);

   if (watchflag && (ma == watchaddr))
   {
      run = FALSE;
      runled = FALSE;
      showstatus = TRUE;
   }
   return (ma);
}

/***********************************************************************
* getinst - Get an Instruction.
***********************************************************************/

uint16
getinst (int bkpt)
{
   uint32 ma;

   ma = GETPC;
   ma = mapaddr (ma, 0, FALSE);

   if (ma < memlen)
      return (memory[ma]);
   return (0xFFFF);
}

/***********************************************************************
* getmem - Get a word from memory.
***********************************************************************/

uint16
getmem (uint16 pa)
{
   uint32 ma;

   ma = pa;
   ma = mapaddr (ma, 0, FALSE);

   if (watchflag && (ma == watchaddr))
   {
      run = FALSE;
      runled = FALSE;
      showstatus = TRUE;
   }
   if (ma < memlen)
      return (memory[ma]);
   return (0xFFFF);
}

/***********************************************************************
* putmem - Put a word in memory.
***********************************************************************/

void
putmem (uint16 pa, uint16 v)
{
   uint32 ma;

   ma = pa;
   ma = mapaddr (ma, 0, FALSE);

   if (watchflag && (ma == watchaddr))
   {
      run = FALSE;
      runled = FALSE;
      showstatus = TRUE;
   }
   if (ma < memlen)
      memory[ma] = v;
}

/***********************************************************************
* getword - Get a word from memory.
***********************************************************************/

uint16
getword (uint16 ixb, uint16 disp, int inc)
{
   uint32 ma;

   ma = process_ixb (ixb, disp, inc);
   if (ma < memlen)
      return (memory[ma]);
   return (0xFFFF);
}

/***********************************************************************
* putword - Put a word in memory.
***********************************************************************/

void
putword (uint16 ixb, uint16 disp, uint16 v, int inc)
{
   uint32 ma;

   ma = process_ixb (ixb, disp, inc);
   if (ma < memlen)
      memory[ma] = v;
}

/***********************************************************************
* getbyte - Get a byte from memory.
***********************************************************************/

uint8
getbyte (uint16 pa, uint16 bno)
{
   uint32 ma;
   uint16 word = 0xFFFF;

   ma = pa;
   ma = mapaddr (ma, 0, FALSE);

   if (watchflag && (ma == watchaddr))
   {
      run = FALSE;
      runled = FALSE;
      showstatus = TRUE;
   }
   if (ma < memlen)
      word = memory[ma];
   return (bno ? (word & 0x00FF) : (word >> 8));
}

/***********************************************************************
* putbyte - Put a byte in memory.
***********************************************************************/

void
putbyte (uint16 pa, uint16 bno, uint8 v)
{
   uint32 ma;
   uint16 word;

   ma = pa;
   ma = mapaddr (ma, 0, FALSE);

   if (watchflag && (ma == watchaddr))
   {
      run = FALSE;
      runled = FALSE;
      showstatus = TRUE;
   }
   word = memory[ma];
   if (bno)
      word = (word & 0xFF00) | v;
   else
      word = (word & 0x00FF) | (v << 8);
   if (ma < memlen)
      memory[ma] = word;
}

/***********************************************************************
* getaddr - Get an address.
***********************************************************************/

uint16
getaddr (uint16 ixb, uint16 disp)
{
   uint32 ma;

   ma = process_ixb (ixb, disp, 0);
   return (ma);
}

/***********************************************************************
* getoperand - Get next inst.
***********************************************************************/

uint16
getoperand (void)
{
   uint16 addr;

   addr = getinst (TRUE);
   INCPC (1);
   trcbuf[trcidx].ops[0] = addr;
   trcopr = 1;
   return (addr);
}

/***********************************************************************
* chkaddword - Check addition for carry and overflow
***********************************************************************/

static void
chkaddword (int32 acc, int16 sv1, int16 sv2)
{
   CLR_CARRY;
   CLR_OVER;

   if ((acc & 0xFFFF) < (sv1 & 0xFFFF)) SET_CARRY;

   if (((sv1 & 0x8000) == (sv2 & 0x8000)) &&
       ((sv2 & 0x8000) != (acc & 0x8000))) SET_OVER;
}

/***********************************************************************
* chksubword - Check subtraction for carry and overflow
***********************************************************************/

static void
chksubword (int32 acc, int16 sv1, int16 sv2)
{
   CLR_CARRY;
   CLR_OVER;

   if ((uint16)sv2 >= (uint16)sv1) SET_CARRY;

   if (((sv1 & 0x8000) != (sv2 & 0x8000)) &&
       ((sv2 & 0x8000) != (acc & 0x8000))) SET_OVER;
}

/***********************************************************************
* cmpword - Compare word values.
***********************************************************************/

void
cmpword (uint16 val1, uint16 val2, int logical)
{
   CLR_CMP;
   if (logical)
   {
      if (val1 == val2) SET_EQ;
      else if (val1 > val2) SET_GT;
      else SET_LT;
   }
   else
   {
      if (val1 == val2) SET_EQ;
      else if ((int16)val1 > (int16)val2) SET_GT;
      else SET_LT;
   }
}

/***********************************************************************
* proc1op - Process type 1 operations.
***********************************************************************/

static int
proc1op (uint16 inst)
{
   int incpc;
   int32 acc;
   uint16 val1, val2;
   int16 sval1, sval2;
   uint16 ixb = (inst & 0x0700) >> 8;
   uint16 disp = (inst & 0x00FF);

   incpc = 0;
   if (ixb == 7)
      val1 = (int16)((int8)disp);
   else
      val1 = getword (ixb, disp, 0);

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", inst >> 11, inst >> 11);
#endif

   switch (inst >> 11)
   {
   case 0: /* LDA */

      regs[0] = val1;
      break;

   case 1: /* LDE */

      regs[1] = val1;
      break;

   case 2: /* LDX */

      regs[2] = val1;
      break;

   case 3: /* LDM */

      regs[3] = val1;
      break;

   case 4: /* ADD */

      sval1 = (int16)regs[0];
      sval2 = (int16)val1;
      acc = sval1 + sval2;
      chkaddword (acc, sval1, sval2);
      regs[0] = acc & 0xFFFF;
      break;

   case 5: /* SUB */

      sval1 = (int16)regs[0];
      sval2 = (int16)val1;
      acc = sval1 - sval2;
      chksubword (acc, sval1, sval2);
      regs[0] = acc & 0xFFFF;
      break;

   case 6: /* IOR */

      regs[0] |= val1;
      break;

   case 7: /* AND */

      regs[0] &= val1;
      break;

   case 8: /* BIX */

      regs[2]++;
      if (regs[2])
      {
	 SETPC (getaddr (ixb, disp));
	 return (0);
      }
      break;

   case 9: /* DMT */

      val1 = getaddr (ixb, disp);
      val2 = getmem (val1) - 1;
      putmem (val1, val2);
      if (val2 == 0)
         INCPC (1);
      break;

   case 10: /* IMO */

      val1 = getaddr (ixb, disp);
      putmem (val1, getmem (val1) + 1);
      break;

   case 11: /* DIV */

      acc = ((int32)regs[0] << 15) | (regs[1] & 0x7FFF);
      if (val1 <= regs[0])
         SET_OVER;
      else
      {
	 CLR_OVER;
	 regs[0] = acc / val1;
	 regs[1] = acc % val1;
      }
      break;

   case 12: /* CPL */

      cmpword (regs[0], val1, TRUE);
      break;

   case 13: /* CPA */

      cmpword (regs[0], val1, FALSE);
      break;

   case 14: /* BRL */

      regs[5] = GETPC;
      if ((ixb & 4) && (disp == 0))
         regs[5]++;
      /* Fall through */

   case 15: /* BRU */

      val1 = getaddr (ixb, disp);
      SETPC (val1);
      return (0);

   case 16: /* STA */

      putword (ixb, disp, regs[0], 0);
      break;

   case 17: /* STE */

      putword (ixb, disp, regs[1], 0);
      break;

   case 18: /* STX */

      putword (ixb, disp, regs[2], 0);
      break;

   case 19: /* MPY */

      sval1 = (int16)regs[0];
      sval2 = (int16)val1;
      acc = sval1 * sval2;
      regs[0] = (acc >> 15) & 0xFFFF;
      regs[1] = acc & 0x7FFF;
      regs[1] |= regs[0] & 0x8000;
      break;

   case 20: /* DST */

      putword (ixb, disp, regs[0], 0);
      putword (ixb, disp, regs[1], 1);
      break;

   case 21: /* DSB */

      if (ixb == 7)
      {
         val2 = val1;
	 if (val2 & 0x8000)
	    val1 = 0xFFFF;
	 else
	    val1 = 0;
      }
      else
	 val2 = getword (ixb, disp, 1);

      acc = ((int32)regs[0] << 15) | (regs[1] & 0x7FFF);
      acc = acc - (((int32)val1 << 15) | (val2 & 0x7FFF));
      regs[0] = acc >> 15;
      regs[1] = acc & 0x7FFF;
      if ((ixb == 0) && (disp == 0)) incpc++;
      break;

   case 22: /* DLD */

      if (ixb == 7)
      {
         val2 = val1;
	 if (val2 & 0x8000)
	    val1 = 0xFFFF;
	 else
	    val1 = 0;
      }
      else
	 val2 = getword (ixb, disp, 1);

      regs[0] = val1;
      regs[1] = val2;
      if ((ixb == 0) && (disp == 0)) incpc++;
      break;

   case 23: /* DAD */

      if (ixb == 7)
      {
         val2 = val1;
	 if (val2 & 0x8000)
	    val1 = 0xFFFF;
	 else
	    val1 = 0;
      }
      else
	 val2 = getword (ixb, disp, 1);

      acc = ((int32)regs[0] << 15) | (regs[1] & 0x7FFF);
      acc = acc + (((int32)val1 << 15) | (val2 & 0x7FFF));
      regs[0] = acc >> 15;
      regs[1] = acc & 0x7FFF;
      if ((ixb == 0) && (disp == 0)) incpc++;
      break;

   default:
      return (-1);
   }

   if ((disp == 0) && ((ixb == 0) || (ixb == 4) || (ixb == 6))) incpc++;
   if ((ixb == 0) && (disp == 0))
   {
      trcbuf[trcidx].ops[0] = val1;
      trcopr = 1;
   }
   INCPC (incpc);

   return (0);
}

/***********************************************************************
* proc2op - Process type 2 operations.
***********************************************************************/

static int
proc2op (uint16 inst)
{
   int32 acc;
   uint16 val1;
   int16 sval1, sval2;
   uint16 sr, dr;

   dr = inst & 0x0007;
   sr = (inst >> 4) & 0x0007;

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", (inst & 0x0F80) >> 7, (inst & 0x0F80) >> 7);
#endif

   switch ((inst & 0x0F80) >> 7)
   {
   case 0: /* RSU */

      sval1 = (int16)regs[sr];
      sval2 = (int16)regs[dr];
      acc = sval2 - sval1;
      chksubword (acc, sval1, sval2);
      regs[dr] = acc & 0xFFFF;
      break;

   case 1: /* RAD */

      sval1 = (int16)regs[sr];
      sval2 = (int16)regs[dr];
      acc = sval2 + sval1;
      chkaddword (acc, sval1, sval2);
      regs[dr] = acc & 0xFFFF;
      break;

   case 2: /* RCO */

      if (regs[sr] == 0xFFFF)
         SET_OVER;
      else
         CLR_OVER;
      regs[dr] = ~regs[sr] + 1;
      break;

   case 4: /* RIV */

      regs[dr] = ~regs[sr];
      break;

   case 5: /* REO */

      regs[dr] = regs[dr] ^ regs[sr];
      break;

   case 6: /* RIN */

      regs[dr] = regs[sr] + 1;
      break;

   case 8: /* RCA */

      cmpword (regs[sr], regs[dr], FALSE);
      break;

   case 9: /* ROR */

      regs[dr] = regs[dr] | regs[sr];
      break;

   case 10: /* RMO */

      regs[dr] = regs[sr];
      break;

   case 12: /* RCL */

      cmpword (regs[sr], regs[dr], TRUE);
      break;

   case 13: /* RAN */

      regs[dr] = regs[dr] & regs[sr];
      break;

   case 14: /* RDE */

      regs[dr] = regs[sr] - 1;
      break;

   case 15: /* REX */

      val1 = regs[dr];
      regs[dr] = regs[sr];
      regs[sr] = val1;
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* circshift - Circular shift
***********************************************************************/

static void
circshift  (int reg, int count)
{
   int bit;

   for (; count; count--)
   {
      bit = regs[reg] & 0x0001;
      regs[reg] >>= 1;
      regs[reg] |= bit ? 0x8000 : 0;
   }
}

/***********************************************************************
* proc3op - Process type 3 operations.
***********************************************************************/

static int
proc3op (uint16 inst)
{
   uint32 acc;
   int count;
   int bit0, bit1;

   count = inst & 0x001F;

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", (inst & 0x03F0) >> 5, (inst & 0x03F0) >> 5);
#endif

   switch ((inst & 0x03F0) >> 5)
   {
   case 0: /* ARA */

      regs[0] = (int16)regs[0] >> count;
      break;

   case 1: /* ARD */

      acc = ((uint32)regs[0] << 16) | (((uint32)regs[1] & 0x7FFF) << 1);
      acc = (int32)acc >> count;
      regs[0] = (acc >> 16) & 0xFFFF;
      regs[1] = (acc & 0xFFFE) >> 1;
      regs[1] |= regs[0] & 0x8000;
      break;

   case 2: /* LRA */

      regs[0] = regs[0] >> count;
      break;

   case 3: /* LRD */

      acc = ((uint32)regs[0] << 16) | (uint32)regs[1];
      acc >>= count;
      regs[0] = (acc >> 16) & 0xFFFF;
      regs[1] = acc & 0xFFFF;
      break;

   case 4: /* ALA */

      CLR_OVER;
      bit0 = regs[0] & 0x8000;
      acc = regs[0];
      for (;count; count--)
      {
	 acc <<= 1;
	 if (bit0 != (acc & 0x8000))
	    SET_OVER;
      }
      regs[0] = bit0 | (acc & 0x7FFF);
      break;

   case 5: /* ALD */

      CLR_OVER;
      bit0 = (regs[0] & 0x8000) ? TRUE : FALSE;
      acc = ((uint32)regs[0] & 0x7FFF) << 15 | ((uint32)regs[1] & 0x7FFF);
      for (; count; count--)
      {
	 bit1 = (acc & 0x20000000) ? TRUE : FALSE;
	 if (bit0 && !bit1)
	    SET_OVER;
	 acc <<= 1;
      }
      regs[0] = (regs[0] & 0x8000) | ((acc >> 15) & 0x7FFF);
      regs[1] = (regs[0] & 0x8000) | (acc & 0x7FFF);
      break;

   case 6: /* LLA */

      regs[0] = regs[0] << count;
      break;

   case 7: /* LLD */

      acc = (uint32)regs[0] << 16 | (uint32)regs[1];
      acc <<= count;
      regs[0] = (acc >> 16) & 0xFFFF;
      regs[1] = acc & 0xFFFF;
      break;

   case 8: /* RTO */

      regs[2] = 0;
      for (; count; count--)
      {
	 if (regs[0] & 0x0001)
	 {
	    regs[0] &= 0xFFFE;
	    break;
	 }
         regs[0] >>= 1;
	 regs[2]++;
      }
      break;

   case 10: /* RTZ */

      regs[2] = 0;
      for (; count; count--)
      {
	 if (!(regs[0] & 0x0001))
	 {
	    regs[0] |= 0x0001;
	    break;
	 }
         regs[0] >>= 1;
	 regs[2]++;
      }
      break;

   case 12: /* LTO */

      regs[2] = 0;
      for (; count; count--)
      {
	 if (regs[0] & 0x8000)
	 {
	    regs[0] &= 0x7FFF;
	    break;
	 }
         regs[0] <<= 1;
	 regs[2]++;
      }
      break;

   case 14: /* LTZ */

      regs[2] = 0;
      for (; count; count--)
      {
	 if (!(regs[0] & 0x8000))
	 {
	    regs[0] |= 0x8000;
	    break;
	 }
         regs[0] <<= 1;
	 regs[2]++;
      }
      break;

   case 16: /* CRA */

      circshift (0, count);
      break;

   case 17: /* CRE */

      circshift (1, count);
      break;

   case 18: /* CRX */

      circshift (2, count);
      break;

   case 19: /* CRM */

      circshift (3, count);
      break;

   case 20: /* NRM */

      regs[2] = 0;
      for (; count; count--)
      {
	 bit0 = (regs[0] & 0x8000) ? TRUE : FALSE;
	 bit1 = (regs[0] & 0x4000) ? TRUE : FALSE;
	 if (bit0 != bit1)
	    break;
         regs[0] <<= 1;
	 regs[0] |= (regs[1] & 0x4000) ? 1 : 0;
	 regs[1] <<= 1;
	 regs[2]++;
      }
      regs[1] = (regs[0] & 0x8000) | (regs[1] & 0x7FFF);
      break;

   case 25: /* CRS */

      circshift (4, count);
      break;

   case 26: /* CRL */

      circshift (5, count);
      break;

   case 27: /* CRB */

      circshift (6, count);
      break;

   case 28: /* CLD */

      for (; count; count--)
      {
         bit0 = regs[0] & 0x8000;
	 regs[0] <<= 1;
	 regs[0] |= regs[1] & 0x8000 ? 0x0001 : 0;
	 regs[1] <<= 1;
	 regs[1] |= bit0 ? 0x0001 : 0;
      }
      break;

   case 30: /* CRD */

      for (; count; count--)
      {
         bit0 = regs[0] & 0x0001;
	 regs[0] >>= 1;
	 regs[0] |= regs[1] & 0x0001 ? 0x8000 : 0;
	 regs[1] >>= 1;
	 regs[1] |= bit0 ? 0x8000 : 0;
      }
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc4op - Process type 4 operations.
***********************************************************************/

static int
proc4op (uint16 inst)
{

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", (inst & 0x03F0) >> 4, (inst & 0x03F0) >> 4);
#endif

   switch ((inst & 0x03F0) >> 4)
   {
   case 0: /* SZE */

      if (regs[inst & 7] == 0x0000)
         INCPC (1);
      break;

   case 1: /* SSE */

      if (ssw & (inst & 0x000F))
         INCPC (1);
      break;

   case 2: /* SOO */

      if (regs[inst & 7] == 0xFFFF)
         INCPC (1);
      break;

   case 4: /* SOD */

      if (regs[inst & 7] & 0x0001)
         INCPC (1);
      break;

   case 6: /* SMI */

      if (regs[inst & 7] & 0x8000)
         INCPC (1);
      break;

   case 8: /* SNZ */

      if (regs[inst & 7] != 0x0000)
         INCPC (1);
      break;

   case 9: /* SSN */

      if (!(ssw & (inst & 0x000F)))
         INCPC (1);
      break;

   case 10: /* SNO */

      if (!(regs[inst & 7] == 0xFFFF))
         INCPC (1);
      break;

   case 12: /* SEV */

      if (!(regs[inst & 7] & 0x0001))
         INCPC (1);
      break;

   case 14: /* SPL */

      if (!(regs[inst & 7] & 0x8000))
         INCPC (1);
      break;

   case 16: /* SLT */

      if (IS_LT)
         INCPC (1);
      break;

   case 18: /* SEQ */

      if (IS_EQ)
         INCPC (1);
      break;

   case 20: /* SGT */

      if (IS_GT)
         INCPC (1);
      break;

   case 22: /* SOV */

      if (IS_OVER)
         INCPC (1);
      break;

   case 24: /* SGE */

      if (!IS_LT)
         INCPC (1);
      break;

   case 26: /* SNE */

      if (!IS_EQ)
         INCPC (1);
      break;

   case 28: /* SLE */

      if (!IS_GT)
         INCPC (1);
      break;

   case 30: /* SNV */

      if (!IS_OVER)
         INCPC (1);
      break;

   case 32: /* IDL */

      if (IS_PIF)
      {
	 geninterrupt (INTINTLOC, PIFVIO_BIT);
	 return (0);
      }
      idle = TRUE;
      idleled = TRUE;
      idlenum = inst & 0x000F;
      if (!IS_INTENAB)
         run = FALSE;
      instcount--;
      if (run)
	 DECPC (1);
      break;

   case 54: /* SOC */

      if (IS_CARRY)
         INCPC (1);
      break;

   case 62: /* SNC */

      if (!IS_CARRY)
         INCPC (1);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc5op - Process type 4 operations.
***********************************************************************/

static int
proc5op (uint16 inst)
{
   int i;
   uint16 addr;
   uint16 dev;

   addr = getoperand();

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", (inst & 0x00F0) >> 4, (inst & 0x00F0) >> 4);
#endif

   if ((inst & 0xD9A0) == 0xD800) /* RDS */
   {
      if (IS_PIF)
      {
	 geninterrupt (INTINTLOC, PIFVIO_BIT);
	 return (0);
      }
      dev = inst & 0x065F;
      doRDS (dev, addr);
   }
   else if ((inst & 0xD9A0) == 0xD820) /* WDS */
   {
      if (IS_PIF)
      {
	 geninterrupt (INTINTLOC, PIFVIO_BIT);
	 return (0);
      }
      dev = inst & 0x065F;
      doWDS (dev, addr);
   }
   else switch ((inst & 0x00F0) >> 4)
   {
   case 9: /* LSR */
      if (model < CPU_TI980A)
         return (-1);
      /* Fall through */

   case 8: /* LSB */
      if (IS_PIF)
      {
	 geninterrupt (INTINTLOC, PIFVIO_BIT);
	 return (0);
      }
      SETPC (getmem (addr));
      statreg = getmem (addr + 1) & 0xF9FC;
      deferint++;
      break;

   case 10: /* LRF */
      if (model < CPU_TI980A)
         return (-1);

      for (i = 0; i < 7; i++)
         regs[i] = getmem (addr + i);
      break;

   case 12: /* SSB */
      if (IS_PIF)
      {
	 geninterrupt (INTINTLOC, PIFVIO_BIT);
	 return (0);
      }
      putmem (addr, GETPC);
      putmem (addr + 1, statreg);
      SETPC (getmem (addr + 2));
      break;

   case 14: /* SRF */
      if (model < CPU_TI980A)
         return (-1);

      for (i = 0; i < 7; i++)
         putmem (addr + i, regs[i]);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc6op - Process type 6 operations.
***********************************************************************/

static int
proc6op (uint16 inst)
{
   uint16 addr;
   uint16 mask;

   if (model < CPU_TI980A)
      return (-1);

   mask = 1 << (15 - (inst & 0x000F));

#ifdef DEBUGINSTR
   printf (" switch %d(%x)", (inst & 0x00F0) >> 4, (inst & 0x00F0) >> 4);
#endif
   switch ((inst & 0x00F0) >> 4)
   {
   case 0: /* TABZ */
      if (!(regs[0] & mask))
         INCPC (1);
      break;

   case 1: /* TABO */
      if (regs[0] & mask)
         INCPC (1);
      break;

   case 2: /* TMBZ */
      addr = getoperand();
      if (!(getmem(addr) & mask))
         INCPC (1);
      break;

   case 3: /* TMBO */
      addr = getoperand();
      if (getmem(addr) & mask)
         INCPC (1);
      break;

   case 4: /* SABZ */
      regs[0] &= ~mask;
      break;

   case 5: /* SABO */
      regs[0] |= mask;
      break;

   case 6: /* SMBZ */
      addr = getoperand();
      putmem (addr, getmem(addr) & ~mask);
      break;

   case 7: /* SMBO */
      addr = getoperand();
      putmem (addr, getmem(addr) | mask);
      break;

   default:
      return (-1);
   }
   return (0);
}

/***********************************************************************
* proc7op - Process type 7 operations.
***********************************************************************/

static int
proc7op (uint16 inst)
{
   uint16 addr;

   addr = getoperand();

   if ((inst & 0xFF00) == 0xD900) /* ATI */
   {
      if (IS_PIF)
      {
	 geninterrupt (INTINTLOC, PIFVIO_BIT);
	 return (0);
      }
      doATI (inst & 0xFF, addr);
   }
   else if ((inst & 0xFF00) == 0xDD00) /* API */
   {
      if (model < CPU_TI980A)
         return (-1);
   }
   else 
      return (-1);

   return (0);
}

/***********************************************************************
* proc8op - Process type 8 operations.
***********************************************************************/

static int
proc8op (uint16 inst)
{
   uint16 saddr, daddr;
   uint16 bc;
   uint16 sbit, dbit;

   sbit = regs[1] & 0x0001;
   saddr = ((regs[0] & 0x0003) << 14) | (regs[1] >> 1);

   dbit = regs[4] & 0x0001;
   daddr = ((regs[3] & 0x0003) << 14) | (regs[4] >> 1);
   
   bc = regs[2];

   if ((inst & 0xFF80) == 0xDF00) /* MVC */
   {
      if (model < CPU_TI980A)
         return (-1);

      for (;bc; bc--)
      {
         putbyte (daddr, dbit, getbyte (saddr, sbit));
	 sbit ^= 1;
	 dbit ^= 1;
	 if (!sbit) saddr++;
	 if (!dbit) daddr++;
      }
   }
   else if ((inst & 0xFF80) == 0xDF80) /* CLC */
   {
      uint8 sbyte, dbyte;

      if (model < CPU_TI980A)
         return (-1);

      SET_EQ;
      for (;bc; bc--)
      {
         sbyte = getbyte (saddr, sbit);
         dbyte = getbyte (daddr, dbit);
	 sbit ^= 1;
	 dbit ^= 1;
	 if (!sbit) saddr++;
	 if (!dbit) daddr++;
	 if (sbyte < dbyte)
	 {
	    SET_LT;
	    break;
	 }
	 else if (sbyte > dbyte)
	 {
	    SET_GT;
	    break;
	 }
      }
      regs[0] = saddr >> 14;
      regs[1] = sbit | ((saddr << 1) & 0x7FFE);
      regs[2] = bc;
      regs[3] = daddr >> 14;
      regs[4] = dbit | ((daddr << 1) & 0x7FFE);
   }
   else 
      return (-1);

   return (0);
}

/***********************************************************************
* proc9op - Process type 9 operations. MOV inst.
***********************************************************************/

static int
proc9op (uint16 inst)
{
   int smap, dmap;
   uint32 ma;
   uint16 val;
   uint16 sr, dr;
   uint16 smode, dmode;

   if (model < CPU_CDD980)
      return (-1);

   dr = inst & 0x000F;
   dmode = (inst >> 8) & 0x0003; 
   dmap = (dr >> 3) & 1;
   if (dmode < 3) dr |= 0x8;

   sr = (inst >> 4) & 0x000F;
   smode = (inst >> 10) & 0x0003;
   smap = (sr >> 3) & 1;
   if (smode < 3) sr |= 0x8;

   switch (smode)
   {
   case 0:
      ma = regs[sr];
      ma = mapaddr (ma, smap, TRUE);
      val = memory[ma];
      break;
   case 1:
      ma = regs[sr];
      ma = mapaddr (ma, smap, TRUE);
      val = memory[ma];
      regs[sr]++;
      break;
   case 2:
      regs[sr]--;
      ma = regs[sr];
      ma = mapaddr (ma, smap, TRUE);
      val = memory[ma];
      break;
   case 3:
      val = regs[sr];
   }

   switch (dmode)
   {
   case 0:
      ma = regs[dr];
      ma = mapaddr (ma, dmap, TRUE);
      memory[ma] = val;
      break;
   case 1:
      ma = regs[dr];
      ma = mapaddr (ma, dmap, TRUE);
      memory[ma] = val;
      regs[dr]++;
      break;
   case 2:
      regs[dr]--;
      ma = regs[dr];
      ma = mapaddr (ma, dmap, TRUE);
      memory[ma] = val;
      break;
   case 3:
      regs[dr] = val;
   }

   return (0);
}

/***********************************************************************
* proc10op - Process type 10 operations.
***********************************************************************/

static int
proc10op (uint16 inst)
{
   int i;
   uint16 sr, dr;
   uint16 addr;

   if (model < CPU_CDD980)
      return (-1);

   sr = ((inst >> 4) & 7) | 0x08;
   dr = (inst & 7) | 0x08;

   if ((inst & 0x0088) == 0x0088) /* LAR */
   {
      addr = getoperand();
      for (i = 0; i < 8; i++)
      {
         regs[i|0x8] = getmem (addr);
	 addr++;
      }
      for (i = 0; i < 7; i++)
      {
         regs[i] = getmem (addr);
	 addr++;
      }
   }
   else if ((inst & 0x0088) == 0x0080) /* SAR */
   {
      addr = getoperand();
      for (i = 0; i < 8; i++)
      {
         putmem (addr, regs[i|0x8]);
	 addr++;
      }
      for (i = 0; i < 7; i++)
      {
         putmem (addr, regs[i]);
	 addr++;
      }
   }
   else if ((inst & 0x0088) == 0x0008) /* RFI */
   {
      regs[dr] = regs[sr] + 1;
   }
   else if ((inst & 0x0088) == 0x0000) /* RFD */
   {
      regs[dr] = regs[sr] - 1;
   }
   else
      return (-1);

   return (0);
}

/***********************************************************************
* proc11op - Process type 11 operations.
***********************************************************************/

static int
proc11op (uint16 inst)
{
   uint16 val;
   uint16 curinst;
   uint16 opc, npc;
   uint16 ost;

   if (model < CPU_CDD980)
      return (-1);

   val = inst & 0x00FF;

   switch ((inst & 0x0F00) >> 8)
   {
   case 0: /* SWTX */
      regs[2] = val;
      break;

   case 1: /* SWT */
      break;

   case 2: /* SWTS */
      regs[4] = val;
      break;

   default:
      return (-1);
   }

   npc = 0;
   curinst = memory[npc];

   /*
   ** If trap location is not set or illegal, stop execution.
   */

   if (!curinst || (decodeinst (curinst) == TYPE_ILL))
   {
      run = FALSE;
      return (0);
   }

   ost = statreg;
   opc = GETPC;
   deferint = 2;

   /*
   ** If the trap inst is an SSB, do it here.
   */

   if (curinst == SSBINST)
   {
      statreg &= 0xF034;
      npc = memory[npc+1];
      memory[npc] = opc;
      memory[npc+1] = ost;
      SETPC (memory[npc+2]);
   }

   else
   {
      SETPC (opc);
      statreg = ost;
   }

   return (0);
}

/***********************************************************************
* runinst - Run an instruction.
***********************************************************************/

static void 
runinst (uint16 inst)
{
   int error;
   int type;
   int i;

   if (traceenable && GETPC >= tracelim[0] && GETPC <= tracelim[1] &&
      instcount >= tracestart) 
      traceit = TRUE;
   else
      traceit = FALSE;

   if (!IS_IDLE (inst)) idleled = FALSE;
   curinst = inst;
   type = decodeinst (inst);

#if defined(DEBUGINSTR)
   if (enablepanel)
   {
      screenposition (windowlen,1);
      clearline ();
   }
   printf ("PC = %04X INST = %04X TYPE = %d", GETPC-1, inst, type+1);
#endif

   error = -1;

   if (trcidx >= TRCBUFMAX) trcidx = 0;
   trcbuf[trcidx].pc = GETPC - 1;
   trcbuf[trcidx].st = statreg;
   trcbuf[trcidx].inst = inst;
   for (i = 0; i < 16; i++)
   {
      trcbuf[trcidx].regs[i] = regs[i];
   }
   trcbuf[trcidx].ops[0] = trcbuf[trcidx].ops[1] = -1;
   trcopr = 0;

   if (type != TYPE_ILL)
      error = inst_proc[type].proc (inst);

#if defined(DEBUGINSTR)
   if (!enablepanel) printf ("\r\n");
#endif
   if (traceit)
   {
      fprintf (tracefd, "\n");
      fprintf (tracefd, " A %04X  E %04X  X %04X  M %04X\n",
      		trcbuf[trcidx].regs[0], trcbuf[trcidx].regs[1],
		trcbuf[trcidx].regs[2], trcbuf[trcidx].regs[3]);
      fprintf (tracefd, " S %04X  L %04X  B %04X  P %04X\n",
      		trcbuf[trcidx].regs[4], trcbuf[trcidx].regs[5],
		trcbuf[trcidx].regs[6], trcbuf[trcidx].regs[7]);
      if (model == CPU_CDD980)
      {
	 fprintf (tracefd, " 8 %04X  9 %04X 10 %04X 11 %04X\n",
		   trcbuf[trcidx].regs[8], trcbuf[trcidx].regs[9],
		   trcbuf[trcidx].regs[10], trcbuf[trcidx].regs[11]);
	 fprintf (tracefd, "12 %04X 13 %04X 14 %04X 15 %04X\n",
		   trcbuf[trcidx].regs[12], trcbuf[trcidx].regs[13],
		   trcbuf[trcidx].regs[14], trcbuf[trcidx].regs[15]);
      }
      if (traceline >= 60)
      {
         traceline = 0;
	 fprintf (tracefd, " PC   ST       INST           CODE\n");
      }
      printinst (outbuf, trcidx);
      fprintf (tracefd, "%s\n", outbuf);
      traceline++;
   }
   trcidx++;

   if (error < 0)
   {
#ifdef DEBUGERRINT
      fprintf (stderr,
	       "%04X: Illegal instruction: inst = >%04X, pc = >%04X\n",
	       curpc, inst, curpc);
         
      runled = FALSE;
#endif
      geninterrupt (INTINTLOC, 0);
   }

   instcount++;
}

/***********************************************************************
* stepprog - Step through a program.
***********************************************************************/

void 
stepprog (void)
{
   uint16 inst;

   inst = getinst (TRUE);
   INCPC (1);
   runinst (inst);
}

/***********************************************************************
* runprog - Run a program.
***********************************************************************/

void
runprog (void)
{
   ttrunstate ();

   deferint = 0;
   while (run)
   {
      idle = FALSE;
      runled = TRUE;
      stepprog ();

      if (breakflag && GETPC == breakaddr)
      {
         run = FALSE;
         runled = FALSE;
	 showstatus = TRUE;
      }
      else if (run)
      {
	 if (enablepanel && pancount) panel ();
	 if (!deferint)
	    checkinterrupts ();
	 else
	    deferint--;
	 if (delayclock)
	 {
	    long i, j;
	    j = delayclock;
	    for (i = 0; i < j; i++) ;
	 }
      }
   }
   if (enablepanel) panel ();

   ttcmdstate ();
}
