/***********************************************************************
*
* simsdsk.c - SCSI Disk IO processing for the TI 980 Simulator.
*
* Changes:
*   08/23/16   DGP   Cloned SCSI disk support.
*
* The disk is formatted as follows:
*    4 bytes (big-endian): # blocks
*    4 bytes (big-endian): disk type << 16 | 1
*    4 bytes (big-endian): overhead << 8 | 1
*    4 bytes (big-endian): # bytes/block
*    N bytes: raw data for each block - the order is:
*      block 0 
*      block 1
*      block 2
*      block 3 ... 
*
***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <errno.h>

#if defined (WIN32)
#define __TTYROUTINES 0
#include <conio.h>
#include <windows.h>
#include <signal.h>
#endif

#if defined(UNIX)
#include <sys/time.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <pthread.h>
#endif

#include "simdef.h"

#define EXTERN extern
#include "simdsk.h"

extern uint16 statreg;	/* The program status register */

extern int run;
extern int devcnt;
extern int model;
extern uint32 memlen;
extern unsigned long instcount;

extern char view[MAXVIEW][MAXVIEWLEN+1];
extern uint16 regs[16];
extern uint16 memory[MAXMEMSIZE];
extern Device devices[MAXDEVICES];

/*
** Disk status word bits
*/

#define WPERR	0x4000		/* Write Protect Error */
#define XFRERR  0x1000		/* Data transfer error */
#define EODERR  0x0800		/* End of Disk error */
#define PGMERR	0x0400		/* Program error */
#define CMPERR  0x0200		/* Compare error */
#define PARERR  0x0100		/* Parity error */
#define IDERR   0x0080		/* ID error */
#define DEVBSY  0x0040		/* Device Busy */
#define CHNTKN  0x0020		/* Chain taken */
#define WPBIT	0x0010		/* Write Protect */

/***********************************************************************
* sdskreadint - Read an integer.
***********************************************************************/

static int
sdskreadint (FILE *fd)
{
   int r;
   int i;

   r = 0;
   for (i = 0; i < 4; i++)
   {
      int c;
      if ((c = fgetc (fd)) < 0)
      {
	 sprintf (view[0], "sdskreadint: read failed: %s", strerror (ERRNO));
         return (-1);
      }
      r = (r << 8) | (c & 0xFF);
   }
   return (r);
}

/***********************************************************************
* sdskopen - Open disk and read geometry.
***********************************************************************/

int 
sdskopen (Device *dev)
{
   int diskdata;

#ifdef DEBUGSDSK
   fprintf (stderr, "sdskopen: ENTERED\n");
#endif

   if (dev->infd == DEV_NULL)
   {
      return (0);
   }

   /*
   ** Seek to beginning of the disk
   */

   if (fseek (dev->infd, 0, SEEK_SET) < 0)
   {
      sprintf (view[0], "sdskopen: seek failed: %s", strerror (ERRNO));
      snprintf (view[1], MAXVIEWLEN,"filename: %s", dev->file);
      return (-1);
   }

   /*
   ** Read the disk geometry
   */

   if ((dev->info.dskinfo.cyls = sdskreadint (dev->infd)) < 0) return (-1);
   if ((diskdata = sdskreadint (dev->infd)) < 0) return (-1);
   dev->info.dskinfo.type = (diskdata >> 28) & 0xF;
   dev->info.dskinfo.model = (diskdata >> 16) & 0xFFF;
   dev->info.dskinfo.heads = diskdata & 0xFFFF;
   if ((diskdata = sdskreadint (dev->infd)) < 0) return (-1);
   dev->info.dskinfo.overhead = 0;
   dev->info.dskinfo.secttrk = diskdata & 0xFFFF;
   if ((dev->info.dskinfo.sectlen = sdskreadint (dev->infd)) < 0) return (-1);
	       
   /*
   ** Check geometry against model
   */

   if ((dev->info.dskinfo.model > MAXDISKS) ||
       (dev->info.dskinfo.type != DISK_SCSI) ||
       (dev->info.dskinfo.type != disks[dev->info.dskinfo.model].type) ||
       (dev->info.dskinfo.cyls != disks[dev->info.dskinfo.model].cyls) ||
       (dev->info.dskinfo.heads != disks[dev->info.dskinfo.model].heads) ||
       (dev->info.dskinfo.secttrk != disks[dev->info.dskinfo.model].sectrk) ||
       (dev->info.dskinfo.sectlen != disks[dev->info.dskinfo.model].bytsec) ||
       (dev->info.dskinfo.overhead != disks[dev->info.dskinfo.model].overhead))
   {
      sprintf (view[0], "Unknown disk type or geometry mismatch");
      snprintf (view[1], MAXVIEWLEN, "filename: %s", dev->file);
      sprintf (view[2], "   model     = %d", dev->info.dskinfo.model);
      sprintf (view[3], "   type      = %d", dev->info.dskinfo.type);
      sprintf (view[4], "   cyls      = %d", dev->info.dskinfo.cyls);
      sprintf (view[5], "   heads     = %d", dev->info.dskinfo.heads);
      sprintf (view[6], "   secttrk   = %d", dev->info.dskinfo.secttrk);
      sprintf (view[7], "   sectlen   = %d", dev->info.dskinfo.sectlen);
      sprintf (view[8], "   overhead  = %d", dev->info.dskinfo.overhead);
      return (-1);
   }

#ifdef DEBUGSDSK
   fprintf (stderr, "Disk geometry:\n");
   fprintf (stderr, "  model     = %s\n", disks[dev->info.dskinfo.model].model);
   fprintf (stderr, "  type      = %d\n", dev->info.dskinfo.type);
   fprintf (stderr, "  blocks    = %d\n", dev->info.dskinfo.cyls);
   fprintf (stderr, "  blklen    = %d\n", dev->info.dskinfo.sectlen);
   fprintf (stderr, "  disk size = %d bytes\n", dev->info.dskinfo.sectlen *
	       dev->info.dskinfo.cyls);
#endif

   return (0);
}

/***********************************************************************
* sdskreset - Reset device
***********************************************************************/

void
sdskreset (Device *dev)
{
#ifdef DEBUGSDSK
   fprintf (stderr, "sdskreset: ENTERED\n");
#endif

   return;
}

/***********************************************************************
* sdskseek - Seek to specified disk location.
***********************************************************************/

static int
sdskseek (Device *dev, int block, int statloc)
{
   int dskloc;

   /*
   ** Ensure block is OK
   */

   memory[statloc] = 0;
   if (block >= dev->info.dskinfo.cyls)
   {
#ifdef DEBUGSDSK
      fprintf (stderr, "sdskseek: ERROR: block = %d\n", block);
#endif
      memory[statloc] |= EODERR | ((dev->switches & SWPROTECT) ? WPBIT : 0);
      return (-1);
   }

   /*
   ** Calculate location
   */

   dskloc = (block * dev->info.dskinfo.sectlen) + DSKOVERHEAD;
#ifdef DEBUGSDSK
   fprintf (stderr, "sdskseek: block = %d, dskloc = %d\n",
	    block, dskloc);
#endif

   /*
   ** Go to the specified location on disk
   */

   if (fseek (dev->infd, dskloc, SEEK_SET) < 0)
   {
      sprintf (view[0], "sdskseek: seek failed: %s", strerror (ERRNO));
      snprintf (view[1], MAXVIEWLEN, "filename: %s", dev->file);
      return (-1);
   }

   return (0);
}

/***********************************************************************
* sdskbumpblock - Bump to next block
***********************************************************************/

static void 
sdskbumpblock (Device *dev, int *block)
{
   (*block)++;
   if (*block == dev->info.dskinfo.cyls)
   {
      *block = 0;
   }
}

/***********************************************************************
* sdskdocmd - Do disk command
***********************************************************************/

int 
sdskdocmd (Device *dev, uint16 device, uint16 word2)
{
   Device *udev;
   int i;
   uint16 statloc;
   uint16 addr;
   uint16 len;
   uint16 chain;
   uint16 cmd;
   uint16 unit;
   uint16 genint;

   statloc = ((device & 0x0007) << 1) + 0x0098;
   memory[statloc] = 0;
   memory[statloc+1] = word2;

   unit = (device & 0x0030) >> 4;
   cmd = (device >> 7) & 1;
   genint = 0;

#ifdef DEBUGSDSK
   fprintf (stderr,
   	    "sdskdocmd: device = >%04X, word2 = >%04X, unit = %d, cmd = %d"
	    ", statloc = >%04X\n",
	    device, word2, unit, cmd, statloc);
#endif
   if (cmd)
      goto dsk_exit;

   /*
   ** Find selected unit
   */

   for (i = 0; i < devcnt; i++) /* find selected unit */
      if (devices[i].devaddr == dev->devaddr && devices[i].unit == unit)
	 break;
   if (i == devcnt)
   {
#ifdef DEBUGSDSK
      fprintf (stderr, "Unit not configured\n", unit);
#endif
      /* Drive not ready */
      memory[statloc] = OPCOMP | OFFLIN;
      goto dsk_exit;
   }
   udev = &devices[i];

   do
   {
      int i;
      int block;
      uint16 subcmd;
      uint16 blkcnt;
      uint16 d, c;

      memory[statloc+1] = word2;
      subcmd = getmem (word2);

      /*
      ** Set initial sate
      */

      udev->select = TRUE;
      chain = 0;

      switch (subcmd)
      {
      case 0: /* Get Info */
	 addr = getmem (word2 + 1);
	 len = getmem (word2 + 2);
#ifdef DEBUGSDSK
	 fprintf (stderr, "   GETINFO: addr = >%04X, len = %d\n",
		  addr, len);
#endif
	 if (len--)
	    dmaputword (udev, addr, 1);
	 if (len--)
	    dmaputword (udev, addr+1, 0x1234);
	 if (len--)
	    dmaputword (udev, addr+2, 0x5600 | statloc);
	 if (len--)
	    dmaputword (udev, addr+3, 0x789A);
         break;

      case 3: /* Inquire */
	 chain = getmem (word2 + 2);
	 addr = getmem (word2 + 5);
	 len = getmem (word2 + 6);
	 
#ifdef DEBUGSDSK
	 fprintf (stderr, "   INQUIRE: addr = >%04X, len = %d\n",
		  addr, len);
#endif
	 if (len--)
	    dmaputword (udev, addr, 0x0000);
	 if (len--)
	    dmaputword (udev, addr+1, udev->info.dskinfo.cyls >> 16);
	 if (len--)
	    dmaputword (udev, addr+2, udev->info.dskinfo.cyls & 0xFFFF);
	 if (len--)
	    dmaputword (udev, addr+3, 0);
	 if (len--)
	    dmaputword (udev, addr+4, udev->info.dskinfo.sectlen);
         break;

      case 4: /* Read */
	 chain = getmem (word2 + 2);
	 addr = getmem (word2 + 5);
	 blkcnt = getmem (word2 + 6);
	 block = (getmem (word2 + 7) << 16) | getmem (word2 + 8);
	 len = udev->info.dskinfo.sectlen * blkcnt;
	 
#ifdef DEBUGSDSK
	 fprintf (stderr, "   READ: block = %d, addr = >%04X, len = %d\n",
		  block, addr, len);
#endif
	 if (sdskseek (udev, block, statloc) < 0)
	 {
	    memory[statloc] |= OPCOMP | PGMERR |
			((udev->switches & SWPROTECT) ? WPBIT : 0);
	    return (genint);
	 }

	 while (len)
	 {
	    int slen;

	    /*
	    ** Read one block at a time
	    */

	    slen = (len < udev->info.dskinfo.sectlen) ?
		  len : udev->info.dskinfo.sectlen;

	    for (i = 0; i < slen; i += 2)
	    {
	       d = fgetc (udev->infd);
	       c = fgetc (udev->infd);
	       d = (d << 8) | c;
	       dmaputword (udev, addr, d);
	       addr++;
	    }

	    len -= slen;
	    sdskbumpblock (udev, &block);

	    if (len)
	       if (sdskseek (udev, block, statloc) < 0)
		  return (genint);
	 }

	 udev->info.dskinfo.curcyl = block;
         break;

      case 5: /* Write */
	 chain = getmem (word2 + 2);
	 addr = getmem (word2 + 5);
	 blkcnt = getmem (word2 + 6);
	 block = (getmem (word2 + 7) << 16) | getmem (word2 + 8);
	 len = udev->info.dskinfo.sectlen * blkcnt;
	 
#ifdef DEBUGSDSK
	 fprintf (stderr, "   WRITE: block = %d, addr = >%04X, len = %d\n",
		  block, addr, len);
#endif
	 if (sdskseek (udev, block, statloc) < 0)
	 {
	    memory[statloc] |= OPCOMP | PGMERR |
			((udev->switches & SWPROTECT) ? WPBIT : 0);
	    return (genint);
	 }

	 if (!(udev->switches & SWPROTECT));
	 {
	    while (len)
	    {
	       int slen;

	       /*
	       ** Write one block at a time
	       */

	       slen = (len < udev->info.dskinfo.sectlen) ?
		     len : udev->info.dskinfo.sectlen;

	       for (i = 0; i < slen; i += 2)
	       {
		  d = dmagetword (udev, addr);
		  fputc ((d >> 8) & 0xFF, udev->infd);
		  fputc (d & 0xFF, udev->infd);
		  addr++;
	       }

	       /*
	       ** Zero fill short blocks
	       */

	       for (; i < udev->info.dskinfo.sectlen; i++)
		  fputc ('\0', udev->infd);

	       len -= slen;
	       sdskbumpblock (udev, &block);

	       if (len)
		  if (sdskseek (udev, block, statloc) < 0)
		     return (genint);
	    }

	    udev->info.dskinfo.curcyl = block;
	 }
         break;

      case 12: /* FORMAT */
	 chain = getmem (word2 + 2);
	 
#ifdef DEBUGSDSK
	 fprintf (stderr, "   FORMAT: \n");
	 for (i = 0; i < 5; i++)
	    fprintf (stderr, "      pad[%d] = >%04X\n",
		     i, getmem (word2 + i + 5));
#endif
	 if (sdskseek (udev, block, statloc) < 0)
	 {
	    memory[statloc] |= OPCOMP | PGMERR |
			((udev->switches & SWPROTECT) ? WPBIT : 0);
	    return (genint);
	 }

         break;

      default: ;
#ifdef DEBUGSDSK
	 fprintf (stderr, "   UNKNOWN: block = %d, len = %d\n",
		  block, len);
#endif
      }
      if (chain)
         word2 = chain;
   } while (chain);

dsk_exit:
   memory[statloc] |= OPCOMP;
   return (genint);
}

/***********************************************************************
* sdskboot - Boot from device.
***********************************************************************/

int
sdskboot (Device *dev)
{
   int i;
   int addr;
   uint16 d, c;

   if (fseek (dev->infd, DSKOVERHEAD, SEEK_SET) < 0)
   {
      sprintf (view[0], "Device %s: seek failed: %s",
	       dev->name, strerror (ERRNO));
      sprintf (view[1], "filename: %s", dev->file);
      return (-1);
   }

   addr = 0;
   SETPC (addr);
   for (i = 0; i < dev->info.dskinfo.sectlen; i += 2)
   {
      d = fgetc (dev->infd);
      c = fgetc (dev->infd);
      d = (d << 8) | c;
      dmaputword (dev, addr, d);
      addr++;
   }
   run = TRUE;

   return (0);
}
