/***********************************************************************
*
* simtape.c - Tape IO processing for the TI 980 Simulator.
*
* Changes:
*   06/24/11   DGP   Original.
*   08/31/11   DGP   Fix up interrupts.
*
* Tape record format:
*   4 bytes (little-endian): block length
*   N bytes: raw data for the record
*   4 bytes (little-endian): repeat block length
*
* Tape mark (EOF):
*   4 bytes: zero data
*
***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <errno.h>

#if defined (WIN32)
#define __TTYROUTINES 0
#include <conio.h>
#include <windows.h>
#include <signal.h>
#endif

#if defined(UNIX)
#include <sys/time.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <pthread.h>
#endif

#include "simdef.h"

extern uint16 statreg;	/* The program status register */

extern int run;
extern int devcnt;
extern uint32 memlen;

extern char view[MAXVIEW][MAXVIEWLEN+1];
extern uint16 regs[16];
extern uint16 memory[MAXMEMSIZE];
extern Device devices[MAXDEVICES];

/*
** Tape status word bits
*/

#define WPBIT	0x4000		/* Write Protect */
#define CTLBUSY 0x1000		/* Controller busy */
#define DMAPER  0x0800		/* DMA Parity error */
#define TAPPER  0x0400		/* Tape Parity error */
#define EORBIT	0x0200		/* End Of Record */
#define EOFBIT	0x0100		/* End Of File */
#define EOTBIT	0x0080		/* End Of Tape */
#define BOTBIT	0x0040		/* Beginning Of Tape */
#define ERRBIT	0x0020		/* Error */

/***********************************************************************
* tapereadint - Read an integer.
***********************************************************************/

static int
tapereadint (FILE *fd)
{
   int r;

   r = fgetc (fd);
   r = r | (fgetc (fd) << 8);
   r = r | (fgetc (fd) << 16);
   r = r | (fgetc (fd) << 24);
   if (feof(fd)) return (-2);
   return (r & 0x7FFF);
}

/***********************************************************************
* tapewriteint - Write an integer.
***********************************************************************/

static void
tapewriteint (FILE *fd, int v)
{
   v &= 0x7FFF;
   fputc (v & 0xFF, fd);
   fputc ((v >> 8) & 0xFF, fd);
   fputc ((v >> 16) & 0xFF, fd);
   fputc ((v >> 24) & 0xFF, fd);
}

/***********************************************************************
* tapeopen - Open tape.
***********************************************************************/

int 
tapeopen (Device *dev)
{
#ifdef DEBUGTAPE
   fprintf (stderr, "tapeopen: ENTERED\n");
   fprintf (stderr, "   file = %s\n", dev->file);
#endif

   if (dev->infd == DEV_NULL)
   {
      return (0);
   }

   /*
   ** Seek to beginning of the tape file.
   */

   if (fseek (dev->infd, 0, SEEK_SET) < 0)
   {
      sprintf (view[0], "tapeopen: seek failed: %s", strerror (ERRNO));
      sprintf (view[1], "filename: %s", dev->file);
      return (-1);
   }

   return (0);
}

/***********************************************************************
* tapereset - Reset device
***********************************************************************/

void
tapereset (Device *dev)
{
#ifdef DEBUGTAPE
   fprintf (stderr, "tapereset: ENTERED\n");
#endif

   return;
}

/***********************************************************************
* tapedocmd - Do tape command
***********************************************************************/

int 
tapedocmd (Device *dev, uint16 device, uint16 word2)
{
   Device *udev;
   int i;
   int reclen;
   uint16 statloc;
   uint16 addr;
   uint16 length;
   uint16 chain;
   uint16 cmd;
   uint16 unit;
   uint16 genint;

   statloc = ((device & 0x0007) << 1) + 0x0098;
   memory[statloc] = 0;
   memory[statloc+1] = word2;
   genint = 0;

   cmd = (device & 0x00C0) >> 6;
   unit = (device & 0x0018) >> 3;

#ifdef DEBUGTAPE
   fprintf (stderr,
   	    "tapedocmd: device = %d, word2 = >%04X, unit = %d, cmd = %d\n",
	    device & 7, word2, unit, cmd);
   fprintf (stderr, "   statloc = >%04X\n", statloc);
#endif
   if (unit == 0)
      goto tape_exit;

   /*
   ** Find selected unit
   */

   for (i = 0; i < devcnt; i++) /* find selected unit */
      if (devices[i].devaddr == dev->devaddr && devices[i].unit == (unit - 1))
	 break;
   if (i == devcnt)
   {
#ifdef DEBUGTAPE
      fprintf (stderr, "Unit not configured\n", unit);
#endif
      /* Drive not ready */
      memory[statreg] |= OFFLIN;
      goto tape_exit;
   }
   udev = &devices[i];

   if ((cmd == 0) || (cmd == 2))
   {
#ifdef DEBUGTAPE
      fprintf (stderr, "   REWIND: \n");
#endif
      udev->info.tapeinfo.curpos = 0;

      if (fseek (udev->infd, 0, SEEK_SET) < 0)
      {
#ifdef DEBUGTAPE
	 sprintf (view[0], "tapedocmd: REWIND: seek failed: %s",
		  strerror (ERRNO));
	 snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
#endif
	 memory[statloc] |= ERRBIT;
	 goto tape_exit;
      }
      memory[statloc] |= (BOTBIT | (1 << ((3 - unit) + 2)));
      if (cmd == 0)
         genint = 1;
   }

   else if (cmd == 1)
   {
#ifdef DEBUGTAPE
      fprintf (stderr, "   REWUNLOAD: \n");
#endif
      udev->info.tapeinfo.curpos = 0;
      if (fseek (udev->infd, 0, SEEK_SET) < 0)
      {
#ifdef DEBUGTAPE
	 sprintf (view[0], "tapedocmd: REWUNLOAD: seek failed: %s",
		  strerror (ERRNO));
	 snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
#endif
	 memory[statloc] |= ERRBIT;
      }
      genint = 1;
   }

   else do
   {
      int i, j;
      uint16 reverse;
      uint16 subcmd;
      uint16 w2val;

      memory[statloc+1] = word2;
      addr = getmem (word2);
      length = getmem (word2+1);
      w2val = getmem (word2+2);
      genint = w2val & 0x8000;
      reverse = w2val & 0x4000;
      chain = w2val & 0x2000;
      subcmd = (w2val >> 10) & 7;
#ifdef DEBUGTAPE
      fprintf (stderr,
      	       "   subcmd = %d, buffer = >%04X, length = %d, chain = %x\n",
      	       subcmd, addr, length, chain);
#endif

      switch (subcmd)
      {
      case 0: /* Read */
	 reclen = tapereadint (udev->infd);
#ifdef DEBUGTAPE
	 fprintf (stderr, "   READ: reclen = %d\n", reclen);
	 fprintf (stderr, "   curpos = %ld\n", udev->info.tapeinfo.curpos);
#endif

	 if (reclen > 0)
	 {
	    for (j = i = 0; i < reclen;)
	    {
	       uint16 d, c;

	       d = fgetc (udev->infd);
	       i++;
	       if (i < reclen)
	       {
		  c = fgetc (udev->infd);
		  i++;
	       }
	       else
		  c = 0xFF;
	       d = (d << 8) | c;
	       if (j < length)
	       {
		  dmaputword (dev, addr, d);
		  j += 2;
		  addr++;
	       }
	    }
	    if (i < length)
	    {
	       length -= i;
	       memory[statloc] |= EORBIT;
	    }
	    else length = 0;
	    memory[statloc+1] = length;
	    i = tapereadint (udev->infd);
	    udev->info.tapeinfo.curpos += reclen + 8;
#ifdef DEBUGTAPE
	    fprintf (stderr, "   length = %d\n", length);
	    fprintf (stderr, "   traillen = %d\n", i);
#endif
	    if (i != reclen)
	    {
	       memory[statloc] |= ERRBIT;
	    }

#ifdef DEBUGTAPEDATA
	    HEXDUMP (stderr, &memory[saddr], j, saddr);
#endif
	 }
	 else if (reclen == 0)
	 {
	    memory[statloc] |= EOFBIT;
	 }
	 else if (reclen == -2)
	 {
	    memory[statloc] |= EOTBIT;
	 }
         break;

      case 1: /* Skip */
#ifdef DEBUGTAPE
	 fprintf (stderr, "   SKIP: reverse = %d\n", reverse);
	 fprintf (stderr, "   curpos = %ld\n", udev->info.tapeinfo.curpos);
#endif

	 if (reverse)
	 {
	    udev->info.tapeinfo.curpos -= 4;
	    if (fseek (udev->infd, udev->info.tapeinfo.curpos, SEEK_SET) < 0)
	    {
#ifdef DEBUGTAPE
	       sprintf (view[0], "tapedocmd: SKIP: seek-1 failed: %s",
			strerror (ERRNO));
	       snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
#endif
	       memory[statloc] |= ERRBIT;
	       goto tape_exit;
	    }
	 }
	 reclen = tapereadint (udev->infd);
	 if (reclen > 0)
	 {
	    if (reverse)
	       udev->info.tapeinfo.curpos -= reclen + 4;
	    else
	       udev->info.tapeinfo.curpos += reclen + 4;
	    if (fseek (udev->infd, udev->info.tapeinfo.curpos, SEEK_SET) < 0)
	    {
#ifdef DEBUGTAPE
	       sprintf (view[0], "tapedocmd: SKIP: seek-2 failed: %s",
			strerror (ERRNO));
	       snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
#endif
	       memory[statloc] |= ERRBIT;
	       goto tape_exit;
	    }
	    i = tapereadint (udev->infd);
	    if (reverse)
	    {
	       if (fseek (udev->infd, udev->info.tapeinfo.curpos, SEEK_SET) < 0)
	       {
#ifdef DEBUGTAPE
		  sprintf (view[0], "tapedocmd: SKIP: seek-3 failed: %s",
			   strerror (ERRNO));
		  snprintf (view[1], MAXVIEWLEN, "filename: %s", udev->file);
#endif
		  memory[statloc] |= ERRBIT;
		  goto tape_exit;
	       }
	    }
	    else
	       udev->info.tapeinfo.curpos += 4;
#ifdef DEBUGTAPE
	    fprintf (stderr, "   length = %d\n", length);
	    fprintf (stderr, "   traillen = %d\n", i);
#endif
	    if (i != reclen)
	    {
	       memory[statloc] |= ERRBIT;
	    }

	 }
	 else if (reclen == 0)
	 {
	    memory[statloc] |= EOFBIT;
	 }
	 else if (reclen == -2)
	 {
	    memory[statloc] |= EOTBIT;
	 }
         break;

      case 4: /* Write */
#ifdef DEBUGTAPE
	 fprintf (stderr, "   WRITE: \n");
	 fprintf (stderr, "   curpos = %ld\n", udev->info.tapeinfo.curpos);
#endif
	 if (!(udev->switches & SWPROTECT))
	 {
#ifdef DEBUGTAPEDATA
	    HEXDUMP (stderr, &memory[saddr], GETLEN(udev), saddr);
#endif
	    tapewriteint (udev->infd, length);

	    for (i = 0; i < length; i += 2)
	    {
	       uint16 d;

	       d = dmagetword (dev, addr);
	       fputc ((d >> 8) & 0xFF, udev->infd);
	       fputc (d & 0xFF, udev->infd);
	       addr++;
	    }
	    tapewriteint (udev->infd, length);
	    memory[statloc+1] = 0;
	    udev->info.tapeinfo.curpos += length + 8;
	 }
         break;

      case 5: /* Erase */
         break;

      case 6: /* Write EOF */
#ifdef DEBUGTAPE
	 fprintf (stderr, "   WRITEEOF: \n");
#endif
	 if (!(udev->switches & SWPROTECT))
	 {
	    tapewriteint (udev->infd, 0);
	    udev->info.tapeinfo.curpos += 4;
	 }
         break;

      default:
	 memory[statloc] |= ERRBIT;
      }
      if (chain)
         word2 = getmem (word2+3);
   } while (chain);

tape_exit:
   memory[statloc] |= OPCOMP;
   return (genint);
}

/***********************************************************************
* tapeboot - Boot from device.
***********************************************************************/

int
tapeboot (Device *dev)
{
   return (0);
}

