/***********************************************************************
*
* simterm.c - Terminal processing for the TI 980 Simulator.
*
* Changes:
*   08/25/16   DGP   Split from simio.c
*
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>

#if defined (WIN32)
#define __TTYROUTINES 0
#include <conio.h>
#include <windows.h>
#include <signal.h>
#include <process.h>
#ifndef HZ
#define HZ 100
#endif
#endif

#if defined(UNIX)
#include <unistd.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <pthread.h>
#endif

#include "simdef.h"
#include "screen.h"

extern int run;
extern int enablepanel;
extern int viewlen;
extern int prtviewlen;
extern int prtviewcol;
extern int windowlen;

extern char view[MAXVIEW][MAXVIEWLEN+1];
extern char prtview[MAXVIEW][MAXVIEWLEN+1];

static int sim_int_char = 005;

/***********************************************************************
* Terminal routines borrowed from Bub Supnik's simh
***********************************************************************/

#if defined (WIN32)
static volatile int sim_win_ctlc = 0;

static void
win_handler (int sig)
{
   sim_win_ctlc = 1;
   return;
}

int
ttinit (void)
{
   return 0;
}

int
ttrunstate (void)
{
   if (signal (SIGINT, win_handler) == SIG_ERR) return -1;
   SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_BELOW_NORMAL);
   return 0;
}

int
ttcmdstate (void)
{
   SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_NORMAL);
   return 0;
}

int
ttclose (void)
{
   return 0;
}

static int
os_getchar (void)
{
   int c;

   if (sim_win_ctlc)
   {
      sim_win_ctlc = 0;
      signal (SIGINT, win_handler);
      return 003;
   }
   if (!kbhit ()) return EOF;
   c = _getch ();
   if ((c & 0177) == '\b') c = 0177;
   if ((c & 0177) == sim_int_char)
   {
      run = FALSE;
      return EOF;
   }
   return c;
}

static int
os_putchar (uint8 c)
{
   c &= 0x7F;
   if (c != 0x7F && c != 0)
      _putch (c);
   return 0;
}
#endif

#if defined(UNIX) || defined(MINGW)
struct termios cmdtty, runtty;

int
ttinit (void)
{
   if (!isatty (fileno (stdin))) return 0;		/* skip if !tty */
   if (tcgetattr (0, &cmdtty) < 0) return -1;		/* get old flags */
   runtty = cmdtty;
   runtty.c_lflag = runtty.c_lflag & ~(ECHO | ICANON);	/* no echo or edit */
   runtty.c_oflag = runtty.c_oflag & ~OPOST;		/* no output edit */
   runtty.c_iflag = runtty.c_iflag & ~ICRNL;		/* no cr conversion */
   runtty.c_cc[VINTR] = sim_int_char;			/* interrupt */
   runtty.c_cc[VQUIT] = 0;				/* no quit */
   runtty.c_cc[VERASE] = 0;
   runtty.c_cc[VKILL] = 0;
   runtty.c_cc[VEOF] = 0;
   runtty.c_cc[VEOL] = 0;
   runtty.c_cc[VSTART] = 0;				/* no host sync */
   runtty.c_cc[VSUSP] = 0;
   runtty.c_cc[VSTOP] = 0;
#if defined (VREPRINT)
   runtty.c_cc[VREPRINT] = 0;				/* no specials */
#endif
#if defined (VDISCARD)
   runtty.c_cc[VDISCARD] = 0;
#endif
#if defined (VWERASE)
   runtty.c_cc[VWERASE] = 0;
#endif
#if defined (VLNEXT)
   runtty.c_cc[VLNEXT] = 0;
#endif
   runtty.c_cc[VMIN] = 0;				/* no waiting */
   runtty.c_cc[VTIME] = 0;
#if defined (VDSUSP)
   runtty.c_cc[VDSUSP] = 0;
#endif
#if defined (VSTATUS)
   runtty.c_cc[VSTATUS] = 0;
#endif
   return 0;
}

int
ttrunstate (void)
{
   if (!isatty (fileno (stdin))) return 0;		/* skip if !tty */
   runtty.c_cc[VINTR] = sim_int_char;			/* in case changed */
   if (tcsetattr (0, TCSAFLUSH, &runtty) < 0) return -1;
   return 0;
}

int
ttcmdstate (void)
{
   if (!isatty (fileno (stdin))) return 0;		/* skip if !tty */
   if (tcsetattr (0, TCSAFLUSH, &cmdtty) < 0) return -1;
   return 0;
}

int
ttclose (void)
{
   return ttcmdstate ();
}

static int
os_getchar (void)
{
   int status;
   unsigned char buf[1];

   status = read (0, buf, 1);
   if (status != 1) return EOF;
   if (buf[0] == sim_int_char)
   {
      run = FALSE;
      return (EOF);
   }
   return (buf[0]);
}

static int
os_putchar (uint8 out)
{
   char c;

   c = out & 0x7F;
   if (c != 0x7F && c != 0)
      write (1, &c, 1);
   return 0;
}
#endif

int 
serialpollchar (Device *dev, long delay)
{
   int c;
#if defined(UNIX)
   int ifd;
   struct timeval tv;
   fd_set readfds;
#endif

   if (delay > 0)
   {
#if defined(WIN32)
      Sleep (delay/10);
      c = 1;
#endif
#if defined(UNIX)
      ifd = fileno (dev->infd);
      FD_ZERO (&readfds);

      FD_SET (ifd, &readfds);

      tv.tv_sec = 0;
      tv.tv_usec = delay;
      if ((c = select (ifd+1, &readfds, NULL, NULL, &tv)) < 0)
      {
	 if (ERRNO != EINTR && ERRNO != EAGAIN)
	    PERROR ("poll_kbd: select failed");
      }
#endif
   }
   else c = 1;

   if (c > 0)
   {
      if (dev->infd == stdin)
      {
	 c = os_getchar ();
	 if (c == 0x03)
	 {
	    dev->cbreak = TRUE;
	    c = EOF;
	 }
      }
      else
      {
	 c = fgetc (dev->infd);
      }
   }
   else
      c = EOF;

   return (c);
}

/***********************************************************************
* prtpanel - Prints messages on screen.
***********************************************************************/

void
prtpanel (int index, int refresh)
{
   int i, j;

   prtviewcol = index + 1;
   if (refresh || index == 0)
   {
      for (j = 0; j < prtviewlen; j++)
      {
	 screenposition (enablepanel, OUTPUTLINE+j,1);
	 clearline (enablepanel);
	 if (!refresh)
	    strcpy (prtview[j], prtview[j+1]);
	 printf ("%s", prtview[j]);
      }
#ifdef DEBUGCONPANEL
      fprintf (stderr, "prtpanel: \n");
      fprintf (stderr, "%s", prtview[j-2]);
      HEXDUMP (stderr, prtview[j-2], 80, 0);
#endif
   }
   i = prtviewlen - 1;
   screenposition (enablepanel, OUTPUTLINE+i, 1);
   clearline (enablepanel);
   printf ("%s", prtview[i]);
   screenposition (enablepanel, OUTPUTLINE+i, prtviewcol);

}

/***********************************************************************
* putpanel - Puts chars in the "panel"
***********************************************************************/

static void
putpanel (Device *dev, uint8 ch)
{
   ch &= 0x7F;
   if (ch > 0)
   {
      if (ch == 0x08)
      {
	 dev->viewndx--;
      }
      else if (ch > 0x1F)
      {
	 prtview[prtviewlen - 1][dev->viewndx++] = ch; 
      }
      if (ch != 0x0D)
      {
	 prtview[prtviewlen - 1][dev->viewndx] = '\0';
	 if (dev->viewndx > MAXVIEWLEN || ch == 0x0A)
	 {
	    dev->viewndx = 0;
	 }
	 prtpanel (dev->viewndx, FALSE);
      }
   }
}

/***********************************************************************
* putview - Puts messages on screen.
***********************************************************************/

void
putview (int interactive, int wait)
{
   int i, j;
   int ch;

   if (interactive && enablepanel)
      for (j = 0; j < viewlen; j++)
      {
         screenposition (enablepanel, OUTPUTLINE + j, 1);
         clearline (enablepanel);
      }

   for (j = i = 0; i < MAXVIEW; i++)
   {
      if (interactive && enablepanel)
      {
         j++;
         if (j >= viewlen)
         {
            screenposition (enablepanel, windowlen-2, 1);
            clearline(enablepanel);
            printf ("press enter to continue:");
            fflush(stdout);

            ch = fgetc (stdin);
            if (ch == 'q')
	    {
	       while ((ch = fgetc (stdin)) != '\n') ;
	       return;
	    }

	    if (interactive && enablepanel)
	       for (j = 0; j < viewlen; j++)
	       {
		  screenposition (enablepanel, OUTPUTLINE + j, 1);
		  clearline (enablepanel);
	       }
            j = 0;
         }

         screenposition (enablepanel, OUTPUTLINE + j, 1);
         clearline (enablepanel);
      }
      if (view[i][0])
      {
         printf ("%s", view[i]);
         if (!enablepanel) printf ("\n");
      }
      else return;
      view[i][0] = '\0';
   }
   fflush(stdout);

   if (wait && interactive && enablepanel)
   {
      screenposition (enablepanel, windowlen-2, 1);
      clearline (enablepanel);
      printf ("press enter to continue:");
      fflush (stdout);

      ch = fgetc (stdin);
   }
}

/***********************************************************************
* serialgetchar - Gets a char from a non-Telnet serial device.
***********************************************************************/

int
serialgetchar (Device *dev)
{
   int c;

   if (dev->infd == stdin)
   {
      c = os_getchar ();
      if (c == 0x03)
      {
	 c = EOF;
      }
   }
   else
   {
      c = fgetc (dev->infd);
   }

   return (c);
}

/***********************************************************************
* serialputchar - Puts a char on non-Telnet serial device.
***********************************************************************/

void
serialputchar (Device *dev, uint16 dat)
{
   uint8 ch;

   ch = (uint8)dat;

#ifdef DEBUGSERIALPUT
   fprintf (stderr, "serialputchar: c = >%02X(%c)\n", dat,
	    isprint (dat) ? dat : '.');
#endif

   if (enablepanel && dev->outfd == stdout)
   {
      putpanel (dev, ch);
   }
   else
   {
      if (dev->outfd == stdout)
      {
	 os_putchar (ch);
      }
      else
      {
	 fputc (ch, dev->outfd);
      }
      fflush (dev->outfd);
   }
}
